//                              -*- Mode: C++ -*-
//: EALinkAnalyzer.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 13/8/97
// Last Update ........
//

#ifndef EALINKANALYZER_HH_
#define EALINKANALYZER_HH_

#include "arte/arte.hh"
#include "care/EAClusterAnalyzer.hh"

/**
class EALinkAnalyzer.
  Class for defining and updating the links between the
  RCCL bank and the banks HITC and RTRA.
*/

class EALinkAnalyzer : public EAClusterAnalyzer {
public:
  /// Constructor
  EALinkAnalyzer();
  /// Destructor
  virtual ~EALinkAnalyzer();

  /**@name Beheavioral methods */
  //@{ 
  /// EMClusterMember Analyzer.
  virtual void doAnalysis(const EMClusterMember * clsMem);
  //--
  /// EMHitCell Analyzer
  virtual void doAnalysis(const EMHitCell * hitC) ;
  //--
  /// EMCellFraction Analyzer  
  inline 
  virtual void doAnalysis(const EMCellFraction * hitT);
  //--
  /// EMTrackOnEcal  Analyzer
  virtual void doAnalysis(const EMTrackOnEcal * track);
  //--
  /// EMCluster Analyzer
  virtual void doAnalysis(const EMCluster * cluster);
  //--
  /// EMClusterTree Analyzer
  virtual void doAnalysis(const EMClusterTree * tree);

  //--
  /**  
    Evaluate the information for the cluster 
    and fill a NEW entry in the bank. */
  void fillNewEntry(const EMCluster * clus, ArtePointer<RCCL> rccl);
  //--
  /**  Evaluate the information for the cluster 
    and fill an OLD entry in the bank. */
  void updateOldEntry(const EMCluster * clus, ArtePointer<RCCL> rccl);
  //--
  /**  Evaluate the information for the cluster 
    and fill the TRACK relation bank only. */
  void updateTrackRelEntry(const EMCluster * clus, ArtePointer<RCCL> rccl);
  //--
  ///  Test the links
  void testLinks(const EMCluster * clus, ArtePointer<RCCL> rccl);
  //@}

private:
  //--
  ///  Data for making relations in RCCL bank.
  bool  mkHitRel;
  bool  mkTraRel;
  unsigned int   nrccl;
};
/// EMCellFraction Analyzer  
void  EALinkAnalyzer::doAnalysis(const EMCellFraction * hitT){
  doAnalysis((const EMHitCell*) hitT);
};



#endif // EALINKANALYZER_HH_

