//                              -*- Mode: C++ -*-
//: ESCOGPos.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 14/4/97
// Last Update ........
//

#ifndef ESCOGPOS_HH_
#define ESCOGPOS_HH_

#include "care/ECALtypes.hh"
#include "care/ESPosStrategy.hh"

/**
class ESCOGPos.
  Concrete class for cluster analysis: it evaluates total energy and
  the center of gravity of a cluster
*/

class ESCOGPos : public ESPosStrategy {
public:
  /// Constructor
  ESCOGPos();
  /// Destructor: currently void but needs to be virtual
  virtual ~ESCOGPos();
 
  /**@name Beheavioral methods */
  //@{ 
  ///   Evaluate the position for a cluster.
  virtual void evaluatePos(const EMCluster* clus);

  /// EMClusterMember Analyzer
  virtual void doAnalysis(const EMClusterMember * clsMem);
  //--
  /// EMHitCell Analyzer
  virtual void doAnalysis(const EMHitCell * hitT);
  //--
  /// EMCellFraction Analyzer  
  inline 
  virtual void doAnalysis(const EMCellFraction * hitT);
  //--
  /// Do a Cluster analysis
  virtual void doAnalysis(const EMCluster* clus);
  //--
  /// Do a ClusterTree analysis
  virtual void doAnalysis(const EMClusterTree* clus);
  //--
  /// EMTrackOnEcal analysis
  virtual void doAnalysis(const EMTrackOnEcal * Track);

  /// Z correction constants
  void setValues(SubCal cal, float constant, float slope);
  //@}
 
  /// access the constant value for the Z correction
  inline static float getConst(SubCal cal);
  /// access the constant value for the Z correction
  inline static float getSlope(SubCal cal);

  struct COGdata{
    int  ndata; // number of data
    Energy_t energySum; // partial energy sum
    Energy_t energyErr; // partial sum of energy errors
    Coord_t  xSum;   // sum of all x coord
    Coord_t  ySum;   // sum of all y coord
    Coord_t  xxSum;  // sum of all x*x coord
    Coord_t  xySum;  // sum of all x*y coord
    Coord_t  yySum;  // sum of all y*y coord
    COGdata() : ndata(0), energySum(0.0), energyErr(0.0),
      xSum(0.0), ySum(0.0),
      xxSum(0.0), xySum(0.0), yySum(0.0) {};
  };
  
protected:
  COGdata currClust;
  // the analyzer for the peak
  static ESPosStrategy* peakSt;

  static float cnst[6];
  static float logE[6];
};

/// EMCellFraction Analyzer  
void  ESCOGPos::doAnalysis(const EMCellFraction * hitT){
  doAnalysis((const EMHitCell*) hitT);
};

/// access the constant value for the Z correction
float ESCOGPos::getConst(SubCal cal){
  float cns = 0.0;
  switch(cal){
  case Inner: cns = cnst[5];
    break;
  case Middle: cns = cnst[2];
    break;
  case Outer: cns = cnst[1];
    break;
  default:
    cns = 0.0;
  }
  return cns;
}

/// access the constant value for the Z correction
float ESCOGPos::getSlope(SubCal cal){
  float slp = 0.0;
  switch(cal){
  case Inner: slp = logE[5];
    break;
  case Middle: slp = logE[2];
    break;
  case Outer: slp = logE[1];
    break;
  default:
    slp = 0.0;
  }
  return slp;
}

#endif // ESCOGPOS_HH_
