// 
//  Authors:  Silvia Masciocchi                Created:  05-Feb-99 ver. 1.01
//            Ivan Kisel                       Modified: 07-Jun-99 ver. 2.01
//            Max-Planck-Institut fuer Physik            01-Jul-99 ver. 2.02 
//            Munchen                                    03-Jul-99 ver. 2.03 
//  Email:    sma@mppmu.mpg.de                           14-Jul-99 ver. 2.04
//            ikisel@mppmu.mpg.de                        30-Jul-99 ver. 2.05
//            HERA-B Collaboration                       01-Sep-99 ver. 2.06
//                                                       19-Sep-99 ver. 3.01
//                                                       07-Oct-99 ver. 3.02
//                                                       07-Oct-99 ver. 3.03
//                                                       12-Oct-99 ver. 3.04
//                                                       25-Oct-99 ver. 3.05
//                                                       05-Nov-99 ver. 3.06
//                                                       18-Nov-99 ver. 3.07
//                                                       01-Dec-99 ver. 3.08
//                                                       11-Dec-99 ver. 3.09
//                                                       17-Dec-99 ver. 3.10
//                                                       12-Jan-00 ver. 3.11
//                                                       23-Jan-00 ver. 3.12
//                                                       11-Feb-00 ver. 3.13
//                                                       13-Mar-00 ver. 3.14
//                                                       23-Apr-00 ver. 3.15
//                                                       12-Jun-00 ver. 3.16
//
// cats.c
// VDS reconstruction program - main part, uses fill_banks.C .
// The algorithm is based on the cellular automaton NIM A 387 (1997) 433-442. 
//
// Version 3.xx does reconstruction of tracks from space points.
//
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <arte/Arte.h>
#include <arte/ArteUtil.h>

#include <auto/MTRA.h>
#include <auto/MIMP.h>
#include <auto/HITB.h>
#include <auto/GEDE.h>
#include <auto/GESL.h>
#include <auto/GTAR.h>
#include <auto/RSEG.h>
#include <auto/RTRA.h>
#include <auto/TOPP.h>
#include <auto/MVER.h>
#include <auto/RVER.h>

#include "../cats/include/hbook.h"
#include "../cats/include/higz.h"
 
#include "../cats/include/cats.h"

#define NCHAMBERS  10
#define NPLANES   200
#define NGEWA    1500
#define NPROJ       4
#define NTARGETS    8
#define MAXNUMBEROFHITS     20000
#define MAXNUMBEROFPOINTS   20000
#define MAXNUMBEROFSEGMENTS 40000
#define MAXNUMBEROFTRACKS   30000
#define MAXNUMBEROFVERTICES 100
#define MAXHITSPERPLANE     100

//extern void sma_histos(int);
//extern void sma_effi(int);
extern int acglobaltolocal(float *, float *, int);
extern int EventFilter(void);
extern void fill_GEDE_array(void);
extern void fill_banks(int, int *, 
		       float *, float *, float *, float *,
		       float *, float *, float *, int *);
extern int  fill_mimp_points(int, float *, float *, float *);
extern void dinv_(int* n, double* da, int* m, int* work, int* ifail);
extern float prob_(float *, int *);
//extern void time_start();
//extern void time_stop();

extern void prima_love(void);

static int Black =  1, 
    DarkGray     =  2, 
    Blue         =  3, 
    Green        =  4, 
    Cyan         =  5, 
    Red          =  6, 
    Magenta      =  7, 
    Brown        =  8, 
    LightBlue    =  9, 
    LightGreen   = 10, 
    LightCyan    = 11, 
    LightRed     = 12, 
    LightMagenta = 13, 
    Yellow       = 14, 
    LightGray    = 15, 
    White        =  0;

void fill_cats();
static void Draw_Event();
static void graph_ini();
static void graph_close();
static void Space_Points();
static void Space_Automaton();
static void Space_Track_Finder();
static void Space_Find_Track_Chi_2(int n, int k);


                 /* distances in cm */


float z_nsl[NCHAMBERS+1];              /* z-pos of the 1st plane      */


static float xpl[1000], ypl[1000];
static int   n_mimps;
static float mimp_x[100], mimp_y[100], mimp_z[100];

static int   nchambers, ichamber[NCHAMBERS+1];/* real number of superlayers  */
                                              /*and numbers in the GESL table*/
static int   nplanes;
//////////////////////////////////////////////////////////////////////////
int   Nsl_Sec[NCHAMBERS+1][5][5];      /* plane from nsl, sect, proj  */
//////////////////////////////////////////////////////////////////////////

static int   ntargets,  itarget[NTARGETS+1];  /* numbers of targets wires    */

static float Z_TARGET;                        /* mean target z-position      */
static float GTAR_x[NTARGETS+1], GTAR_y[NTARGETS+1], GTAR_z[NTARGETS+1];

static int   fh[NPLANES], lh[NPLANES], fhc[NPLANES], lhc[NPLANES];
static int   npl; 

/////////////////////////////////////////
int   npr[NPLANES];
float zpl[NPLANES], coeff[NPLANES];

int   GEDE_cmp[NPLANES],  GEDE_nsl[NPLANES],
  GEDE_sec[NPLANES],  GEDE_lay[NPLANES];
float GEDE_sxmn[NPLANES], GEDE_sxmx[NPLANES];
float GEDE_symn[NPLANES], GEDE_symx[NPLANES];
float GEDE_cos[NPLANES],  GEDE_sin[NPLANES]; 
float GEDE_xfw[NPLANES],  GEDE_nwi_x_pit[NPLANES];
int   GEDE_gewa[NPLANES], ngewa;

float GEWA_xmin[NGEWA], GEWA_xmax[NGEWA];
float GEWA_ymin[NGEWA], GEWA_ymax[NGEWA];
float GEWA_zmin[NGEWA], GEWA_zmax[NGEWA];
float GEWA_zmean[NGEWA];
int   GEWA_gewa[NGEWA];
float GEWA_u[NGEWA], GEWA_z[NGEWA], GEWA_ms2[NGEWA];


int   cell_f_tmp[201][51], cell_l_tmp[201][51];
int   cell_f[201][51], cell_l[201][51];
/////////////////////////////////////////
static int   ihit[MAXNUMBEROFHITS+1];
static int   ifl[MAXNUMBEROFHITS+1], jfl[MAXNUMBEROFHITS+1];
static int   ihc[MAXNUMBEROFHITS+1], jhc[MAXNUMBEROFHITS+1];

static int   h_t_pointer[MAXNUMBEROFHITS+1]; // hit to track pointer

/////////////////////////////////////////////
float HITB_x[MAXNUMBEROFHITS+1];    /* original value                */
////////////////////////////////////////////

static float X_Hit[MAXNUMBEROFHITS+1];     /* coordinate ???                */
static float Z_Hit[MAXNUMBEROFHITS+1];     /* coordinate ???                */
static int   L_Hit[MAXNUMBEROFHITS+1];     /* layer                         */
static int   M_Hit[MAXNUMBEROFHITS+1];     /* which part of a layer 1,2,3,4 */
static int   IPL_Hit[MAXNUMBEROFHITS+1];   /* plane number of a hit         */
static int   Used_Hit[MAXNUMBEROFHITS+1];  /* 1 - used, 0 - not             */

static int   N_Points;           /* total number of points */
static int   NC_Points;          /* total number of unused points */

static float X_Point[MAXNUMBEROFPOINTS+1];     /* x-coordinate               */
static float Y_Point[MAXNUMBEROFPOINTS+1];     /* y-coordinate               */
static float Z_Point[MAXNUMBEROFPOINTS+1];     /* z-coordinate               */
static int   NSL_Point[MAXNUMBEROFPOINTS+1];   /* layer                      */
static int   SEC_Point[MAXNUMBEROFPOINTS+1];   /* detector number - see M_Hit*/
static int   Hit_Point[MAXNUMBEROFPOINTS+1][5];/* pointers to user hits      */
static int   N_Point[MAXNUMBEROFPOINTS+1];     /* number of used hits        */
static float E_Point[MAXNUMBEROFPOINTS+1];     /* error of reconstruction    */
static int   Good_Point[MAXNUMBEROFPOINTS+1];  /* to remove bad combinations */
static int   IC_Point[MAXNUMBEROFPOINTS+1];    /* index for sorting          */

static int   TrackPoint[MAXNUMBEROFPOINTS+1];  /* free or on a track         */
static int   FirstPoint[MAXNUMBEROFPOINTS+1];  /* first point on a track     */
static int   OnTheTrack[MAXNUMBEROFPOINTS+1];  /* track number               */


static int   N_Segments, Total_N_Segments;

static float X_Start[MAXNUMBEROFSEGMENTS+1],   X_End[MAXNUMBEROFSEGMENTS+1];
static float Y_Start[MAXNUMBEROFSEGMENTS+1],   Y_End[MAXNUMBEROFSEGMENTS+1];
static float Z_Start[MAXNUMBEROFSEGMENTS+1],   Z_End[MAXNUMBEROFSEGMENTS+1];

static int   FirstSegment[MAXNUMBEROFHITS+1], LastSegment[MAXNUMBEROFHITS+1];
static int   TrackHit[MAXNUMBEROFHITS+1]; // free or on a track at the depth level


/*
The first and the last segments starting at a hit and ending at
another hit on the next plane. Index is the number of the hit.
*/

/* Arrays describing segments. Index is the number of the segment.  */


static int   StartOfSegment[MAXNUMBEROFSEGMENTS+1], 
      EndOfSegment[MAXNUMBEROFSEGMENTS+1];


/* Starting and ending hits of a segment. */

static int   First_Segment[MAXNUMBEROFSEGMENTS+1], 
      Last_Segment[MAXNUMBEROFSEGMENTS+1];

/*
The first and the last segments starting at the ending hit of a
segment and making a small [ < arccos(COSPHIMAX) ] angle with the segment.
I: I would prefer FirstNeighbor & LastNeighbor instead.
*/

static int   HasNeighbour[MAXNUMBEROFSEGMENTS+1];
static int   LevelOfSegment[MAXNUMBEROFSEGMENTS+1];
static int   CorrectIndex[MAXNUMBEROFSEGMENTS+1];

static int   StartOnPlane[NCHAMBERS+1], EndOnPlane[NCHAMBERS+1];
/*
The first and the last hit on a plane. Index is the number of the plane.
(!) StartOnPlane>=EndOnPlane.
*/

static int   SetNo;

static int   PlaneSet[10][10]={{8,1,2,3,4,5,6,7,8},{4,1,2,4,5},{4,1,3,4,5}};
//static int   CoPlaneSet[10][10]={{0},{1,3},{1,2}};

static int   NumberOfTracks, Level;
static int   LengthOfTrack[MAXNUMBEROFTRACKS+1];
static int   GoodTrack[MAXNUMBEROFTRACKS+1];
static float Chi_2[MAXNUMBEROFTRACKS+1];
static float Qual[MAXNUMBEROFTRACKS+1];
static int   idx_tr[MAXNUMBEROFTRACKS+1];
//static int   n_sector[MAXNUMBEROFTRACKS+1];

static int   CurrentTrackStart[NCHAMBERS+1];
static int   TrackStart[MAXNUMBEROFTRACKS+1][NCHAMBERS+1];

static float x_start_track, y_start_track, z_start_track;
//static float x_end_track,   y_end_track,   z_end_track;

static float X_Start_Track[MAXNUMBEROFTRACKS+1];
static float Y_Start_Track[MAXNUMBEROFTRACKS+1];
static float Z_Start_Track[MAXNUMBEROFTRACKS+1];
static float X_End_Track[MAXNUMBEROFTRACKS+1];
static float Y_End_Track[MAXNUMBEROFTRACKS+1];
static float Z_End_Track[MAXNUMBEROFTRACKS+1];


static int   Call;

static float SHIFT      =  10.0;
static float shift;
////////////////////////////////////////////////////
int   base_plane[5] = {0, 1, 11, 2, 7};
////////////////////////////////////////////////////
static int   shift_coeff[5] = {0, 3, 1, -1, -3};

static int   Ntk, nh_trk[MAXNUMBEROFTRACKS+1];
static float Xp_trk[MAXNUMBEROFTRACKS+1], X0_trk[MAXNUMBEROFTRACKS+1];
static float Yp_trk[MAXNUMBEROFTRACKS+1], Y0_trk[MAXNUMBEROFTRACKS+1];
static float Z1_trk[MAXNUMBEROFTRACKS+1], Z2_trk[MAXNUMBEROFTRACKS+1];
static float Chi2_trk[MAXNUMBEROFTRACKS+1];
static int   hits[MAXNUMBEROFHITS+3*MAXNUMBEROFTRACKS+1];

static int   nhits, nfound, hh[NPLANES];
static int   gfound, gg[NPLANES];


static int   MINLEVEL    = 3;
static int   n_at_least  = 5;
static float TARGETSCALE = 10.0;

static float DOUBLETARGETCUT;// = 2.0*TARGETREGION;
static float MINDIST;        /* distance for hit searching around a track  */
static float TARGETREGION;   /* half of a target window                    */
static int   ind_pl, ind_ba; /* plane to calculate det. efficiency         */
static int   DRAW;           /* switch on the display                      */

static int EventNumber = 0;
static int back[NPLANES];

static float x_station[3], y_station[3], z_station[3];
static int   n_station[3];
static int   no_targets = 0;

/* structure for IOL */
Cats_Iol_t cats_iol = {
  "cats"                  , IOL_NVERBS(Cats_Iol_t,0),
  {"search window"        ,0.0100            ,  " " , IOL_TYPE_FLOAT},
  {"target window"        ,0.3               ,  " " , IOL_TYPE_FLOAT},
  {"det. plane off"       ,0                 ,  " " , IOL_TYPE_INT},
  {"graphics on"          ,0                 ,  " " , IOL_TYPE_FLAG},

};

static int   cats_first = 0;
extern int   ROVER_DRAW;
static int   CATS_DRAW = 0;

void cats_param_init()
{
  /* initialize iol parameters of silarte package */
  iol_declare_structure((IOL_GENERIC *) &cats_iol);

}


void  CatsEvent(){
  int   ih, nh, ifh, ilh, ipl, ifpl, ilpl, ich, nch, ic, ihfirst;
  float xt, yt, zt;
  float local[5], global[5];
  int   ista;

  MINDIST      = cats_iol.search_window.value;
  TARGETREGION = cats_iol.target_window.value;
  ind_pl       = cats_iol.gede_plane_off.value;
  DRAW         = cats_iol.draw_display.value;

  if (DRAW){        // exchange to draw track 
    ROVER_DRAW = 0; // 0;
    CATS_DRAW  = 1; // 1;
  }

  if (ind_pl<0)
    ind_pl = 0; // to mask previous way to run rover

  if (cats_first==0){
    cats_first = 10;
    fill_cats();
    cats_first = 1;
  }

  EventNumber++;
  //if (EventNumber<1398) return;

  
  /* check quality of real data */
  if(!EventFilter()) return;

  /* recall the target wires for each event */
  ntargets = 0;
  Z_TARGET = 0.0;
  z_station[1] = 0.0; x_station[1] = 0.0; y_station[1] = 0.0; n_station[1] = 0;
  z_station[2] = 0.0; x_station[2] = 0.0; y_station[2] = 0.0; n_station[2] = 0;
  global[3] = 0.; global[4] = 0.; /* for transformation of target wires as space points */

  for(ih=1; ih <=NGTAR(); ih++){
    if (GTAR(ih).wnr!=0){
      xt = GTAR(ih).x;
      yt = GTAR(ih).y;
      zt = GTAR(ih).z;

      if (xt*xt+yt*yt<1.0){
	ntargets++;
        itarget[ntargets] = ih;
	Z_TARGET = Z_TARGET + zt;

	/* seems that GTAR is in global coordinates starting from 03-06-r3... */
	     
	global[0] = xt; global[1] = yt; global[2] = zt;
	acglobaltolocal(local, global, 1);
	     
	GTAR_x[ntargets] = local[0];
	GTAR_y[ntargets] = local[1];
	GTAR_z[ntargets] = local[2];

	if (zt<-3.0) 
	  ista = 1;
	else 
	  ista = 2;

	n_station[ista]++;
	x_station[ista]+= GTAR_x[ntargets];
	y_station[ista]+= GTAR_y[ntargets];
	z_station[ista]+= GTAR_z[ntargets];

      }
    }
    else
      goto no_more_targets;
  }
no_more_targets:;  

  if (ntargets!=0){
    Z_TARGET = Z_TARGET/ntargets;
    if (n_station[1]!=0){
      x_station[1] = x_station[1]/n_station[1];
      y_station[1] = y_station[1]/n_station[1];
      z_station[1] = z_station[1]/n_station[1];
    }
    else
      z_station[1] = -5.0;
    if (n_station[2]!=0){
      x_station[2] = x_station[2]/n_station[2];
      y_station[2] = y_station[2]/n_station[2];
      z_station[2] = z_station[2]/n_station[2];
    }
    else
      z_station[2] = -1.0;
  }
  else{
    no_targets++;
    ntargets = 1;
    Z_TARGET = -4.7;
    GTAR_x[ntargets] = 0.0;
    GTAR_y[ntargets] = 0.0;
    GTAR_z[ntargets] = -4.7;
    z_station[1] = -5.0;
    z_station[2] = -1.0;
    if (no_targets<=10){
      printf("CATS : ******                  NO TARGETS                 ******\n");
      printf("CATS : ****** Dummy target at (0.0, 0.0, -4.7) introduced ******\n");
    }
    if (no_targets==10)
      printf("CATS : ******        the last message of this type        ******\n");

    if (no_targets>10)
      no_targets = 11;
}

  shift = SHIFT; 
  
  ind_ba = back[ind_pl];

  if (CATS_DRAW) Draw_Event();

  for(ipl=1; ipl <= nplanes; ipl++){ 
    fh[ipl] = 0;
    lh[ipl] =-1;
    fhc[ipl]= 0;
    lhc[ipl]=-1;
    for(ic=0; ic <= 50; ic++){ 
      cell_f[ipl][ic]= 100000;
      cell_l[ipl][ic]=-100000;
      cell_f_tmp[ipl][ic]= 100000;
      cell_l_tmp[ipl][ic]=-100000;
    }
  }


  npl = 0;
  nh = 0;
  for(ich=1; ich <= nchambers; ich++){
    nch = ichamber[ich];
    ifpl = GESL(nch).fpl;
    ilpl = ifpl + GESL(nch).npl;
    for(ipl=ifpl; ipl < ilpl; ipl++){ 
      if (ipl==ind_pl || 
	  ipl==ind_ba) continue; /* hide tested plane and its back*/

      ifh = GEDE(ipl).hit1;
      ilh = GEDE(ipl).hitl;
      if (ilh-ifh>=MAXHITSPERPLANE)
	printf("CATS : Noise: VDS plane %3d with %d hits excluded\n", ipl, (ilh-ifh));
      if ((ilh>=ifh)&&(ilh-ifh<MAXHITSPERPLANE)){ 
        /*
	  fh[ipl] = nh+1;
	  lh[ipl] = nh+1+(ilh-ifh);
	  fhc[ipl] = nh+1;
	  lhc[ipl] = nh+1+(ilh-ifh);
	  if (ipl > npl) npl = ipl;
	  */
        if (nh+1+(ilh-ifh)>MAXNUMBEROFHITS){
	  printf("CATS : *** MXNUMBEROFHITS reached %d ***\n", nh);
	  return;
	}
        ihfirst = 0;
        for(ih=ifh; ih <= ilh; ih++){ 
          if (HITB(ih).gede!=ipl){
            printf("CATS : HITB problem on plane %3d : hitb->gede %3d, hitb->z %5.2f, hitb->x %5.2f\n",  
                   ipl, HITB(ih).gede, HITB(ih).z, HITB(ih).x);
            continue;
          }

          nh++;
          
          if (ihfirst == 0){
            fh[ipl]  = nh;
            fhc[ipl] = nh;
            ihfirst = 1;
          }
          lh[ipl]  = nh;
          lhc[ipl] = nh;
          if (ipl > npl) npl = ipl;
          
	  ihit[nh] = ih;
	  ihc[nh]  = nh;
	  jhc[nh]  = nh;
	  ifl[nh]  = 0;
	  jfl[nh]  = 0;
	  h_t_pointer[nh] = 0;
 
	  HITB_x[nh] = HITB(ih).x;
	  X_Hit[nh]  = coeff[ipl]*HITB(ih).x;
	  Z_Hit[nh]  = HITB(ih).z;
	  L_Hit[nh]  = GEDE_nsl[ipl];
	  M_Hit[nh]  = GEDE_sec[ipl];
	  IPL_Hit[nh]= ipl;

	  Used_Hit[nh] = (0==1);
	  TrackHit[nh] = 0;
	  FirstSegment[nh]= 0;
	  LastSegment[nh] =-1;

	  ic = 50.0*((HITB_x[nh]-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	  
	  if ((ic<0)||(ic>50)) continue;
	  if (cell_f[ipl][ic] == 100000){
	    cell_f[ipl][ic]     = nh;
	    cell_f_tmp[ipl][ic] = nh;
	  }
	  cell_l[ipl][ic]     = nh+1;
	  cell_l_tmp[ipl][ic] = nh+1;

	}

	if (cell_f_tmp[ipl][1]<cell_f[ipl][0])
	  cell_f[ipl][0]=cell_f_tmp[ipl][1];
	if (cell_l_tmp[ipl][1]>cell_l[ipl][0])
	  cell_l[ipl][0]=cell_l_tmp[ipl][1];
	
	for(ic=1; ic <= 49; ic++){
	  if (cell_f_tmp[ipl][ic-1]<cell_f[ipl][ic])
	    cell_f[ipl][ic]=cell_f_tmp[ipl][ic-1];
	  if (cell_f_tmp[ipl][ic+1]<cell_f[ipl][ic])
	    cell_f[ipl][ic]=cell_f_tmp[ipl][ic+1];
	  if (cell_l_tmp[ipl][ic+1]>cell_l[ipl][ic])
	    cell_l[ipl][ic]=cell_l_tmp[ipl][ic+1];
	  if (cell_l_tmp[ipl][ic-1]>cell_l[ipl][ic])
	    cell_l[ipl][ic]=cell_l_tmp[ipl][ic-1];
	}
	
	if (cell_f_tmp[ipl][49]<cell_f[ipl][50])
	  cell_f[ipl][50]=cell_f_tmp[ipl][49];
	if (cell_l_tmp[ipl][49]>cell_l[ipl][50])
	  cell_l[ipl][50]=cell_l_tmp[ipl][49];

      }
    }
  }



  //  Space_Points();
  //time_start();
  Space_Track_Finder();
  //time_stop();

  //time_start();

  fill_banks(Ntk, nh_trk, 
	     Xp_trk, X0_trk, Yp_trk, Y0_trk,
	     Z1_trk, Z2_trk, Chi2_trk, hits);

  //time_stop();
  
  //time_start();
  //  prima_love();  /* run rover for primary vertices */
  //time_stop();
  
  return;
  
}


void Space_Track_Finder(){
  
  int   i, j, jj, ic, jc, ipr, ibpl, ipl, it, ipc, npc;
  int   k, kc, first, last, level_i, level_j, first_track, First_Track;
  int   ReCorrectIdx[MAXNUMBEROFSEGMENTS+1];
  int   ip, ih, jh, kh;
  int   itr, jtr, itrc, jtrc, jjc;
  int   MinLevel, is, ie, call, ich;
  float MinDist, Ax, Bx, Ay, By, x, xx, yy, qual;
  float x_n, y_n, z_n, x_m, y_m, z_m;
  int   ipr12_first, ipr12_last, ipr34_first, ipr34_last;
  float gMinDist;
  int   il, hits_on_nsl[11], hits_in_pr[5];
  int   start, stop, isec;
  int   local_hits[100];

  if (CATS_DRAW && NMIMP()){
    ismksc(0.5);
    ismk(3);
    ispmci(Green);
    
    for(ipr=1; ipr <= 4; ipr++){ 
      shift = shift_coeff[ipr]*SHIFT;
      ibpl = base_plane[ipr];
      for(ih=1; ih <= NMIMP(); ih++){ 
	if ( GEDE(MIMP(ih).gede).cmp != 1) continue; 
	xpl[0] = MIMP(ih).z;
	ypl[0] = MIMP(ih).x*GEDE_cos[ibpl] - MIMP(ih).y*GEDE_sin[ibpl]+shift;
	ipm(1,xpl,ypl);
      }
    }

    isplci(Green);
    
    for(itr=1; itr <= NMTRA(); itr++){ 
      n_mimps = fill_mimp_points(itr, mimp_x, mimp_y, mimp_z);
      for(ipr=1; ipr <= 4; ipr++){ 
	shift = shift_coeff[ipr]*SHIFT;
	ibpl = base_plane[ipr];
	for(ih=1; ih < n_mimps; ih++){ 
	  xpl[0] = mimp_z[ih];
	  ypl[0] = mimp_x[ih]*GEDE_cos[ibpl] - 
	    mimp_y[ih]*GEDE_sin[ibpl]+shift;
	  xpl[1] = mimp_z[ih+1];
	  ypl[1] = mimp_x[ih+1]*GEDE_cos[ibpl] - 
	    mimp_y[ih+1]*GEDE_sin[ibpl]+shift;
	  //ipl(2,xpl,ypl);
	}
      }
    }
    
    iuwk(1,1);
    
  }




  if (CATS_DRAW){
    ismksc(0.5);
    ismk(3);
    ispmci(Red);
    
    for(ipr=1; ipr <= 4; ipr++){ 
      shift = shift_coeff[ipr]*SHIFT;
      for(ipl=1; ipl <= npl; ipl++){ 
	if (npr[ipl] != ipr) continue;
	for (ih=fh[ipl]; ih<=lh[ipl]; ih++){
	  xpl[0] = Z_Hit[ih];
	  ypl[0] = X_Hit[ih]+shift;
	  ipm(1,xpl,ypl);
	}
      }
    }
    
    iuwk(1,1);
    
  }


  Ntk = 0;
  nhits=0;

  //  if (N_Points==MAXNUMBEROFPOINTS) return; // no tracking for very noisy events


  NumberOfTracks = 0;
  TrackStart[NumberOfTracks][1] = 0; /* dummy track */
  TrackStart[NumberOfTracks][2] = 0;

  MinLevel = MINLEVEL;

  for (call=0; call<=2; call++){
    Call = call;
    SetNo = 0;
    //time_start();
    Space_Points();
    //time_stop();
    if (N_Points==MAXNUMBEROFPOINTS) return; // no tracking for very noisy events
    //time_start();
    Space_Automaton();
    //time_stop();
    First_Track = NumberOfTracks+1;

    if ((Call == 0)||(Call == 2))
      DOUBLETARGETCUT = 2.0*TARGETREGION;
    else
      DOUBLETARGETCUT = 2.0*TARGETSCALE*TARGETREGION;

    if (call==0){
      //time_start();
    while (Level>=MinLevel){
      first_track = NumberOfTracks+1;

      
      for (k=1;k<=N_Segments;k++){
	kc = CorrectIndex[k];
	
	if (LevelOfSegment[kc] == Level){
	  
	  CurrentTrackStart[1] = StartOfSegment[kc];
	  x_start_track = X_Start[kc];
	  y_start_track = Y_Start[kc];
	  z_start_track = Z_Start[kc];
	  
	  Space_Find_Track_Chi_2(1, kc); 
	  /*
	  NumberOfTracks++;
	  TrackStart[NumberOfTracks][1] = StartOfSegment[kc];
	  TrackStart[NumberOfTracks][2] = EndOfSegment[kc];
	  LengthOfTrack[NumberOfTracks] = 2;
	  X_Start_Track[NumberOfTracks] = X_Start[kc];
	  Y_Start_Track[NumberOfTracks] = Y_Start[kc];
	  Z_Start_Track[NumberOfTracks] = Z_Start[kc];
	  X_End_Track[NumberOfTracks]   = X_End[kc];
	  Y_End_Track[NumberOfTracks]   = Y_End[kc];
	  Z_End_Track[NumberOfTracks]   = Z_End[kc];
	  
	  Qual[NumberOfTracks] = 0.0;
	  idx_tr[NumberOfTracks]=NumberOfTracks;
	  GoodTrack[NumberOfTracks]=1;
	  */
	}
      }
      
      if (NumberOfTracks == MAXNUMBEROFTRACKS){
	printf("CATS : (info) <<< MAXIMUM NUMBER OF TRACKS EXCEEDED >>>\n");
	goto max_tracks;
      }

     
      for (itr=first_track;itr<=NumberOfTracks;itr++){

	qual = 0.0;   /* not really chi_2 but quality of a track */
                       /* the higher number - the better track    */
	
	x_n = X_Start_Track[itr];
	y_n = Y_Start_Track[itr];
	z_n = Z_Start_Track[itr];
	x_m = X_End_Track[itr];
	y_m = Y_End_Track[itr];
	z_m = Z_End_Track[itr];
	
	Ax = (x_m-x_n)/(z_m-z_n); /* slope in proj. x */
	Bx = x_m-Ax*z_m;
	Ay = (y_m-y_n)/(z_m-z_n); /* slope in proj. y */
	By = y_m-Ay*z_m;
	
	
	nfound = 0;
	for(ipl=1; ipl <= npl; ipl++){ 
	  xx=Bx+Ax*zpl[ipl];
	  if ( xx < GEDE_sxmn[ipl] ||
	       xx > GEDE_sxmx[ipl] ) continue;
	  yy=By+Ay*zpl[ipl]; 
	  if ( yy < GEDE_symn[ipl] ||
	       yy > GEDE_symx[ipl] ) continue;
	  
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl]; 
	  
	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	  
	  if ((ic<0)||(ic>50)) continue;
	  
	  MinDist = MINDIST;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){
	    if (fabs(x-HITB_x[ih])<MinDist) {
	      MinDist    = fabs(x-HITB_x[ih]);
	    }
	  }
	  
	  if (MinDist<MINDIST){
	    nfound++;
	    qual += 1.0-MinDist/MINDIST;
	  }
	}
	
	qual += nfound;

	/* crossing the target wires? */

	MinDist = DOUBLETARGETCUT;

	for (it=1; it<=ntargets; it++){
	  xx=Bx+Ax*GTAR_z[it];
	  yy=By+Ay*GTAR_z[it];
	  if ( fabs(xx-GTAR_x[it])+fabs(yy-GTAR_y[it])<MinDist )
	    MinDist = fabs(xx-GTAR_x[it])+fabs(yy-GTAR_y[it]);
	}
	
	qual += 4.0*(1.0-MinDist/DOUBLETARGETCUT);

	Qual[NumberOfTracks] = qual;


	/* flag used hits */
	for (i=1;i<=LengthOfTrack[itr];i++) {
	  ip = TrackStart[itr][i];
	  for (ih=1;ih<=N_Point[TrackStart[itr][i]];ih++) {
	    jfl[Hit_Point[ip][ih]]=1;  /* flag hit as used */
	  }
	}
      }
      

      npc = 0;
      for (ipc=1;ipc<=NC_Points;ipc++) {
	ip = IC_Point[ipc];
	nfound = 0;
	for (ih=1;ih<=N_Point[ip];ih++) {
	  if (jfl[Hit_Point[ip][ih]]==0) nfound++;;
	}
	if (nfound<2)
	  TrackPoint[ip] = 1;
	else
	  IC_Point[++npc] = ip;
      }
      NC_Points = npc;

      
      if (Level>=MinLevel){
	jc=0; //calculate new set of segments
	for (ic=1; ic<=N_Segments; ic++){ // on previouse set
	  i = CorrectIndex[ic];           // real segment
	  ReCorrectIdx[ic] = 0;           // pointer to the new set
	  if ((TrackPoint[StartOfSegment[i]] == 0)&&
	      (TrackPoint[EndOfSegment[i]] == 0)){
	    jc++;                         // add to the new set
	    ReCorrectIdx[ic] = jc;        // point from old set to the new one
	    CorrectIndex[jc] = i;         // point from new set to real segment  
	    first=First_Segment[i];       // remember first/last
	    last=Last_Segment[i];
	    First_Segment[i]=0;
	    Last_Segment[i]=-1;
	    level_i=0;
	    for (k=first; k<=last; k++) { // check status of neighbours
	      kc=ReCorrectIdx[k];         // if there is corresp. in new set
	      if (kc>0) {                 // found
		if (First_Segment[i]==0) First_Segment[i]=kc;
		Last_Segment[i]=kc;      
		level_i=1;
		LevelOfSegment[CorrectIndex[kc]]=1;
	      }
	    }
	    LevelOfSegment[i]=level_i;
	  }
	}
	N_Segments = jc;
	
	Level = 0;
	for (ic=1; ic<=N_Segments; ic++){
	  i = CorrectIndex[ic];
	  if (LevelOfSegment[i]){
	    first = First_Segment[i];
	    last  = Last_Segment[i];
	    level_i = 0;
	    for (jc=first;jc<=last;jc++){
	      level_j = LevelOfSegment[CorrectIndex[jc]];
	      if (level_j > level_i)
		level_i = level_j;
	    }
	    LevelOfSegment[i] = ++level_i;
	    if (level_i > Level)
	      Level = level_i;
	  }
	}
	
      }
    }/* while (Level>= MinLevel); */
    //time_stop();
    }

    /* start working with single tracks */
    
    first_track = NumberOfTracks+1;
    
    //time_start();

    for (i=1; i<=Total_N_Segments; i++){
      if (NumberOfTracks == MAXNUMBEROFTRACKS){
	printf("CATS : (info) <<< MAXIMUM NUMBER OF TRACKS EXCEEDED >>>\n");
	goto max_tracks;
      }
      is = StartOfSegment[i];
      ie = EndOfSegment[i];


      if ((TrackPoint[is]!=0)&&(TrackPoint[ie]!=0)) continue;


      qual = 0.0;   /* not really chi_2 but quality of a track */
                     /* the higher number - the better track    */
      
      x_n = X_Start[i]; y_n = Y_Start[i]; z_n = Z_Start[i];
      x_m = X_End[i];   y_m = Y_End[i];   z_m = Z_End[i];
      
      Ax = (x_m-x_n)/(z_m-z_n); // slope in proj. x
      Bx = x_m-Ax*z_m;
      Ay = (y_m-y_n)/(z_m-z_n); // slope in proj. y
      By = y_m-Ay*z_m;

      nfound = 0;
      for(ipl=1; ipl <= npl; ipl++){ 
	xx=Bx+Ax*zpl[ipl];
	if ( xx < GEDE_sxmn[ipl] ||
	     xx > GEDE_sxmx[ipl] ) continue;
	yy=By+Ay*zpl[ipl]; 
	if ( yy < GEDE_symn[ipl] ||
	     yy > GEDE_symx[ipl] ) continue;
	
	x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl]; 
	
	ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	
	if ((ic<0)||(ic>50)) continue;
	
	MinDist = MINDIST;
	for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){
	  if (fabs(x-HITB_x[ih])<MinDist) {
	    MinDist    = fabs(x-HITB_x[ih]);
	  }
	}
	
	if (MinDist<MINDIST){
	  qual += 1.0-MinDist/MINDIST;
	  nfound++;
	}
      }
      
      if (nfound < n_at_least) continue; //weak track? 
      
      qual += nfound;
      
      /* crossing the target wires? */

      MinDist = DOUBLETARGETCUT;

      for (it=1; it<=ntargets; it++){
	xx=Bx+Ax*GTAR_z[it];
	yy=By+Ay*GTAR_z[it];
	if ( fabs(xx-GTAR_x[it])+fabs(yy-GTAR_y[it])<MinDist )
	  MinDist = fabs(xx-GTAR_x[it])+fabs(yy-GTAR_y[it]);
      }
      
      qual += 4.0*(1.0-MinDist/DOUBLETARGETCUT);

     
      NumberOfTracks++;
      TrackStart[NumberOfTracks][1] = is;
      TrackStart[NumberOfTracks][2] = ie;
      LengthOfTrack[NumberOfTracks] = 2;
      X_Start_Track[NumberOfTracks] = X_Start[i];
      Y_Start_Track[NumberOfTracks] = Y_Start[i];
      Z_Start_Track[NumberOfTracks] = Z_Start[i];
      X_End_Track[NumberOfTracks]   = X_End[i];
      Y_End_Track[NumberOfTracks]   = Y_End[i];
      Z_End_Track[NumberOfTracks]   = Z_End[i];
      
      Qual[NumberOfTracks] = qual;
      idx_tr[NumberOfTracks]=NumberOfTracks;
      GoodTrack[NumberOfTracks]=1;



    }


    
    
    
  max_tracks:;

    //time_stop();
    
    //time_start();

    for (itrc=First_Track;itrc<=NumberOfTracks;itrc++){
      itr = idx_tr[itrc];
      /* sort tracks */
      if (GoodTrack[itr]==1){
	jjc=itrc;
	jj =itr ;
	for (jtrc=(itrc+1);jtrc<=NumberOfTracks;jtrc++){
	  jtr = idx_tr[jtrc];
	  if (GoodTrack[jtr]==1){
	    if (Qual[jtr] > Qual[jj]){
	      jjc=jtrc;
	      jj =jtr ;
	    }
	  }
	}
	
	idx_tr[itrc]=jj;
	idx_tr[jjc ]=itr;
	
	/* collect track hits */
	
	itr = jj;
	
	x_n = X_Start_Track[itr];
	y_n = Y_Start_Track[itr];
	z_n = Z_Start_Track[itr];
	x_m = X_End_Track[itr];
	y_m = Y_End_Track[itr];
	z_m = Z_End_Track[itr];
	
	Ax = (x_m-x_n)/(z_m-z_n); /* slope in proj. x */
	Bx = x_m-Ax*z_m;
	Ay = (y_m-y_n)/(z_m-z_n); /* slope in proj. y */
	By = y_m-Ay*z_m;


	nfound = 0;
	gfound = 0;
	ipr12_first = 0;
	ipr12_last  = 0;
	ipr34_first = 0;
	ipr34_last  = 0;

	for(ipr=1; ipr <= 4; ipr++){ 
	  hits_in_pr[ipr] = 0;
	}

	for( il=1; il<=10;il++){
	  hits_on_nsl[il] = 0;
	}

	for(ipl=1; ipl <= npl; ipl++){ 
	  /*	  if (GEDE_sec[ipl]==n_s_co) continue; */
	  
	  xx=Bx+Ax*zpl[ipl];
	  if ( xx < GEDE_sxmn[ipl] ||
	       xx > GEDE_sxmx[ipl] ) continue;
	  yy=By+Ay*zpl[ipl]; 
	  if ( yy < GEDE_symn[ipl] ||
	       yy > GEDE_symx[ipl] ) continue;
	  
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl]; 
	  
	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	  
	  if ((ic<0)||(ic>50)) continue;

	  MinDist = 2.0*MINDIST;
	  gMinDist = 2.0*MINDIST;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){
	    if (fabs(x-HITB_x[ih])<gMinDist) {
	      gMinDist   = fabs(x-HITB_x[ih]);
	      gg[gfound] = ih;
	    }
	    if (ifl[ih]==0)
	      if (fabs(x-HITB_x[ih])<MinDist) {
		MinDist    = fabs(x-HITB_x[ih]);
		hh[nfound] = ih;
	      }
	  }
	  
	  if (gMinDist<2.0*MINDIST){
	    gfound++;
	    //	    hits_in_pr[npr[ipl]]++;
	    //	    hits_on_nsl[GEDE_nsl[ipl]]++;
	    if ((npr[ipl]==1)||(npr[ipl]==2)){
	      if (ipr12_first==0)
		ipr12_first = GEDE_nsl[ipl];
	      ipr12_last = GEDE_nsl[ipl];
	    }
	    else{
	      if (ipr34_first==0)
		ipr34_first = GEDE_nsl[ipl];
	      ipr34_last = GEDE_nsl[ipl];
	    }
	  }
	  
	  if (MinDist<2.0*MINDIST){
	    hits_in_pr[npr[ipl]]++;
	    hits_on_nsl[GEDE_nsl[ipl]]++;
	    nfound++;
	  }
	}

	if (nfound < n_at_least){
	  GoodTrack[itr] = 0;
	  continue; 
	}

	if ((ipr12_last-ipr12_first==0)||(ipr34_last-ipr34_first==0)){
	  GoodTrack[itr] = 0;
	  continue; 
	}

	/* store found hits */
	/*	
	for(i=0; i < nfound; i++) {
	  ih = hh[i];
	  hits[nhits++]=ihit[ih];
	  ifl[ih]=1;  // flag hit as used 
	}
      */

	for(i=0; i < nfound; i++) {
	  ih = hh[i];
	  local_hits[i]=ihit[ih];
	  ifl[ih]=1;  /* flag hit as used */
	}

	//sort according to z
	for(i=0; i < nfound; i++) {
	  ih = local_hits[i];
	  for(j=i+1; j < nfound; j++) {
	    jh = local_hits[j];
	    if (HITB(ih).z>HITB(jh).z){
	      kh = ih;
	      ih = jh;
	      jh = kh;
	      local_hits[i] = ih;
	      local_hits[j] = jh;
	    }
	  }
	}

	for(i=0; i < nfound; i++) {
	  hits[nhits++]=local_hits[i];
	}

	hits[nhits++]=-1; /* end of hit list */
	
	/* store found track parameters */
	X0_trk[Ntk]=Bx;
	Xp_trk[Ntk]=Ax;
	Y0_trk[Ntk]=By;
	Yp_trk[Ntk]=Ay;
	nh_trk[Ntk]=nfound;
	
	Z1_trk[Ntk]=Z_Start_Track[itr];
	Z2_trk[Ntk]=Z_End_Track[itr];

//	TrackRefit();


	if (CATS_DRAW){
	  isplci(Red);

	  Bx = X0_trk[Ntk];
	  Ax = Xp_trk[Ntk];
	  By = Y0_trk[Ntk];
	  Ay = Yp_trk[Ntk];
	  
	  
	  for(ipr=1; ipr <= 4; ipr++){
	    shift = shift_coeff[ipr]*SHIFT;
	    ibpl = base_plane[ipr];
	    
	    xx=Bx+Ax*Z_Hit[hh[0]];
	    yy=By+Ay*Z_Hit[hh[0]];

	    xpl[0] = Z_Hit[hh[0]];
	    ypl[0] = xx*GEDE_cos[ibpl] - yy*GEDE_sin[ibpl]+shift;

	    xx=Bx+Ax*Z_Hit[hh[nfound-1]];
	    yy=By+Ay*Z_Hit[hh[nfound-1]];

	    xpl[1] = Z_Hit[hh[nfound-1]];
	    ypl[1] = xx*GEDE_cos[ibpl] - yy*GEDE_sin[ibpl]+shift;

	    ipl(2,xpl,ypl);
	  }	
	}


	if(++Ntk >= MAXNUMBEROFTRACKS)  {
	  printf("CATS : (info) *** Space track *** == > too many tracks %d\n", Ntk); 
	  return;
	}

      }
   
    }
    //time_stop();




    if (CATS_DRAW)
      printf("Track candidates: long %3d; short %3d => %3d\n", 
	     first_track - First_Track, NumberOfTracks - first_track, Ntk);


    while ((NumberOfTracks>0)&&(GoodTrack[NumberOfTracks]==0))
      NumberOfTracks--;
    

    for (ich=1; ich <= nchambers; ich++){
      for (isec=1; isec <= 4; isec++){
	for (ipr=1; ipr <= 4; ipr++){
	  
	  ipl =  Nsl_Sec[ich][isec][ipr];
	  if (ipl==0) continue;
	  
	  start = fh[ipl];
	  stop  = lh[ipl];
	  if (stop<start) continue;
	  for (ih=start; ih<=stop; ih++){
	    if (ifl[ih]==1)
	      Used_Hit[ih] = 1;
	    else
	      Used_Hit[ih] = 0;
	  }
	}
      }
    }


    /*
    for (ich=0; ich<=nchambers; ich++) {
      StartOnPlane[ich] = 0;
      EndOnPlane[ich]   =-1;
    }
    
    npc = 0;
    for (ipc=1;ipc<=NC_Points;ipc++) {
      ip = IC_Point[ipc];
      TrackPoint[ip] = 0;
      nfound = 0;
      for (ih=1;ih<=N_Point[ip];ih++) {
	if (ifl[Hit_Point[ip][ih]]==0) nfound++;;
      }
      if (nfound<2)
	TrackPoint[ip] = 1;
      else{
	IC_Point[++npc] = ip;
	ich=NSL_Point[ip];
	if (StartOnPlane[ich] == 0)
	  StartOnPlane[ich] = npc;
	EndOnPlane[ich] = npc;
      }
    }
    NC_Points = npc;
    */
    
  }/*call */

       
  if (CATS_DRAW){
    printf(" *** Tracks found ===> %3d\n",Ntk);
    
    iuwk(1,1);
  }

}






void Space_Find_Track_Chi_2(int n, int k){
  
  int   i;
  int   l, lc, lvl;
  int   first, last;
  float chi_2;
  int   start, end, point;
  float k1, k2, x, y;

  lvl=Level-n; n++;

  if (lvl == 0){

    start = CurrentTrackStart[1];
    end   = EndOfSegment[k];
    chi_2 = 0.0;//N_Point[start];
    for (i=2;i<n;i++){
      point=CurrentTrackStart[i];
      k1=(Z_Point[end]-Z_Point[point])/(Z_Point[end]-Z_Point[start]);
      k2=(Z_Point[point]-Z_Point[start])/(Z_Point[end]-Z_Point[start]);
      x=k1*X_Point[start]+k2*X_Point[end];
      y=k1*Y_Point[start]+k2*Y_Point[end];
      
      chi_2 += (1.0-fabs(X_Point[point]-x)/MINDIST)+
	(1.0-fabs(Y_Point[point]-y)/MINDIST);// + N_Point[point];
    }
    //chi_2 += N_Point[CurrentTrackStart[n]];


    if ((TrackStart[NumberOfTracks][1] != CurrentTrackStart[1])||
	(TrackStart[NumberOfTracks][2] != CurrentTrackStart[2])){

      if (NumberOfTracks < MAXNUMBEROFTRACKS)
	NumberOfTracks++;
      idx_tr[NumberOfTracks]=NumberOfTracks;
      GoodTrack[NumberOfTracks] = 1;
    }
    else
      if ((LengthOfTrack[NumberOfTracks] > n)||
	  ((LengthOfTrack[NumberOfTracks] == n)&&
	   (Chi_2[NumberOfTracks] > chi_2)) )
	goto a;   
    
    Chi_2[NumberOfTracks] = chi_2;
    LengthOfTrack[NumberOfTracks] = n;
    X_Start_Track[NumberOfTracks] = x_start_track;
    Y_Start_Track[NumberOfTracks] = y_start_track;
    Z_Start_Track[NumberOfTracks] = z_start_track;
    X_End_Track[NumberOfTracks]   = X_End[k];
    Y_End_Track[NumberOfTracks]   = Y_End[k];
    Z_End_Track[NumberOfTracks]   = Z_End[k];
    for (i=1;i<n;i++)
      TrackStart[NumberOfTracks][i] = CurrentTrackStart[i];
    TrackStart[NumberOfTracks][n]=EndOfSegment[k];
  a:;
  }   
  else {
    first = First_Segment[k];
    last  = Last_Segment[k];
    for (l=first;l<=last;l++){
      lc = CorrectIndex[l];
      if (LevelOfSegment[lc] == lvl){
	CurrentTrackStart[n] = StartOfSegment[lc];
	
	Space_Find_Track_Chi_2(n, lc);
      }
    }
  }
  return;
}













void Space_Automaton(){

  int   i, ii, j, ij, n, m, nc, mc;
  int   index, ih1, ih2, ipl1, ipl2;
  int   start_on_plane, end_on_plane, iplane, eplane;
  float x_n, y_n, z_n, x_m, y_m, z_m;
  int   start, end, first, last;
  int   has_i, level_i, level_j;
  float x, y, x1, x2, z2, c1, s1, c2, s2, det, A, B, dx;
  float A1, A2, B1, B2, x_b[3], x_u[3], y_b[3], y_u[3], targetcut;
  int   it, from_a_target, skip;

  if ((Call==0)||(Call==2))
    targetcut = TARGETREGION+0.2;
  else
    targetcut = TARGETSCALE*TARGETREGION;

  skip = 0;

  if (Call==2)
    skip = 1;
  
  index = 0;

  /* create segments in upstream direction pointing from the target      */
  /* to have the last segment on each track connected with the target  */

  for (ii=1; ii<PlaneSet[SetNo][0]-skip; ii++){
    iplane=PlaneSet[SetNo][ii];
    eplane=PlaneSet[SetNo][ii+1+skip];
    start_on_plane = StartOnPlane[iplane];
    end_on_plane   = EndOnPlane[iplane];
    for (mc=StartOnPlane[eplane]; mc<=EndOnPlane[eplane]; mc++){
      m = IC_Point[mc];
      if (TrackPoint[m]==0){
	x_m = X_Point[m];
	y_m = Y_Point[m];
	z_m = Z_Point[m];
	FirstSegment[m] = index+1;

	for (it=1; it<=2; it++){
	  A1 = (x_m-(x_station[it]-targetcut))/(z_m-z_station[it]);
	  B1 = x_m-A1*z_m;
	  x_b[it]= A1*z_nsl[iplane]+B1-3.0*MINDIST;
	  A2 = (x_m-(x_station[it]+targetcut))/(z_m-z_station[it]);
	  B2 = x_m-A2*z_m;
	  x_u[it]= A2*z_nsl[iplane]+B2+3.0*MINDIST;
	  
	  A1 = (y_m-(y_station[it]-targetcut))/(z_m-z_station[it]);
	  B1 = y_m-A1*z_m;
	  y_b[it]= A1*z_nsl[iplane]+B1-3.0*MINDIST;
	  A2 = (y_m-(y_station[it]+targetcut))/(z_m-z_station[it]);
	  B2 = y_m-A2*z_m;
	  y_u[it]= A2*z_nsl[iplane]+B2+3.0*MINDIST;
	}

	for (nc = start_on_plane; nc<=end_on_plane; nc++){
	  n = IC_Point[nc];
	  if ((Call==0) && (SEC_Point[n] != SEC_Point[m])) continue;
	  //if ((SEC_Point[n] != SEC_Point[m])&&(eplane!=8)) continue;
	  if (TrackPoint[n]==0){ /* free */
	    x_n = X_Point[n];
	    y_n = Y_Point[n];
	    z_n = Z_Point[n];

	    
	    from_a_target = 0;
	    if (((x_n>x_b[1])&&(x_n<x_u[1])&&(y_n>y_b[1])&&(y_n<y_u[1]))||
		((x_n>x_b[2])&&(x_n<x_u[2])&&(y_n>y_b[2])&&(y_n<y_u[2])))
	      from_a_target = 1;

	    if (from_a_target){
	    
	      if (index==MAXNUMBEROFSEGMENTS){
		printf("CATS : (info) <<< MAXIMUM NUMBER OF SEGEMENTS EXCEEDED >>>\n");
		goto a;
	      }
	      index++;

	      StartOfSegment[index] = n; /* from the target direction */ 
	      EndOfSegment[index] = m;


	      /* determine position of a segment */

	      ih1 = Hit_Point[n][1];
	      ipl1 =  IPL_Hit[ih1];
	      x1 = x_n*GEDE_cos[ipl1] - y_n*GEDE_sin[ipl1];
	      x2 = x_m*GEDE_cos[ipl1] - y_m*GEDE_sin[ipl1];

	      A = (x2-x1)/(z_m-z_n); // slope in proj. ipl1
	      B = x2-A*z_m;

	      dx = A*Z_Hit[ih1]+B-HITB_x[ih1];;

	      ih2 = Hit_Point[n][2];
	      ipl2 =  IPL_Hit[ih2];
	      x2 = HITB_x[ih2];
	      z2 = Z_Hit[ih2];

	      x1 = A*z2+B-dx;

	      c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
	      c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
	      det=c1*s2-c2*s1;

	      /* space point coordinates of a segment's start */
	      X_Start[index] = (x1*s2-s1*x2)/det;
	      Y_Start[index] =-(x1*c2-c1*x2)/det;
	      Z_Start[index] = z2;


	      ih1 = Hit_Point[m][1];
	      ipl1 =  IPL_Hit[ih1];
	      x1 = x_n*GEDE_cos[ipl1] - y_n*GEDE_sin[ipl1];
	      x2 = x_m*GEDE_cos[ipl1] - y_m*GEDE_sin[ipl1];

	      A = (x2-x1)/(z_m-z_n); // slope in proj. ipl1
	      B = x2-A*z_m;

	      dx = A*Z_Hit[ih1]+B-HITB_x[ih1];;

	      ih2 = Hit_Point[m][2];
	      ipl2 =  IPL_Hit[ih2];
	      x2 = HITB_x[ih2];
	      z2 = Z_Hit[ih2];

	      x1 = A*z2+B-dx;

	      c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
	      c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
	      det=c1*s2-c2*s1;
	      
	      /* space point coordinates of a segment's end */
	      X_End[index] = (x1*s2-s1*x2)/det;
	      Y_End[index] =-(x1*c2-c1*x2)/det;
	      Z_End[index] = z2;


	      has_i = 0;
	      start = FirstSegment[n];
	      end   = LastSegment[n];
	      first =  0;
	      last  = -1;

	      for (j=start; j<=end; j++){
		A = (X_End[index]-X_Start[j])/(Z_End[index]-Z_Start[j]); 
		B = X_End[index]-A*Z_End[index];
		x = A*Z_Start[index]+B;
		A = (Y_End[index]-Y_Start[j])/(Z_End[index]-Z_Start[j]); 
		B = Y_End[index]-A*Z_End[index];
		y = A*Z_Start[index]+B;

		if ((fabs(x-X_Start[index])<MINDIST)&&
		    (fabs(y-Y_Start[index])<MINDIST)){
		  has_i = (1==1);
		  HasNeighbour[j] = (1==1);
		  if (first==0) first = j;
		  last = j;
		}
	      }
	      HasNeighbour[index]  = has_i;
	      First_Segment[index] = first;
	      Last_Segment[index]  = last;
	    }
	  }
	}
	LastSegment[m]  = index;
      }
    }
  }

a:

  Total_N_Segments = index;


  j=0;
  Level = 0;
  for (i=1; i<=index; i++){
    if (HasNeighbour[i]){
      j++;
      CorrectIndex[j] = i;
      HasNeighbour[i] = j;
      if (First_Segment[i]>0)
	First_Segment[i] = HasNeighbour[First_Segment[i]];
      if (Last_Segment[i]>0)
	Last_Segment[i]  = HasNeighbour[Last_Segment[i]];
      first = First_Segment[i];
      last  = Last_Segment[i];
      level_i = 0;
      for (ij=first;ij<=last;ij++){
	level_j = LevelOfSegment[CorrectIndex[ij]];
	if (level_j > level_i)
	  level_i = level_j;
      }
      LevelOfSegment[i] = ++level_i;
      if (level_i > Level)
	Level = level_i;
    }
  }
  N_Segments = j;

  
}





 

void Space_Points(){
  int   ipr1, ipr2, ipr3, ipr4, ipl1, ipl2, ipl3, ipl4, ipl, ich, isec; 
  int   ic, ih, ip, ip1;
  int   np_tmp4, np_tmp3, np_tmp2;
  int   start1, stop1, start2, stop2, ih1, ih2, ih3, ih4, ipr, ibpl;
  float c1, s1, c2, s2, det, x1, x2, z1, z2, z3, z4, xx, yy, zz;
  float x, MINDIST_SPACE, MinDist, e_tmp;
  float zst = -1.0;

  MINDIST_SPACE = 1.0*MINDIST;//1.5*MINDIST;

  ih3 = 0; ih4 = 0; /* dummy initialization */

  N_Points = 0;

  for (ich=1; ich <= nchambers; ich++){
    for (isec=1; isec <= 4; isec++){
      /* reconstruct space points from 4 hits */
      
      np_tmp4 = N_Points;

      ipr1=1; 
      ipr2=3;
      ipr3=2;
      ipr4=4;

      ipl1 =  Nsl_Sec[ich][isec][ipr1];
      ipl2 =  Nsl_Sec[ich][isec][ipr2];
      ipl3 =  Nsl_Sec[ich][isec][ipr3];
      ipl4 =  Nsl_Sec[ich][isec][ipr4];
	  
      if ((ipl1==0)||(ipl2==0)||(ipl3==0)||(ipl4==0)) goto tripple;

      c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
      c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
      det=c1*s2-c2*s1;

      start1 = fh[ipl1];
      stop1  = lh[ipl1];
      if (stop1<start1) goto tripple;
      start2 = fh[ipl2];
      stop2  = lh[ipl2];
      if (stop2<start2) goto tripple;

      for (ih1=start1; ih1<=stop1; ih1++){
	if (Used_Hit[ih1]) continue;
	x1 = HITB_x[ih1];
	z1 = Z_Hit[ih1];

	for (ih2 = start2; ih2<=stop2; ih2++){
	  if (Used_Hit[ih2]) continue;
	  x2 = HITB_x[ih2];
	  z2 = Z_Hit[ih2];
	  
	/* space point coordinates with correction */
	  xx= ((x1*(z2-zst)/(z1-zst))*s2-s1*x2)/det;
	  yy=-((x1*(z2-zst)/(z1-zst))*c2-c1*x2)/det;
 	  zz = z2;

	  if ( xx < GEDE_sxmn[ipl2]-0.1 ||
	       xx > GEDE_sxmx[ipl2]+0.1 ) continue;
	  if ( yy < GEDE_symn[ipl2]-0.1 ||
	       yy > GEDE_symx[ipl2]+0.1 ) continue;
	  

	  z3 = 0.0;
	  ipl = ipl3;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) continue;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z3 = Z_Hit[ih];
	      ih3 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) continue;

	  e_tmp = MinDist;

	  z4 = 0.0;
	  ipl = ipl4;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) continue;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z4 = Z_Hit[ih];
	      ih4 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) continue;
	  
	  e_tmp += MinDist;

	  if (N_Points==MAXNUMBEROFPOINTS){
	    printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
	    goto finish;
	  }

	  N_Points++;
	  X_Point[N_Points] = xx;
	  Y_Point[N_Points] = yy;
	  Z_Point[N_Points] = zz;
	  NSL_Point[N_Points] = ich;
	  SEC_Point[N_Points] = isec;
	  Hit_Point[N_Points][1] = ih1;
	  Hit_Point[N_Points][2] = ih2;
	  Hit_Point[N_Points][3] = ih3;
	  Hit_Point[N_Points][4] = ih4;
	  Used_Hit[ih1] = (1==1);
	  Used_Hit[ih2] = (1==1);
	  Used_Hit[ih3] = (1==1);
	  Used_Hit[ih4] = (1==1);
	  N_Point[N_Points] = 4;
	  Good_Point[N_Points] = (1==1);
	  E_Point[N_Points] = e_tmp;
	  IC_Point[N_Points] = N_Points;

	}
      }





      
      if (CATS_DRAW){ 
	ismksc(1.5);
	ispmci(Red);
	ismk(24);
	
	for(ip=np_tmp4+1; ip <= N_Points; ip++){
	  if (Good_Point[ip])
	    for(ipr=1; ipr <= 4; ipr++){ 
	      
	      ibpl = base_plane[ipr];
	      shift = shift_coeff[ipr]*SHIFT;
	      
	      x = X_Point[ip]*GEDE_cos[ibpl] - Y_Point[ip]*GEDE_sin[ibpl]; 
	      
	      xpl[0] = Z_Point[ip];
	      ypl[0] = x+shift;
	      ipm(1,xpl,ypl);

	      hfill(3000+10*ich+4, X_Point[ip], Y_Point[ip], 1.0);

	    }
	}    
      }
      
    tripple:;      

      np_tmp3 = N_Points;
      
      /* reconstruct space points from 3 hits */
      
      ipr1=1; 
      ipr2=3;
      ipr3=2;
      ipr4=4;
      
      ipl1 =  Nsl_Sec[ich][isec][ipr1];
      if (ipl1==0) goto second_tripple;
      ipl2 =  Nsl_Sec[ich][isec][ipr2];
      if (ipl2==0) goto second_tripple;
      
      c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
      c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
      det=c1*s2-c2*s1;
      
      start1 = fh[ipl1];
      stop1  = lh[ipl1];
      if (stop1<start1) goto second_tripple;
      start2 = fh[ipl2];
      stop2  = lh[ipl2];
      if (stop2<start2) goto second_tripple;
      
      for (ih1=start1; ih1<=stop1; ih1++){
	if (Used_Hit[ih1]) continue;
	x1 = HITB_x[ih1];
	z1 = Z_Hit[ih1];
	for (ih2 = start2; ih2<=stop2; ih2++){
	if (Used_Hit[ih2]) continue;
	  x2 = HITB_x[ih2];
	  z2 = Z_Hit[ih2];
	  
	/* space point coordinates with correction */

	  xx= (x1*s2*(z2-zst)/(z1-zst)-s1*x2)/det;
	  yy=-(x1*c2*(z2-zst)/(z1-zst)-c1*x2)/det;
	  zz = z2;
	  
	  if ( xx < GEDE_sxmn[ipl2]-0.1 ||
	       xx > GEDE_sxmx[ipl2]+0.1 ) continue;
	  if ( yy < GEDE_symn[ipl2]-0.1 ||
	       yy > GEDE_symx[ipl2]+0.1 ) continue;
	  

	  ipl3 =  Nsl_Sec[ich][isec][ipr3];
	  if (ipl3==0) goto label_4;

	  z3 = 0.0;
	  ipl = ipl3;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) goto label_4;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z3 = Z_Hit[ih];
	      ih3 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) goto label_4;

	  e_tmp = MinDist;
	  
	  if (N_Points==MAXNUMBEROFPOINTS){
	    printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
	    goto finish;
	  }
	  N_Points++;
	  X_Point[N_Points] = xx;
	  Y_Point[N_Points] = yy;
	  Z_Point[N_Points] = zz;
	  NSL_Point[N_Points] = ich;
	  SEC_Point[N_Points] = isec;
	  Hit_Point[N_Points][1] = ih1;
	  Hit_Point[N_Points][2] = ih2;
	  Hit_Point[N_Points][3] = ih3;
	  Hit_Point[N_Points][4] = 0;
	  N_Point[N_Points] = 3;
	  Good_Point[N_Points] = (1==1);
	  E_Point[N_Points] = e_tmp;
	  IC_Point[N_Points] = N_Points;
	  
	  continue;

	label_4:;

	  ipl4 =  Nsl_Sec[ich][isec][ipr4];
	  if (ipl4==0) continue;
	  	  
	  z4 = 0.0;
	  ipl = ipl4;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) continue;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z4 = Z_Hit[ih];
	      ih4 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) continue;
	  
	  e_tmp = MinDist;

	  if (N_Points==MAXNUMBEROFPOINTS){
	    printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
	    goto finish;
	  }
	  N_Points++;
	  X_Point[N_Points] = xx;
	  Y_Point[N_Points] = yy;
	  Z_Point[N_Points] = zz;
	  NSL_Point[N_Points] = ich;
	  SEC_Point[N_Points] = isec;
	  Hit_Point[N_Points][1] = ih1;
	  Hit_Point[N_Points][2] = ih2;
	  Hit_Point[N_Points][3] = ih4;
	  Hit_Point[N_Points][4] = 0;
	  N_Point[N_Points] = 3;
	  Good_Point[N_Points] = (1==1);
	  E_Point[N_Points] = e_tmp;
	  IC_Point[N_Points] = N_Points;
	}
      }

    second_tripple:;

      ipr1=2; 
      ipr2=4;
      ipr3=1;
      ipr4=3;
      
      ipl1 =  Nsl_Sec[ich][isec][ipr1];
      if (ipl1==0) goto end_tripple;
      ipl2 =  Nsl_Sec[ich][isec][ipr2];
      if (ipl2==0) goto end_tripple;

      c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
      c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
      det=c1*s2-c2*s1;

      start1 = fh[ipl1];
      stop1  = lh[ipl1];
      if (stop1<start1) goto end_tripple;
      start2 = fh[ipl2];
      stop2  = lh[ipl2];
      if (stop2<start2) goto end_tripple;

      for (ih1=start1; ih1<=stop1; ih1++){
	if (Used_Hit[ih1]) continue;
	x1 = HITB_x[ih1];
	z1 = Z_Hit[ih1];
	for (ih2 = start2; ih2<=stop2; ih2++){
	if (Used_Hit[ih2]) continue;
	  x2 = HITB_x[ih2];
	  z2 = Z_Hit[ih2];
	  
	/* space point coordinates with correction */

	  xx= (x1*s2*(z2-zst)/(z1-zst)-s1*x2)/det;
	  yy=-(x1*c2*(z2-zst)/(z1-zst)-c1*x2)/det;
 	  zz = z2;
	  
	  if ( xx < GEDE_sxmn[ipl2]-0.1 ||
	       xx > GEDE_sxmx[ipl2]+0.1 ) continue;
	  if ( yy < GEDE_symn[ipl2]-0.1 ||
	       yy > GEDE_symx[ipl2]+0.1 ) continue;
	  

	  ipl3 =  Nsl_Sec[ich][isec][ipr3];
	  if (ipl3==0) goto label_3;

	  z3 = 0.0;
	  ipl = ipl3;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) goto label_3;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z3 = Z_Hit[ih];
	      ih3 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) goto label_3;

	  e_tmp = MinDist;
	  
	  if (N_Points==MAXNUMBEROFPOINTS){
	    printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
	    goto finish;
	  }
	  N_Points++;
	  X_Point[N_Points] = xx;
	  Y_Point[N_Points] = yy;
	  Z_Point[N_Points] = zz;
	  NSL_Point[N_Points] = ich;
	  SEC_Point[N_Points] = isec;
	  Hit_Point[N_Points][1] = ih1;
	  Hit_Point[N_Points][2] = ih2;
	  Hit_Point[N_Points][3] = ih3;
	  Hit_Point[N_Points][4] = 0;
	  N_Point[N_Points] = 3;
	  Good_Point[N_Points] = (1==1);
	  E_Point[N_Points] = e_tmp;
	  IC_Point[N_Points] = N_Points;

	  continue;

	label_3:;

	  ipl4 =  Nsl_Sec[ich][isec][ipr4];
	  if (ipl4==0) continue;
	  	  
	  z4 = 0.0;
	  ipl = ipl4;
	  x = xx*GEDE_cos[ipl] - yy*GEDE_sin[ipl];

	  ic = 50.0*((x-GEDE_xfw[ipl])/GEDE_nwi_x_pit[ipl]);
	    
	  if ((ic<0)||(ic>50)) continue;
	    
	  MinDist = MINDIST_SPACE;
	  for (ih=cell_f[ipl][ic]; ih < cell_l[ipl][ic]; ih++){ 
	    if (Used_Hit[ih]) continue;
	    if (fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih])<MinDist){
	      MinDist = fabs(x*(Z_Hit[ih]-zst)/(zz-zst)-HITB_x[ih]);
	      z4 = Z_Hit[ih];
	      ih4 = ih;
	    }
	  }
	    
	  if (MinDist==MINDIST_SPACE) continue;
	  
	  e_tmp = MinDist;

	  if (N_Points==MAXNUMBEROFPOINTS){
	    printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
	    goto finish;
	  }
	  N_Points++;
	  X_Point[N_Points] = xx;
	  Y_Point[N_Points] = yy;
	  Z_Point[N_Points] = zz;
	  NSL_Point[N_Points] = ich;
	  SEC_Point[N_Points] = isec;
	  Hit_Point[N_Points][1] = ih1;
	  Hit_Point[N_Points][2] = ih2;
	  Hit_Point[N_Points][3] = ih4;
	  Hit_Point[N_Points][4] = 0;
	  N_Point[N_Points] = 3;
	  Good_Point[N_Points] = (1==1);
	  E_Point[N_Points] = e_tmp;
	  IC_Point[N_Points] = N_Points;
	}
      }

    end_tripple:;

      /* label 3-points */
      
      //if ((Call==0)||(ich>3))
	for(ip1=np_tmp3+1; ip1 <= N_Points; ip1++){
	  Used_Hit[Hit_Point[ip1][1]] = (1==1);
	  Used_Hit[Hit_Point[ip1][2]] = (1==1);
	  Used_Hit[Hit_Point[ip1][3]] = (1==1);
	}
      

      if (CATS_DRAW){ 
	ismksc(1.5);
	ispmci(Green);
	ismk(24);
	
	for(ip=np_tmp3+1; ip <= N_Points; ip++){
	  if (Good_Point[ip])
	    for(ipr=1; ipr <= 4; ipr++){ 
	      
	      ibpl = base_plane[ipr];
	      shift = shift_coeff[ipr]*SHIFT;
	      
	      x = X_Point[ip]*GEDE_cos[ibpl] - Y_Point[ip]*GEDE_sin[ibpl]; 
	      
	      xpl[0] = Z_Point[ip];
	      ypl[0] = x+shift;
	      ipm(1,xpl,ypl);
	      hfill(3000+10*ich+3, X_Point[ip], Y_Point[ip], 1.0);
	    }
	}    
      }

      np_tmp2 = N_Points;
      
      /* reconstruct space points from 2 hits */
      /* use only perpendicular projections   */

      for (ipr1=1; ipr1 <= 2; ipr1++){
	ipl1 =  Nsl_Sec[ich][isec][ipr1];
	if (ipl1==0) continue;
	c1=GEDE_cos[ipl1]; s1=-GEDE_sin[ipl1];
	start1 = fh[ipl1];
	stop1  = lh[ipl1];
	if (stop1<start1) continue;

	for (ipr2=3; ipr2 <= 4; ipr2++){
	  ipl2 =  Nsl_Sec[ich][isec][ipr2];
	  if (ipl2==0) continue;

	  c2=GEDE_cos[ipl2]; s2=-GEDE_sin[ipl2];
	  det=c1*s2-c2*s1;

	  start2 = fh[ipl2];
	  stop2  = lh[ipl2];
	  if (stop2<start2) continue;

	  for (ih1=start1; ih1<=stop1; ih1++){
	    if (Used_Hit[ih1]) continue;
	    x1 = HITB_x[ih1];
	    z1 = Z_Hit[ih1];
	    for (ih2 = start2; ih2<=stop2; ih2++){
	    if (Used_Hit[ih2]) continue;
	      x2 = HITB_x[ih2];
	      z2 = Z_Hit[ih2];
	      
	      /* space point coordinates with correction */

	      xx= (x1*s2*(z2-zst)/(z1-zst)-s1*x2)/det;
	      yy=-(x1*c2*(z2-zst)/(z1-zst)-c1*x2)/det;
 	      zz = z2;

	      if ( xx < GEDE_sxmn[ipl2]-0.1 ||
		   xx > GEDE_sxmx[ipl2]+0.1 ) continue;
	      if ( yy < GEDE_symn[ipl2]-0.1 ||
		   yy > GEDE_symx[ipl2]+0.1 ) continue;



	      if (N_Points==MAXNUMBEROFPOINTS){
		printf("CATS : (info) Max number of space points exceeded %d\n", N_Points);
		goto finish;
	      }
	      N_Points++;
	      X_Point[N_Points] = xx;
	      Y_Point[N_Points] = yy;
	      Z_Point[N_Points] = zz;
	      NSL_Point[N_Points] = ich;
	      SEC_Point[N_Points] = isec;
	      Hit_Point[N_Points][1] = ih1;
	      Hit_Point[N_Points][2] = ih2;
	      Hit_Point[N_Points][3] = 0;
	      Hit_Point[N_Points][4] = 0;
	      N_Point[N_Points] = 2;
	      Good_Point[N_Points] = (1==1);
	      E_Point[N_Points] = 0.0;
	      IC_Point[N_Points] = N_Points;
	    }
	  }
	}
      }


      if (CATS_DRAW){ 
	ismksc(1.5);
	ispmci(Blue);
	ismk(24);
	
	for(ip=np_tmp2+1; ip <= N_Points; ip++){
	  for(ipr=1; ipr <= 4; ipr++){ 
	    
	    ibpl = base_plane[ipr];
	    shift = shift_coeff[ipr]*SHIFT;
	    
	    x = X_Point[ip]*GEDE_cos[ibpl] - Y_Point[ip]*GEDE_sin[ibpl]; 
	    
	    xpl[0] = Z_Point[ip];
	    ypl[0] = x+shift;
	    ipm(1,xpl,ypl);
	    hfill(3000+10*ich+2, X_Point[ip], Y_Point[ip], 1.0);
	  }
	}    
      }
    }
  }
  
finish:;

  if (CATS_DRAW){ 
    iuwk(1,1);
  }

  for (ich=0; ich<=nchambers; ich++) {
    StartOnPlane[ich] = 0;
    EndOnPlane[ich]   =-1;
  }

  for (ip=1;ip<=N_Points;ip++){
    TrackPoint[ip] = 0;
    FirstPoint[ip] = 0;
    OnTheTrack[ip] = 0;    
    FirstSegment[ip]=0;
    LastSegment[ip]=-1;
    ich=NSL_Point[ip];
    if (StartOnPlane[ich] == 0)
      StartOnPlane[ich] = ip;
    EndOnPlane[ich] = ip;
  }

  NC_Points = N_Points;

}



void fill_cats(){
  
  int   ih, ipl, ipr, is;

  if (cats_first==10)
    printf("CATS  version 3.15 started\n");
  else
    printf("CATS  version 3.15 filled\n");

  ntargets = 0;
  Z_TARGET = 0.0;
  for(ih=1; ih <=NGTAR(); ih++){
    if (GTAR(ih).wnr!=0){
      ntargets++;
      itarget[ntargets] = ih;
      Z_TARGET = Z_TARGET + GTAR(ih).z;
      if ((CATS_DRAW)&&(cats_first==10)){
	printf("Target %1d wire %1d (dx %5.2f - dy %5.2f), %5.2f, %5.2f, %5.2f\n ",
	       ntargets, GTAR(ih).wnr, GTAR(ih).dx, GTAR(ih).dy, 
	       GTAR(ih).x, GTAR(ih).y, GTAR(ih).z);
      }
    }
  }
  if (ntargets!=0) Z_TARGET = Z_TARGET/ntargets;
  
  nchambers = 0;
  for(ih=1; ih <= 20; ih++){
    if (GESL(ih).cmp==1){
      nchambers++;     
      ichamber[nchambers] = ih;
      if ((CATS_DRAW)&&(cats_first==10)){
      printf("%2d (%2d) cmp=%2d, nsl=%2d, lay=%2d, npl=%2d, fpl=%3d z= %5.2f\n", 
	     nchambers, ih, GESL(ih).cmp, GESL(ih).nsl, GESL(ih).lay, 
	     GESL(ih).npl, GESL(ih).fpl, GESL(ih).z1);
      printf("%2d (%2d) cmp=%2d, nsl=%2d, lay=%2d, nwl=%2d, fwl=%3d z= %5.2f\n", 
	     nchambers, ih, GESL(ih).cmp, GESL(ih).nsl, GESL(ih).lay, 
	     GESL(ih).nwl, GESL(ih).fwl, GESL(ih).z1);
      }
      z_nsl[nchambers] = GEDE(GESL(ih).fpl).zfw[0];
      //      printf("Z of SL %1d = %5.2f\n", nchambers, z_nsl[nchambers]);
      for(is=1; is <= 4; is++){
	for(ipr=1; ipr <= 4; ipr++){
	  Nsl_Sec[nchambers][is][ipr] = 0;
	}
      }
      nplanes= GESL(ih).fpl+GESL(ih).npl-1;
    }
    if (GESL(ih).cmp==0) goto no_more_chambers;
  }
no_more_chambers:;


  for(ipl=1; ipl <= nplanes; ipl++){ 

    if ( GEDE(ipl).cmp != 1 ) 
      printf("CATS : ******* WRONG IINITIALIZATION *******\n");

    /*  moved to fill_GEDE_array */

    GEDE_cmp[ipl] = GEDE(ipl).cmp;
    GEDE_nsl[ipl] = GESL(GEDE(ipl).ogesl).nsl;
    GEDE_sec[ipl] = GEDE(ipl).sec;
    GEDE_lay[ipl] = GEDE(ipl).lay;
    GEDE_nwi_x_pit[ipl] = GEDE(ipl).nwi[0]*GEDE(ipl).pit;

    /* Determine the "back" of each plane */

    back[ipl] = 0;
    for (ih=1; ih <= nplanes; ih++){
      if ( GEDE(ih).sec == GEDE(ipl).sec &&
	   GESL(GEDE(ih).ogesl).nsl == GESL(GEDE(ipl).ogesl).nsl ) {
	if ((GEDE(ih).lay == 1 && GEDE(ipl).lay ==2) ||
	    (GEDE(ih).lay == 2 && GEDE(ipl).lay ==1) ||
	    (GEDE(ih).lay == 3 && GEDE(ipl).lay ==4) ||
	    (GEDE(ih).lay == 4 && GEDE(ipl).lay ==3) ){
	  back[ipl] = ih;
	  break;
	}
      }
    }

    //printf("Plane  %3d  has as friend  %3d \n", ipl, back[ipl]);

  }
  back[0] = 0;

  fill_GEDE_array();

  if ((CATS_DRAW)&&(cats_first==10)){
    graph_ini();
  }


}






void stop_cats(){

  if (CATS_DRAW){
    graph_close();
  }
}


void Draw_Event(){
  
  int   iproj, i, ih, ipl;
  
  
  /* test data only in a projection corresponding to this plane slope */
  
  
  for(iproj=1; iproj <= 4; iproj++){ 
    
    shift = shift_coeff[iproj]*SHIFT;
       
    isfaci(White);
    isplci(Red);
    igbox(-15.0,220.0,-10.0+shift,10.0+shift);
       
       /* draw the targets - only their centers - how to project them?.. */

    ispmci(Red);
    ismk(3);
    ismksc(0.5);
    
    for (i=1; i<=ntargets; i++){
      xpl[0] = GTAR(itarget[i]).z;
      ypl[0] = GTAR(itarget[i]).x+shift;
      ipm(1,xpl,ypl);
    }
    
    /* draw layers */
    
    ispmci(Black);
    ismk(3);
    ismksc(0.5);
    

    isplci(Black);
    for (i=1; i<=nchambers; i++){
      ih = ichamber[i];
      for (ipl=GESL(ih).fpl; ipl<GESL(ih).fpl+GESL(ih).npl; ipl++){

        if (GEDE(ipl).ogesl!=0){

          if (npr[ipl]==npr[base_plane[iproj]]){
            xpl[0] = GEDE(ipl).zmin;
            ypl[0] = coeff[ipl]*GEDE(ipl).xfw[0]+shift;
            xpl[1] = GEDE(ipl).zmin;
            ypl[1] = coeff[ipl]*(GEDE(ipl).xfw[0]+
                                 GEDE(ipl).nwi[0]*GEDE(ipl).pit)+shift;
	    if (ipl==ind_pl) 
	      isplci(Magenta);
	    else
	      if (ipl==ind_ba) 
		isplci(Yellow);
	      else
		isplci(Black);
	  ipl(2,xpl,ypl);
          }

        }
      }
    }

  }
  iuwk(1,1);
  
  
}



/***************** Graphics ******************/



/******************************************/
/************ graph_ini() ******************/
/******************************************/



void graph_ini(){
//  int nxpaw=NXPAW;
//  hlimit_(&nxpaw); 

  iginit(0);
  igsse(0, 8);


  iscr(1, Black       ,0.0000, 0.0000, 0.0000);
  iscr(1, DarkGray    ,0.5000, 0.5400, 0.5300);
  iscr(1, Blue        ,0.0000, 0.0000, 1.0000);
  iscr(1, Green       ,0.1333, 0.5451, 0.1333);
  iscr(1, Cyan        ,0.0000, 0.8078, 0.8196);
  iscr(1, Red         ,1.0000, 0.0000, 0.0000);
  iscr(1, Magenta     ,1.0000, 0.0000, 1.0000);
  iscr(1, Brown       ,0.5000, 0.1647, 0.1647);
  iscr(1, LightBlue   ,0.6784, 0.8471, 0.9020);
  iscr(1, LightGreen  ,0.5961, 0.9843, 0.5961);
  iscr(1, LightCyan   ,0.4980, 1.0000, 0.8314);
  iscr(1, LightRed    ,0.9804, 0.5020, 0.4471);
  iscr(1, LightMagenta,0.8549, 0.4392, 0.8392);
  iscr(1, Yellow      ,1.0000, 1.0000, 0.0000);
  iscr(1, LightGray   ,0.8275, 0.8275, 0.8275);
  iscr(1, White       ,1.0000, 1.0000, 1.0000);



  iswn(2,-15.00,220.00,-40.00,40.00);
  isvp(2,0.0,1.0,0.0,1.0);
  iselnt(2);

  isfais(1);

  isfaci(LightGray);
  isplci(Red);

  igbox(-15.00,220.00,-40.00,40.00);

  iuwk(1,1);

}
/**************************************************/
/************** graph_close() *********************/
/**************************************************/

void graph_close(){
  igend();
}















