#ifndef STL_H
#define STL_H
#include <map>
#include <multimap.h>
#include <vector>
#include <set>
#include <algorithm>
#include "roverstldef.h"
#endif

#ifndef _MATH_H_
#define _MATH_H_
#include <math.h>
#endif

using namespace std;

//functions for matrix calculations
//the matrix has to be stored in a vector<double>
//in the form:
//( a11 a12 ... a1n)
//( a21 a22 ... a2n)
//    .........       ={a11,a21,...,am1,a12,a22,....,amn}
//    .........
//( am1 a12 ... amn)
// which means fortran format

//matrixinversion with CERN-routine
extern "C" {
  void dinv_(int *, double *, int *, double *, int *);
}

// functions for user
	 
	//matrix multiplication
	 int mtxmul(vector<double>,vector<double>,vector<double>&,int,int,int);
        //matrix addition 
	 int mtxadd(vector<double>&,vector<double>&,vector<double>&);
        //matrix multiplication with scalar
	 int mtxscal(double,vector<double>&);
        //matrix inversion
	 int mtxinv(vector<double>&,int);
        //matrixinversion with several checks	 
         int mtxinv(vector<double>&,int,bool);
        //singular value decomposition	 
         double mtxsvdcmp(vector<double>&,int,vector<double>&);
        //solve linear equation Ax=b
	 int lusolve(vector<double>,int,vector<double>&);
        //inversion by LUdecomposition(seems to be unstable when used in fit!!) 
	 int luinverse(vector<double>& a,int n);


// internal functions
void nrerror(char *);
void svdcmp_nrc(vector< vector<double> >&,int,int,vector<double>& ,vector< vector<double> >&);
double PYTHAG(double,double);
double MAX(double,double);
double SIGN(double,double);
void lubksb_nrc(vector<double>&,int,vector<int>&,vector<double>&);
int ludcmp_nrc(vector<double>&,int,vector<int>&,double&);
