/****************************************************************************
*                                                                           *
*                  ROVER                                                    *
*                                                                           *
*            Dmitri Emeliyanov                                              *
*            Ivan Kisel                       Modified: 13-Mar-00 ver. 1.00 *
*            Max-Planck-Institut fuer Physik            21-Mar-00 ver. 1.01 *
*            Muenchen                                   04-Apr-00 ver. 1.02 *
*  Email:    ikisel@mppmu.mpg.de                        11-Apr-00 ver. 1.03 *
*            HERA-B Collaboration                       19-Apr-00 ver. 1.04 *
*                                                       23-Apr-00 ver. 1.05 *
*                                                       14-May-00 ver. 1.06 *
*                                                       16-May-00 ver. 1.07 *
*                                                                           *
* ------------------------------------------------------------------------- *
*       vertex reconstruction package                                       *  
*                                                                           *
****************************************************************************/
 
#include "../cats/include/rover.hh" // all includes are there
#include "../cats/include/rovermatrix.h" // all includes are there

// all input is in global coordinates

int rseg_to_rver[1500]; // will be not very much tracks
int rtra_to_rver[1500]; //  -""-
int rver_per_wnr[10];
int ntr_per_rver[10];
int rver_array_size = 1500;
float second_x, second_y, second_z;

static int Black =  1, 
    DarkGray     =  2, 
    Blue         =  3, 
    Green        =  4, 
    Cyan         =  5, 
    Red          =  6, 
    Magenta      =  7, 
    Brown        =  8, 
    LightBlue    =  9, 
    LightGreen   = 10, 
    LightCyan    = 11, 
    LightRed     = 12, 
    LightMagenta = 13, 
    Yellow       = 14, 
    LightGray    = 15, 
    White        =  0;



static Lover* lover = static_cast<Lover *>(0);
int rover_newgeometry = 1;
int ROVER_DRAW;
int RoverHidden = 0;
int RoverNver;

static float zsta = 0.0;
static int   nver;
static int   firstcall   = 1;
static int   MCdata     = 0;

static int   BadCovMatrix = 0;

extern "C"{
  void prima_love(void);
  void second_love(ArtePointer<RTRA>, ArtePointer<RTRA>);
  
  void prima_love(void){

    if (ArteTable<MCIN>::booked())
      MCdata = 1;
    else
      MCdata = 0;
    
    if(firstcall){
      if (!RoverHidden)
	cout<<"ROVER version 1.07 started"<<endl;
      if(lover == static_cast<Lover*>(0))
	lover = new Lover();
      firstcall = 0;
    } 

    for(int itar=1;itar<=8;itar++){
      rver_per_wnr[itar] = 0;
      ntr_per_rver[itar] = 0;
    }

    lover->lover_primaries();

    if (!RoverHidden)
      lover->verona_event();

    RoverNver = nver;
  }
  
  
  void second_love(ArtePointer<RTRA> rtra1, ArtePointer<RTRA> rtra2){
    
    
    if (ArteTable<MCIN>::booked())
      MCdata = 1;
    else
      MCdata = 0;
    
    if(firstcall){
      if (!RoverHidden)
	cout<<"ROVER version 1.07 started"<<endl;
      if(lover == static_cast<Lover*>(0))
	lover = new Lover();
      firstcall = 0;
    } 
    
    lover->create_secondary_map(rtra1, rtra2);
    
    lover->lover_secondary();
    
  }
  
	  }

Lover::Lover(){
  mintracksinvertex = 3;
  if (ROVER_DRAW)
    graph_ini();
}

Lover::~Lover(){
  if (ROVER_DRAW)
    graph_close();

}

void Lover::create_secondary_map(ArtePointer<RTRA> rtra1, ArtePointer<RTRA> rtra2){

  float local[]  = {0,0,0,0,0};
  float global[] = {0,0,0,0,0};

  Tracks_t trackstruct;
  secondary_tracks.clear();
  
  // transform to the local coordinate system
      
  global[0] = rtra1->xf;
  global[1] = rtra1->yf;
  global[2] = rtra1->zf;
  global[3] = rtra1->txf;
  global[4] = rtra1->tyf;
  acGlobalToLocal(local, global, 1);

  trackstruct.rtra_rseg = 1;     // rtra stored
  trackstruct.rtra_index = 1;
  trackstruct.rtra = rtra1;
  trackstruct.x  = local[0];
  trackstruct.y  = local[1];
  trackstruct.z  = local[2];
  trackstruct.tx = local[3];
  trackstruct.ty = local[4];
  trackstruct.p =  rtra1->pf;
  for(int i=0;i<15;++i){
    trackstruct.C[i] = rtra1->cvf[i];
  }
  trackstruct.used  = 0;
  trackstruct.nhitb = rtra1->hsi;
  
  secondary_tracks[1] = trackstruct;

  global[0] = rtra2->xf;
  global[1] = rtra2->yf;
  global[2] = rtra2->zf;
  global[3] = rtra2->txf;
  global[4] = rtra2->tyf;
  acGlobalToLocal(local, global, 1);

  trackstruct.rtra_rseg = 1;     // rtra stored
  trackstruct.rtra_index = 2;
  trackstruct.rtra = rtra2;
  trackstruct.x  = local[0];
  trackstruct.y  = local[1];
  trackstruct.z  = local[2];
  trackstruct.tx = local[3];
  trackstruct.ty = local[4];
  trackstruct.p =  rtra2->pf;
  for(int i=0;i<15;++i){
    trackstruct.C[i] = rtra2->cvf[i];
  }
  trackstruct.used  = 0;
  trackstruct.nhitb = rtra2->hsi;//distance(rseg2->hitb.begin(),rseg2->hitb.end());
  
  secondary_tracks[2] = trackstruct;

}


void Lover::lover_secondary(void){

  typedef struct{
    float cov[4][4];
    float Tcov[2][2];
    float Tpar[2];
    float dzeta[4];
    float zf;
  } TRACK_TYPE, *pTRACK_TYPE;
  
  TRACK_TYPE atr[2]; 
  float Vcov[3][3];
  float Rk[5];
  float Vpar[3];
  float dG[5][5],KH[5][5],Gk[5][5];
  float Hk[4][5];
  float Kk[5][4];
  float L[4][4];
  float Fk[5][4];
  float ax,ay,dz,detr,zref=5.0;
  float res[4],ext[4];
  int i,j,itr,k,jjj;

  Tracks_mi it1 = secondary_tracks.begin();
  Tracks_mi it2 = secondary_tracks.begin(); it2++;
   
  trackincluster[1] = it1;
  trackincluster[2] = it2;
   
  ntrincluster = 2;
   
  vertex.mass = 0.0;vertex.merr = 0.0;
  trackpropagate(it1->second,zref);trackpropagate(it2->second,zref);
  
  detr =((it2->second).tx-(it1->second).tx)*
        ((it2->second).tx-(it1->second).tx);
  detr+=((it2->second).ty-(it1->second).ty)*
        ((it2->second).ty-(it1->second).ty);
  vertex.zv=((it2->second).tx-(it1->second).tx)*
        ((it2->second).x-(it1->second).x);
  vertex.zv+=((it2->second).ty-(it1->second).ty)*
        ((it2->second).y-(it1->second).y);
  vertex.zv=zref-vertex.zv/detr;
  dz=vertex.zv-zref;
  vertex.xv=0.5*((it2->second).x+dz*(it2->second).tx+
		 (it1->second).x+dz*(it1->second).tx);
  vertex.yv=0.5*((it2->second).y+dz*(it2->second).ty+
		 (it1->second).y+dz*(it1->second).ty);
  Vpar[0]=vertex.xv; Vpar[1]=vertex.yv; Vpar[2]=vertex.zv;
  vertex.cv[0]=3.0;vertex.cv[2]=3.0;vertex.cv[5]=25.0;
  atr[0].cov[0][0]=(it1->second).C[0];
  atr[0].cov[0][1]=(it1->second).C[1];
  atr[0].cov[1][1]=(it1->second).C[2];
  atr[0].cov[0][2]=(it1->second).C[3];
  atr[0].cov[1][2]=(it1->second).C[4];
  atr[0].cov[2][2]=(it1->second).C[5];
  atr[0].cov[0][3]=(it1->second).C[6];
  atr[0].cov[1][3]=(it1->second).C[7];
  atr[0].cov[2][3]=(it1->second).C[8];
  atr[0].cov[3][3]=(it1->second).C[9];
  atr[0].cov[1][0]=atr[0].cov[0][1];
  atr[0].cov[2][0]=atr[0].cov[0][2];
  atr[0].cov[3][0]=atr[0].cov[0][3];
  atr[0].cov[2][1]=atr[0].cov[1][2];
  atr[0].cov[3][1]=atr[0].cov[1][3];
  atr[0].cov[3][2]=atr[0].cov[2][3];
  atr[0].dzeta[0]=(it1->second).x;
  atr[0].dzeta[1]=(it1->second).y;
  atr[0].dzeta[2]=(it1->second).tx;
  atr[0].dzeta[3]=(it1->second).ty;
  atr[0].zf=zref;
  atr[0].Tpar[0]= atr[0].dzeta[2];
  atr[0].Tpar[1]= atr[0].dzeta[3];
  for(i=0;i<2;i++) for(j=0;j<2;j++)  
    atr[0].Tcov[i][j]=atr[0].cov[i+2][j+2];  
  atr[1].cov[0][0]=(it2->second).C[0];
  atr[1].cov[0][1]=(it2->second).C[1];
  atr[1].cov[1][1]=(it2->second).C[2];
  atr[1].cov[0][2]=(it2->second).C[3];
  atr[1].cov[1][2]=(it2->second).C[4];
  atr[1].cov[2][2]=(it2->second).C[5];
  atr[1].cov[0][3]=(it2->second).C[6];
  atr[1].cov[1][3]=(it2->second).C[7];
  atr[1].cov[2][3]=(it2->second).C[8];
  atr[1].cov[3][3]=(it2->second).C[9];
  atr[1].cov[1][0]=atr[1].cov[0][1];
  atr[1].cov[2][0]=atr[1].cov[0][2];
  atr[1].cov[3][0]=atr[1].cov[0][3];
  atr[1].cov[2][1]=atr[1].cov[1][2];
  atr[1].cov[3][1]=atr[1].cov[1][3];
  atr[1].cov[3][2]=atr[1].cov[2][3];
  atr[1].dzeta[0]=(it2->second).x;
  atr[1].dzeta[1]=(it2->second).y;
  atr[1].dzeta[2]=(it2->second).tx;
  atr[1].dzeta[3]=(it2->second).ty;
  atr[1].zf=zref;
  atr[1].Tpar[0]= atr[1].dzeta[2];
  atr[1].Tpar[1]= atr[1].dzeta[3];
  for(i=0;i<2;i++) for(j=0;j<2;j++)  
    atr[1].Tcov[i][j]=atr[1].cov[i+2][j+2];
  
  for(jjj=0;jjj<1;jjj++){
    for(i=0;i<3;i++) for(j=0;j<3;j++) Vcov[i][j]=0.0;
    Vcov[0][0]=vertex.cv[0];Vcov[1][1]=vertex.cv[2];Vcov[2][2]=vertex.cv[5];
    for(itr=0;itr<2;itr++)
      {
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
	for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
	for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
	for (i=0;i<3;i++) Rk[i]=Vpar[i];
	Rk[3]=atr[itr].Tpar[0];Rk[4]=atr[itr].Tpar[1];
	
	dz=atr[itr].zf-Rk[2];ax=Rk[3];ay=Rk[4];      
	for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
	Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
	Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
      
	for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
      
	L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	  2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
	L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	  2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
	L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	  ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
	L[1][0]=L[0][1];L[2][2]=Gk[3][3];
	L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
	for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];
	four_dim_inv(L);

	ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
	for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
	
	Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
	Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
	Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
	Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
	Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
	Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
	Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
	Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
	Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
	Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
	Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++)
	  {
	    Kk[i][j]=0.0;
	    for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	  }
	
	for(i=0;i<5;i++) for(j=0;j<4;j++) Rk[i]+=Kk[i][j]*res[j];
	
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    KH[i][j]=0.0;
	    for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	  }
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    dG[i][j]=0.0;
	    for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	  }
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=dG[i][j];
	for(i=0;i<3;i++) Vpar[i]=Rk[i];
	for(i=0;i<3;i++) for (j=0;j<3;j++) Vcov[i][j]=Gk[i][j];
      }

    for(itr=0;itr<2;itr++) 
      {
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
	for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
	for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
      
	for (i=0;i<3;i++) Rk[i]=Vpar[i];
	Rk[3]=atr[itr].Tpar[0];Rk[4]=atr[itr].Tpar[1];
      
	dz=atr[itr].zf-Rk[2];ax=Rk[3];ay=Rk[4];
      
	for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
	Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
	Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
      
	for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
      
	L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	  2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
	L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	  2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
	L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	  ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
	L[1][0]=L[0][1];L[2][2]=Gk[3][3];
	L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
	for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];      
	four_dim_inv(L);
	ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
	for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
	Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
	Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
	Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
	Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
	Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
	Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
	Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
	Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
	Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
	Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
	Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++)
	  {
	    Kk[i][j]=0.0;
	    for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	  }
	
	for(i=0;i<5;i++) for(j=0;j<4;j++) Rk[i]+=Kk[i][j]*res[j];
      
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    KH[i][j]=0.0;
	    for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	  }
	
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    dG[i][j]=0.0;
	    for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	  }
      
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=dG[i][j];

	atr[itr].Tpar[0]=Rk[3];atr[itr].Tpar[1]=Rk[4];
	Vpar[0]=Rk[0];Vpar[1]=Rk[1];Vpar[2]=Rk[2];
      }
  }
  vertex.xv    = Vpar[0];
  vertex.yv    = Vpar[1];
  vertex.zv    = Vpar[2];
  vertex.cv[0] = Vcov[0][0]; 
  vertex.cv[1] = Vcov[0][1];
  vertex.cv[2] = Vcov[1][1];
  vertex.cv[3] = Vcov[0][2];
  vertex.cv[4] = Vcov[1][2];
  vertex.cv[5] = Vcov[2][2];
  vertex.nt = ntrincluster;

  //   cout<<"Secondary (found) : "<<vertex.xv<<", "<<vertex.yv<<", "<<vertex.zv<<endl;
  //   cout<<sqrt(vertex.cv[0])<<", "<<sqrt(vertex.cv[2])<<", "<<sqrt(vertex.cv[5])<<endl;

   second_x = vertex.xv;
   second_y = vertex.yv;
   second_z = vertex.zv;

   
   if (!RoverHidden){
     ArteTable<RVER>::push_back(vertex);

     ArtePointer<RVER> rver(ArteTable<RVER>::size());
     RREL relation;
     
     for(int itr=1; itr<=ntrincluster; itr++){
       Tracks_mi it = trackincluster[itr];
       if ((it->second).rtra_rseg){
	 // state relations to RTRA
	 relation.chi  = 0. ;
	 relation.flag = 1  ;
	 ArteTable<RREL>::push_back(relation);
	 
	 ArtePointer<RREL> rrel(ArteTable<RREL>::size());
	 ArtePointer<RTRA> rtra = (it->second).rtra;
	 createRelation(rver, rrel, &RVER::rrel);
	 createRelation(rtra, rrel, &RTRA::rrel);
	 //rtra->ever = rver;

       }
     }
   }
   

}


void Lover::lover_primaries(void){

  //  const float Chi2_cut=10.0;
  //  float krit2;


  for(int i = 0; i<rver_array_size; i++){ // general clean arrays for Michael
    rseg_to_rver[i] = 0; 
    rtra_to_rver[i] = 0; 
  }

  filltrackmap(tracks);

  get_targets();

  if (ROVER_DRAW)
    draw_event();

  if (!RoverHidden)
    ArteTable<RVER>::clear();
  nver = 0;

  for(Target_mi tarit = targets.begin();tarit != targets.end();++tarit){
    int wnr = (tarit->second).wnr;
    int ntr = 0;
    for(Tracks_mi it = tracks.begin();it != tracks.end();++it){
      trackpropagate(it->second,(tarit->second).z);
      if(target_near(it->second,tarit->first,3.5)){
	ntr++;
	trackonwire[wnr][ntr] = it;
	trackweight[wnr][ntr] = 0.0;;
      }
    }
    ntronwire[wnr] = ntr;
  }

  int   vertexfound = 1;
  ntrincluster = 0;

  while(true){   
    //  while(vertexfound){   
    vertexfound = 0;
    
    Tracks_mi it_best = tracks.begin();
    float w_best      = 0.0;
    int max_clu = 0;
    Target_mi   wire_best   = 0;
    Tracks_mi it, jt;
    int tracksonwire;
    for(Target_mi tarit = targets.begin();tarit != targets.end();++tarit){
      int wnr = (tarit->second).wnr;
      tracksonwire = ntronwire[wnr];
      if(tracksonwire < mintracksinvertex) 
	continue;
      int itrnew = 0;

      for (int itr=1; itr <= tracksonwire; itr++){ // remove used tracks
	it = trackonwire[wnr][itr];
	if ((it->second).used)
	  continue;
	itrnew++;
	trackonwire[wnr][itrnew] = trackonwire[wnr][itr];
	trackweight[wnr][itrnew] = trackweight[wnr][itr];
	trackpropagate(it->second,(tarit->second).z);
      }
      tracksonwire = ntronwire[wnr] = itrnew;
      if(tracksonwire < mintracksinvertex) 
	continue;

      //      cout<<"Wire "<<wnr<<" ";
      for (int itr=1; itr <= tracksonwire; itr++){ // calculate weights
	it = trackonwire[wnr][itr];
	(it->second).w = 0.0;
	float x1 = (it->second).x;
	float y1 = (it->second).y;
	int ntr_clu = 0;
	for (int jtr=1; jtr <= tracksonwire; jtr++){ 
	  jt = trackonwire[wnr][jtr];
	  float x2 = (jt->second).x;
	  float y2 = (jt->second).y;

	  //	  krit2 = (x1-x2)*(x1-x2)/((it->second).C[0]+(jt->second).C[0])+
	  //	          (y1-y2)*(y1-y2)/((it->second).C[2]+(jt->second).C[2]);
	  //	  if(krit2 < 7.0)
	  //	    {
	  //	      (it->second).w+=exp(-krit2);
	  //	      ntr_clu++;
	  //	    }

          float dist2 = (x1-x2)*(x1-x2)+(y1-y2)*(y1-y2);
          float cut2 = 3.5*3.5*
            ((it->second).C[0]+(it->second).C[2]+
             (jt->second).C[0]+(jt->second).C[2]);
          if(dist2 < cut2){
            (it->second).w+=(1.0-(dist2/cut2))*(1.0-(dist2/cut2));///
            //        (sqrt((jt->second).C[0])+sqrt((jt->second).C[2]));
            ntr_clu++;
          }
	}
	//	cout<<(it->second).w<<"/"<<ntr_clu<<" ";
	if ((it->second).w>w_best){
          //if (ntr_clu>max_clu){
	    it_best = it;
	    w_best  = (it->second).w;
	    max_clu = ntr_clu;
	    wire_best = tarit;
	  }
      }
      //      cout<<endl<<endl;
    }
    if (w_best==0.0)
      break;

#ifdef XXX
    if (ROVER_DRAW){
      
      ismk(30); ismksc(7.0);	ispmci(Green); isplci(Green);
      
      float xpm[1], ypm[1];
      	
      xpm[0] = (it_best->second).x;
      ypm[0] = (it_best->second).y;
      ipm(1,xpm,ypm);
	
      cout<<"Best track on wire "<<(wire_best->second).wnr<<" : "
	  <<sqrt((it_best->second).C[0])<<" "
	  <<sqrt((it_best->second).C[2])<<" has "
	  <<max_clu<<" neighbours"<<endl;
    }
#endif

    int wnr = (wire_best->second).wnr;
    tracksonwire = ntronwire[wnr];

    for (int itr=1; itr <= tracksonwire; itr++){ // propagate tracks
      it = trackonwire[wnr][itr];
      trackpropagate(it->second,(wire_best->second).z);
    }

    for (int itr=1; itr <= tracksonwire; itr++){ // calculate weights
      it = trackonwire[wnr][itr];
      (it->second).w = 0.0;
      float x1 = (it->second).x;
      float y1 = (it->second).y;
      for (int jtr=1; jtr <= tracksonwire; jtr++){ 
	jt = trackonwire[wnr][jtr];
	float x2 = (jt->second).x;
	float y2 = (jt->second).y;

	//	krit2 = (x1-x2)*(x1-x2)/((it->second).C[0]+(jt->second).C[0])+
	//	        (y1-y2)*(y1-y2)/((it->second).C[2]+(jt->second).C[2]);
	//	if(krit2 < Chi2_cut) (it->second).w+=exp(-krit2);

        float dist2 = (x1-x2)*(x1-x2)+(y1-y2)*(y1-y2);
        float cut2 = 3.5*3.5*
          ((it->second).C[0]+(it->second).C[2]+
           (jt->second).C[0]+(jt->second).C[2]);
        if(dist2 < cut2)
          (it->second).w+=(1.0-(dist2/cut2))*(1.0-(dist2/cut2));///
        //            (sqrt((jt->second).C[0])+sqrt((jt->second).C[2]));
      }
    }
    float x_best  = (it_best->second).x;
    float y_best  = (it_best->second).y;
    float c0_best = (it_best->second).C[0];
    float c2_best = (it_best->second).C[2];
    float vertx = 0.0, verty = 0.0, vertweight = 0.0;
    ntrincluster = 0;
    for (int itr=1; itr <= tracksonwire; itr++){ // collect cluster
      it = trackonwire[wnr][itr];
      float x = (it->second).x;
      float y = (it->second).y;
      float w = (it->second).w;
      
      //      krit2 = (x-x_best)*(x-x_best)/((it->second).C[0]+c0_best)+
      //	(y-y_best)*(y-y_best)/((it->second).C[2]+c2_best);
      //      if(krit2<Chi2_cut){
      
      float dist2 = (x-x_best)*(x-x_best)+(y-y_best)*(y-y_best);
      float cut2 = 3.5*3.5*
	((it->second).C[0]+(it->second).C[2]+
         c0_best+c2_best);
      if(dist2 < cut2){
	vertx += w*x;
	verty += w*y;
	vertweight += w;
	trackincluster[++ntrincluster] = it;
      }
    }

    if (ntrincluster < mintracksinvertex){
      //      cout<<"BREAK with "<<ntrincluster<<" tracks in cluster"<<endl;
      break;
    }

    if(ntrincluster >= mintracksinvertex){

      vertex.mass = 0.0;
      vertex.merr = 0.0;
      
      vertx /= vertweight;
      verty /= vertweight;      
      vertex.xv = vertx;
      vertex.yv = verty;
      vertex.zv = (wire_best->second).z;
      vertex.wire = (wire_best->second).wnr;
      
      float cov_x = 0.0;
      float cov_y = 0.0;
      float cov_z = 0.0;
      for(int itr=1; itr<=ntrincluster; itr++){
	it = trackincluster[itr];
	cov_x += ((it->second).x-vertx)*((it->second).x-vertx);
	cov_y += ((it->second).y-verty)*((it->second).y-verty);
      }
      cov_x /= ntrincluster;
      cov_y /= ntrincluster;
      cov_z  = 1.0e-2;
      
      vertex.cv[0] = cov_x; //float(vcov[0]) == Vcov[0][0];     
      vertex.cv[1] = 0.0;   //float(vcov[3]) == Vcov[0][1];
      vertex.cv[2] = cov_y; //float(vcov[4]) == Vcov[1][1];
      vertex.cv[3] = 0.0;   //float(vcov[6]) == Vcov[0][2];     
      vertex.cv[4] = 0.0;   //float(vcov[7]) == Vcov[1][2];
      vertex.cv[5] = cov_z; //float(vcov[8]) == Vcov[2][2];
      vertex.nt = ntrincluster;
      
#ifdef XXX
      if (ROVER_DRAW){
	ismk(30); ismksc(3.0); ispmci(Blue); isplci(Blue);
	  
	float xpm[1], ypm[1];
	xpm[0]=vertex.xv;
	ypm[0]=vertex.yv;
	ipm(1,xpm,ypm);
	
	float xv = vertex.xv;
	float dx = sqrt(vertex.cv[0]);
	float yv = vertex.yv;
	float dy = sqrt(vertex.cv[2]);
	
	float xpl[5], ypl[5];
	xpl[0] = xv-dx;
	ypl[0] = yv-dy;
	xpl[1] = xv+dx;
	ypl[1] = yv-dy;
	xpl[2] = xv+dx;
	ypl[2] = yv+dy;
	xpl[3] = xv-dx;
	ypl[3] = yv+dy;
	xpl[4] = xv-dx;
	ypl[4] = yv-dy;
	ipl(5,xpl,ypl);
      }
#endif

      vertex_refit();

      if (vertex_ok){
	//      if (vertex.nt>=mintracksinvertex){
	vertexfound = 1;
	rver_per_wnr[(wire_best->second).wnr]++;
	if (nver<=8)
	  ntr_per_rver[nver] = vertex.nt;


	// modify the target wire position
	
	float dx = 0.1, dy = 0.1;
	
	float xt    = (wire_best->second).x;
	float yt    = (wire_best->second).y;
	float zt    = (wire_best->second).z;
	float dxt   = (wire_best->second).dx;
	float dyt   = (wire_best->second).dy;
	float dzt   = (wire_best->second).dz;
	int   nt_tr = (wire_best->second).n_tr;
	int   nt_vx = (wire_best->second).n_vx;
	int   nv  = vertex.nt;
	
	float x = vertex.xv;
	float y = vertex.yv;
	float z = vertex.zv;
	
	
	if ((fabs(x-xt)<dx)&&(fabs(y-yt)<dy)){
	  xt = xt*(nt_tr+nt_vx)+x*3*nv;
	  yt = yt*(nt_tr+nt_vx)+y*3*nv;
	  zt = zt*nt_vx+z*3*nv;
	  nt_vx+= 3*nv;
	  xt /= (nt_tr+nt_vx);
	  yt /= (nt_tr+nt_vx);
	  zt /= nt_vx;
	  
	  dxt = (dxt*(nt_tr+nt_vx-3*nv)+nv*fabs(x-xt))/(nt_tr+nt_vx);
	  dyt = (dyt*(nt_tr+nt_vx-3*nv)+nv*fabs(y-yt))/(nt_tr+nt_vx);
	  dzt = (dzt*(nt_vx-3*nv)+3*nv*fabs(z-zt))/nt_vx;
	}
	
	if (nt_tr+nt_vx > 10000){
	  nt_tr = 100;
	  nt_vx = 100;
	}	

	(wire_best->second).x    = xt;;
	(wire_best->second).y    = yt;
	(wire_best->second).z    = zt;
	(wire_best->second).dx   = dxt;
	(wire_best->second).dy   = dyt;
	(wire_best->second).dz   = dzt;
	(wire_best->second).n_tr = nt_tr;
	(wire_best->second).n_vx = nt_vx;
	

	if (ROVER_DRAW){

	  ismk(3); ismksc(0.5);	ispmci(LightRed); isplci(LightRed);

	  float xpm[1], ypm[1];
	  int ntr_ver = 0;
	  for(int itr=1; itr<=ntrincluster; itr++){
	    Tracks_mi it = trackincluster[itr];
	    if (!(it->second).used) 
	      continue;

	    ntr_ver++;

	    xpm[0] = (it->second).x;
	    ypm[0] = (it->second).y;
	    ipm(1,xpm,ypm);
	    
	    float xm = (it->second).x;
	    float ym = (it->second).y;
	    float dx = sqrt((it->second).C[0]);
	    float dy = sqrt((it->second).C[2]);
	    
	    igarc(xm, ym, 0.5*(dx+dy),0.5*(dx+dy), 0.0, 0.0);
	  }

	  /*
	  if (ntr_ver<=4){
	    cout<<"vertex with "<<ntr_ver<<" tracks : ";
	    for(int itr=1; itr<=ntrincluster; itr++){
	      Tracks_mi it = trackincluster[itr];
	      if (!(it->second).used) 
		continue;
	      cout<<(it->second).nhitb<<" ";
	    }
	    cout<<endl;
	  }
	  */

	  ismk(30); ismksc(5.0); ispmci(Red); isplci(Green);
	  
	  xpm[0]=vertex.xv;
	  ypm[0]=vertex.yv;
	  ipm(1,xpm,ypm);
	  
	  float xv = vertex.xv;
	  float dx = 10.0*sqrt(vertex.cv[0]);
	  float yv = vertex.yv;
	  float dy = 10.0*sqrt(vertex.cv[2]);
	  
	  float xpl[5], ypl[5];
	  xpl[0] = xv-dx;
	  ypl[0] = yv-dy;
	  xpl[1] = xv+dx;
	  ypl[1] = yv-dy;
	  xpl[2] = xv+dx;
	  ypl[2] = yv+dy;
	  xpl[3] = xv-dx;
	  ypl[3] = yv+dy;
	  xpl[4] = xv-dx;
	  ypl[4] = yv-dy;
	  ipl(5,xpl,ypl);
	}
	
      }

#ifdef XXX

      cout<<"Primary :"<<vertex.xv<<", "<<vertex.yv<<", "<<vertex.zv<<endl;

      Tracks_mi it1 = trackincluster[1];
      Tracks_mi it2 = trackincluster[2];

      float tmp_x = vertex.xv;
      float tmp_y = vertex.yv;
      float tmp_z = vertex.zv;

      if (vertex.nt>10){

	second_rove((it1->second).rseg, (it2->second).rseg);
	
	hfill(2011, vertex.xv-tmp_x, 0.0, 1.0);
	hfill(2012, vertex.yv-tmp_y, 0.0, 1.0);
	hfill(2013, vertex.zv-tmp_z, 0.0, 1.0);
	
	hfill(2021, (vertex.xv-tmp_x)/sqrt(vertex.cv[0]), 0.0, 1.0);
	hfill(2022, (vertex.yv-tmp_y)/sqrt(vertex.cv[2]), 0.0, 1.0);
	hfill(2023, (vertex.zv-tmp_z)/sqrt(vertex.cv[5]), 0.0, 1.0);
      }

#endif

    }
  }


  if (ROVER_DRAW){

    ismk(3); ismksc(0.5); ispmci(Blue); isplci(Blue);

    float xpm[1], ypm[1];

    for(Target_mi tarit = targets.begin();tarit != targets.end();++tarit){
      int wnr = (tarit->second).wnr;
      int tracksonwire = ntronwire[wnr];

      for (int itr=1; itr <= tracksonwire; itr++){ // draw unused tracks
	Tracks_mi it = trackonwire[wnr][itr];
	if ((it->second).used)
	  continue;

	trackpropagate(it->second,(tarit->second).z);
	
	xpm[0] = (it->second).x;
	ypm[0] = (it->second).y;
	ipm(1,xpm,ypm);
	
	float xm = (it->second).x;
	float ym = (it->second).y;
	float dx = sqrt((it->second).C[0]);
	float dy = sqrt((it->second).C[2]);
	
	igarc(xm, ym, 0.5*(dx+dy),0.5*(dx+dy), 0.0, 0.0);
      }
    }


    iuwk(1,1);
    
  }

  if (!RoverHidden)
    for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
	 rverIt != ArteTable<RVER>::end() ; ++rverIt){
      
      // transform to global coordinates !!!
      
      float local[]  = {0,0,0,0,0};
      float global[] = {0,0,0,0,0};
      
      local[0] = (*rverIt)->xv;
      local[1] = (*rverIt)->yv;
      local[2] = (*rverIt)->zv;
      acLocalToGlobal(local, global, 1);
      (*rverIt)->xv = global[0];
      (*rverIt)->yv = global[1];
      (*rverIt)->zv = global[2];
    }
  
#ifdef XXX
  int ntr = 0;
  for( ArteTable<RTRA>::iterator rtraIt = ArteTable<RTRA>::begin(); 
       rtraIt != ArteTable<RTRA>::end(); ++rtraIt ){
    if((*rtraIt)->hsi>0){
      ++ntr;
      cout << "RTRA "<< ntr<<" on vertex "<<rtra_to_rver[ntr]<<endl;
    }
  } 
  // additions for using rsegs also
  ntr = 0;
  for( ArteTable<RSEG>::iterator rsegIt = ArteTable<RSEG>::begin(); 
       rsegIt != ArteTable<RSEG>::end(); ++rsegIt ){
    if(  !((*rsegIt)->cmp & ~(Rsegc::vxd | Rsegc::bit16)) ){
      ++ntr;
      cout << "RSEG "<< ntr<<" on vertex "<<rseg_to_rver[ntr]<<endl;
    }
  } 
  
#endif
  
}


void Lover::filltrackmap(Tracks_m& tracks){
  
  Tracks_t trackstruct;
  tracks.clear();
  int ntracks = 0;
  float local[]  = {0,0,0,0,0};
  float global[] = {0,0,0,0,0};
  
  int nrseg = 0;

  for( ArteTable<RSEG>::iterator rsegIt = ArteTable<RSEG>::begin(); 
       rsegIt != ArteTable<RSEG>::end(); ++rsegIt ){
    if(  !((*rsegIt)->cmp & ~(Rsegc::vxd | Rsegc::bit16)) ){


      // transform to the local coordinate system
	
      global[0] = (*rsegIt)->xf;
      global[1] = (*rsegIt)->yf;
      global[2] = (*rsegIt)->zf;
      global[3] = (*rsegIt)->txf;
      global[4] = (*rsegIt)->tyf;
      acGlobalToLocal(local, global, 1);

      ++ntracks;

      ++nrseg;

      rseg_to_rver[nrseg] = 0; // clean array for Michael


      trackstruct.index = ntracks;
      trackstruct.rtra_rseg = 0;          // rseg stored
      trackstruct.rseg_index = nrseg;
      trackstruct.rseg = *rsegIt;
      trackstruct.rtra = ArteTable<RTRA>::end();
      trackstruct.x  = local[0];
      trackstruct.y  = local[1];
      trackstruct.z  = local[2];
      trackstruct.tx = local[3];
      trackstruct.ty = local[4];
      trackstruct.p  = (*rsegIt)->pf;
      for(int i=0;i<15;++i)
	trackstruct.C[i] = (*rsegIt)->cvf[i];
      trackstruct.used  = 0;
      trackstruct.nhitb = int((*rsegIt)->time+(*rsegIt)->dtim);
      //distance((*rsegIt)->hitb.begin(),(*rsegIt)->hitb.end());

      tracks[trackstruct.index] = trackstruct;
    }
  } 

  int nrtra = 0;

  for( ArteTable<RTRA>::iterator rtraIt = ArteTable<RTRA>::begin();
       rtraIt != ArteTable<RTRA>::end(); ++rtraIt ){
    if( ((*rtraIt)->hsi)>0 ){
      ++nrtra;

      rtra_to_rver[nrtra] = 0; // clean array for Michael

      // transform to the local coordinate system
      
      global[0] = (*rtraIt)->xf;
      global[1] = (*rtraIt)->yf;
      global[2] = (*rtraIt)->zf;
      global[3] = (*rtraIt)->txf;
      global[4] = (*rtraIt)->tyf;
      acGlobalToLocal(local, global, 1);

      int rtra_unique = 1;
      for( ArteRelationIterator<RSEG> rsegIt = (*rtraIt)->rseg.begin();
	   rsegIt != (*rtraIt)->rseg.end(); ++rsegIt){
	if(  !((*rsegIt)->cmp & ~(Rsegc::vxd | Rsegc::bit16)) ){
	  // find corresponding rseg in tracks
	  for(Tracks_mi it = tracks.begin();it != tracks.end();++it){
	    if ((it->second).rseg == *rsegIt){ 
	      rtra_unique = 0;

	      // rseg found -> check rtra
	      if (((*rtraIt)->cvf[0]< 0.0)||((*rtraIt)->cvf[2]< 0.0)||
		  ((*rtraIt)->cvf[5]< 0.0)||((*rtraIt)->cvf[9]< 0.0)) continue;

	      (it->second).rtra_rseg = 1;     // now rtra stored
	      (it->second).rtra_index = nrtra;
	      (it->second).rtra = *rtraIt;
	      (it->second).x  = local[0];
	      (it->second).y  = local[1];
	      (it->second).z  = local[2];
	      (it->second).tx = local[3];
	      (it->second).ty = local[4];
	      (it->second).p =  (*rtraIt)->pf;
	      for(int i=0;i<15;++i){
		(it->second).C[i] = (*rtraIt)->cvf[i];
	      }
	      (it->second).used  = 0;
	      (it->second).nhitb = (*rtraIt)->hsi;

	    }
	  }
	  
	}
      }

      if (rtra_unique) {
	++ntracks;
	trackstruct.index = ntracks;
	trackstruct.rtra_rseg = 1;          // rtra stored
	trackstruct.rtra_index = nrtra;
	trackstruct.rtra = *rtraIt;
	trackstruct.rseg = ArteTable<RSEG>::end();
	trackstruct.x  = local[0];
	trackstruct.y  = local[1];
	trackstruct.z  = local[2];
	trackstruct.tx = local[3];
	trackstruct.ty = local[4];
	trackstruct.p =  (*rtraIt)->pf;
	for(int i=0;i<15;++i){
	  trackstruct.C[i] = (*rtraIt)->cvf[i];
	}
	trackstruct.used  = 0;
	trackstruct.nhitb = (*rtraIt)->hsi;

	tracks[trackstruct.index] = trackstruct;
	
      }
    }
  }

  
}




void Lover::get_targets(void){
  float dx = 0.1, dy = 0.1, dz = 0.1;
  float local[]  = {0,0,0,0,0};
  float global[] = {0,0,0,0,0};

  if (rover_newgeometry){
    rover_newgeometry = 0;

    //	cout<<"Beam "<<BEAM.x<<" "<<BEAM.y<<" "<<BEAM.z<<endl; //???
    
    float beam_x = 0.0, beam_y = 0.0;  // beam center for simu data

    //    if (!ArteTable<MCIN>::booked()){
      beam_x =  0.30;                  // beam center for real data in local coordinates
      beam_y = -0.05;
      //    } // the same for real and simu data
    
    cout.setf(ios::fixed);
    cout.setf(ios::showpoint);
    cout.precision(4);
    if (!RoverHidden)
      cout<<endl;

    float zsta1 = 0.0, zsta2 = 0.0;
    int   nsta1 = 0,   nsta2 = 0;
    //initialize target arrays used for allignemt
    for( ArteTable<GTAR>::iterator gtarIt = ArteTable<GTAR>::begin();
	 gtarIt != ArteTable<GTAR>::end(); ++gtarIt ){

      //      if ((MCdata)&&((*gtarIt)->z<-2.5))
      //      	continue;

      targetstruct target;
      target.wnr = (*gtarIt)->wnr;
      target.x   = (*gtarIt)->x;
      target.y   = (*gtarIt)->y;
      target.z   = (*gtarIt)->z;

      if ((MCdata)&&((*gtarIt)->wnr==2)){

	target.x   = (*gtarIt)->y;
	target.y   = (*gtarIt)->z;
	target.z   = (*gtarIt)->x;

      } 

      if ((fabs(target.x)>1.0)||(fabs(target.y)>1.0))
	continue;

      if (!RoverHidden){
	cout<<"ROVER : target wire "<<target.wnr << " at (";
	if (target.x>=0) 
	  cout<<" ";
	cout<<target.x<<", ";
	if (target.y>=0) 
	  cout<<" ";
	cout<<target.y<<", ";
	if (target.z>=0) 
	  cout<<" ";
	cout<<target.z<<" )"<<endl;
      }



      global[0] = target.x;
      global[1] = target.y;
      global[2] = target.z;
      acGlobalToLocal(local, global, 1);
      target.x = local[0];
      target.y = local[1];
      target.z = local[2];

      target.dx  = dx;
      target.dy  = dy;
      target.dz  = dz;

      if((*gtarIt)->dx > (*gtarIt)->dy){
	target.x = beam_x;
	target.flaghorver = 0; // hor
      }
      else{
	target.y = beam_y;
	target.flaghorver = 1; // ver
      }
      target.n_tr = 100;
      target.n_vx = 100;
      
      if (target.z<-2.5){
	nsta1++;
	zsta1 += target.z;
      }
      else{
	nsta2++;
	zsta2 += target.z;
      }

      targets.insert(pair<int,targetstruct>((*gtarIt)->wnr,target));
    }

    if (!RoverHidden)
      cout<<endl;

    if ((nsta1!=0)||(nsta2!=0)){

      if (nsta1>nsta2)
	zsta = zsta1/nsta1;
      else
	zsta = zsta2/nsta2;
    }

  }
  


  for(Target_mi tarit = targets.begin(); tarit != targets.end(); ++tarit){
    
    float xt     = (tarit->second).x;
    float yt     = (tarit->second).y;
    float dxt    = (tarit->second).dx;
    float dyt    = (tarit->second).dy;
    int   nt_tr  = (tarit->second).n_tr;
    int   nt_vx  = (tarit->second).n_vx;
    
    for(Tracks_mi it = tracks.begin();it != tracks.end();++it){
      trackpropagate(it->second,(tarit->second).z);
      
      float x = (it->second).x;
      float y = (it->second).y;
      
      
      if ((fabs(x-xt)<dx)&&(fabs(y-yt)<dy)){
	xt = xt*(nt_tr+nt_vx)+x;
	yt = yt*(nt_tr+nt_vx)+y;
	nt_tr++;
	xt /= (nt_tr+nt_vx);
	yt /= (nt_tr+nt_vx);
	
	dxt = (dxt*(nt_tr-1+nt_vx)+fabs(x-xt))/(nt_tr+nt_vx);
	dyt = (dyt*(nt_tr-1+nt_vx)+fabs(y-yt))/(nt_tr+nt_vx);
      }
    }

    if ((nt_tr+nt_vx) > 10000){
      nt_tr = 100;
      nt_vx = 100;
    }    

    (tarit->second).x     = xt;;
    (tarit->second).y     = yt;
    (tarit->second).dx    = dxt;
    (tarit->second).dy    = dyt;
    (tarit->second).n_tr  = nt_tr;
    (tarit->second).n_vx  = nt_vx;

    /*
    cout<<"ROVER : target wire "<<(tarit->second).wnr << " at (";
    if ((tarit->second).x>=0) 
      cout<<" ";
    cout<<(tarit->second).x<<", ";
    if ((tarit->second).y>=0) 
      cout<<" ";
    cout<<(tarit->second).y<<", ";
    if ((tarit->second).z>=0) 
      cout<<" ";
    cout<<(tarit->second).z<<" )"<<endl;
    */


  }
  
} 

void Lover::trackpropagate(Tracks_t& track,float z){
  
  if(track.z == z)
    return;
  
  float dz = z - track.z;
  
  if ((fabs(dz)< 5.0)||(dz>0)){
    track.x += dz * track.tx;
    track.y += dz * track.ty;
    track.z  = z;
    
    track.C[0]  += dz * (dz*track.C[5] + track.C[3] + track.C[3]);
    track.C[1]  += dz * (dz*track.C[8] + track.C[6] + track.C[4]);
    track.C[2]  += dz * (dz*track.C[9] + track.C[7] + track.C[7]);
    track.C[3]  += dz * track.C[5]; 
    track.C[4]  += dz * track.C[8]; 
    track.C[6]  += dz * track.C[8];
    track.C[7]  += dz * track.C[9];
    track.C[10] += dz * track.C[12];
    track.C[11] += dz * track.C[13];
  }
  else{
    CatsExtrapolateTrack(track, z);
    track.z = z;
    track.C[10] += dz * track.C[12];
    track.C[11] += dz * track.C[13];
  }
  
}


//=========================================================================================
  
void Lover::CatsExtrapolateTrack(Tracks_t& track, float z0){
    /****************************************************************
     *                                                              *
     *     This routine provides a procedure for the track          *
     *     parameters extrapolation to a vertex point               *
     *                                                              *
     *     (c) Dmitry Emeliyanov, 2000, DESY HH.                    *
     *                                                              *
     ***************************************************************/
    
    int   gewa[NGEWA];
    int   i, k, g, nfound;
    float Rk[4], Gk[4][4];
    float coeff, pinv, xp, yp, zp, l2sqrt, zprev, ms2, D_z;
    float Ax, Bx, Ay, By;


    if ((track.C[0]< 0.0)||(track.C[2]< 0.0)||(track.C[5]< 0.0)||(track.C[9]< 0.0)){
      if (++BadCovMatrix<10){
	if (!RoverHidden){
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------> INPUT bad cov matrix "<<track.C[0]<<" "<<track.C[2]<<" "<<track.C[5]<<" "<<track.C[9]<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	}
      }
      if (BadCovMatrix==10){
	if (!RoverHidden){
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------> INPUT bad cov matrix *** NO FURTHER WARNINGS ***"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	}
      }
    }

    Ax=track.tx;
    Bx=track.x-track.tx*track.z;
    Ay=track.ty;
    By=track.y-track.ty*track.z;
    
    l2sqrt = 1.0+Ax*Ax+Ay*Ay; l2sqrt = l2sqrt*sqrt(l2sqrt);
    
    pinv = track.p;
    coeff = (1.0+0.105*0.105*pinv*pinv)*pinv*pinv; // muon mass correction
      
    Rk[0] = track.x;           Rk[1] = track.y;
    Rk[2] = track.tx;          Rk[3] = track.ty;
      
    Gk[0][0] = track.C[0];    Gk[0][1] = track.C[1];
    Gk[1][1] = track.C[2];    Gk[0][2] = track.C[3];
    Gk[1][2] = track.C[4];    Gk[2][2] = track.C[5];
    Gk[0][3] = track.C[6];    Gk[1][3] = track.C[7];
    Gk[2][3] = track.C[8];    Gk[3][3] = track.C[9];
      
    int ngewaf = 0;
    nfound = 0;
    for(i=0; i<ngewaf; i++) {
      if (GEWA_zmax[i]<z0) continue;
      if (GEWA_zmax[i]>track.z) break;
      ngewaf = i;
      zp = 0.5*(GEWA_zmin[i]+GEWA_zmax[i]);
      xp = Ax*zp+Bx;
      yp = Ay*zp+By;
      
      if ((xp>=GEWA_xmin[i])&&(xp<=GEWA_xmax[i])&&
	  (yp>=GEWA_ymin[i])&&(yp<=GEWA_ymax[i]))
	gewa[nfound++]=i;
    }
    
    /********************* Fit for the point at z0 *********************/
    
    zprev=track.z;
    
    for(k=nfound-1; k>=0; k--) {
      g  = gewa[k];
      ms2= GEWA_ms2[g];
      
      /***** Multiple scattering corrections ****/
      
      Gk[2][2]+=(1.0+Rk[2]*Rk[2])*l2sqrt*coeff*ms2;
      Gk[3][3]+=(1.0+Rk[3]*Rk[3])*l2sqrt*coeff*ms2;
      Gk[2][3]+=Rk[2]*Rk[3]*l2sqrt*coeff*ms2;
      
      /***** Prediction step: from zprev to zmin *****/
      D_z=GEWA_zmin[g]-zprev; 
      zprev=GEWA_zmin[g];
      Rk[0]+=D_z*Rk[2];
      Rk[1]+=D_z*Rk[3];
      Gk[0][0]+=D_z*(2.0*Gk[0][2]+D_z*Gk[2][2]);
      Gk[0][1]+=D_z*(Gk[1][2]+Gk[0][3]+D_z*Gk[2][3]);
      Gk[0][2]+=D_z*Gk[2][2];
      Gk[0][3]+=D_z*Gk[2][3];
      Gk[1][1]+=D_z*(2.0*Gk[1][3]+D_z*Gk[3][3]);
      Gk[1][2]+=D_z*Gk[2][3];
      Gk[1][3]+=D_z*Gk[3][3];
    }
    
    /***** Prediction step : from zprev to z0 *****/
    
    D_z=z0-zprev;
    Rk[0]+=D_z*Rk[2];
    Rk[1]+=D_z*Rk[3];
    Gk[0][0]+=D_z*(2.0*Gk[0][2]+D_z*Gk[2][2]);
    Gk[0][1]+=D_z*(Gk[1][2]+Gk[0][3]+D_z*Gk[2][3]);
    Gk[0][2]+=D_z*Gk[2][2];
    Gk[0][3]+=D_z*Gk[2][3];
    Gk[1][1]+=D_z*(2.0*Gk[1][3]+D_z*Gk[3][3]);
    Gk[1][2]+=D_z*Gk[2][3];
    Gk[1][3]+=D_z*Gk[3][3];


    track.x  = Rk[0];           track.y  = Rk[1];
    track.tx = Rk[2];           track.ty = Rk[3];
    
    track.C[0] = Gk[0][0];    track.C[1] = Gk[0][1];
    track.C[2] = Gk[1][1];    track.C[3] = Gk[0][2];
    track.C[4] = Gk[1][2];    track.C[5] = Gk[2][2];
    track.C[6] = Gk[0][3];    track.C[7] = Gk[1][3];
    track.C[8] = Gk[2][3];    track.C[9] = Gk[3][3];


    if ((track.C[0]< 0.0)||(track.C[2]< 0.0)||(track.C[5]< 0.0)||(track.C[9]< 0.0)){
      if (BadCovMatrix<10){
	if (!RoverHidden){
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------> ROVER bad cov matrix "<<track.C[0]<<" "<<track.C[2]<<" "<<track.C[5]<<" "<<track.C[9]<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	}
      }
      if (BadCovMatrix==10){
	if (!RoverHidden){
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------> ROVER bad cov matrix *** NO FURTHER WARNINGS ***"<<endl;
	  cout<<"------------------>"<<endl;
	  cout<<"------------------>"<<endl;
	}
      }
    }

    
    return;   
}


//=======================================================================


int Lover::target_near(Tracks_t& track, int wire, float sigcut){
  float dist = 0.;
  float cut = 0.;
  
  if(targets[wire].flaghorver == 0){
    dist = track.y - (targets[wire].y);
    cut = sigcut*sigcut*track.C[2];
  }
  else{
    dist = track.x - (targets[wire].x);
    cut = sigcut*sigcut*track.C[0];
  }          

  if(dist*dist < cut )
    return 1;
  else
    return 0;  

} 



void Lover::vertex_refit(void){
  
  typedef struct{
    float cov[4][4];
    float Tcov[2][2];
    float Tpar[2];
    float dzeta[4];
    float zf;
    float ChiInc;
    int flag;
  } TRACK_TYPE, *pTRACK_TYPE;
  
  TRACK_TYPE atr[ROVERMAXTRACKS];
  int n_track;

  const float dchi2=20.0;
  const float Pfalse=0.10;
  const int   MaxIter=3;

  float Vcov[3][3];
  float Rk[5];
  float Vpar[3];
  float dG[5][5],KH[5][5],Gk[5][5];
  float Hk[4][5];
  float Kk[5][4];
  float L[4][4];
  float Fk[5][4];
  float lambda,ax,ay,beta1,dz,chi2;
  float res[4],ext[4];
  int i,j,itr,k,nValidTrack;

  Vpar[0]=vertex.xv; Vpar[1]=vertex.yv; Vpar[2]=vertex.zv;

  n_track=0;
  for(int itr=1; itr<=ntrincluster; itr++){
    Tracks_mi it = trackincluster[itr];
    trackpropagate(it->second,Vpar[2]);
    atr[n_track].cov[0][0]=(it->second).C[0];
    atr[n_track].cov[0][1]=(it->second).C[1];
    atr[n_track].cov[1][1]=(it->second).C[2];
    atr[n_track].cov[0][2]=(it->second).C[3];
    atr[n_track].cov[1][2]=(it->second).C[4];
    atr[n_track].cov[2][2]=(it->second).C[5];
    atr[n_track].cov[0][3]=(it->second).C[6];
    atr[n_track].cov[1][3]=(it->second).C[7];
    atr[n_track].cov[2][3]=(it->second).C[8];
    atr[n_track].cov[3][3]=(it->second).C[9];
    atr[n_track].cov[1][0]=atr[n_track].cov[0][1];
    atr[n_track].cov[2][0]=atr[n_track].cov[0][2];
    atr[n_track].cov[3][0]=atr[n_track].cov[0][3];
    atr[n_track].cov[2][1]=atr[n_track].cov[1][2];
    atr[n_track].cov[3][1]=atr[n_track].cov[1][3];
    atr[n_track].cov[3][2]=atr[n_track].cov[2][3];
    atr[n_track].dzeta[0]=(it->second).x;
    atr[n_track].dzeta[1]=(it->second).y;
    atr[n_track].dzeta[2]=(it->second).tx;
    atr[n_track].dzeta[3]=(it->second).ty;
    atr[n_track].zf=Vpar[2];
    atr[n_track].Tpar[0]= atr[n_track].dzeta[2];
    atr[n_track].Tpar[1]= atr[n_track].dzeta[3];
    for(i=0;i<2;i++) for(j=0;j<2;j++)  
      atr[n_track].Tcov[i][j]=atr[n_track].cov[i+2][j+2];
    //      atr[n_track].Tcov[0][0]=1000.0;atr[n_track].Tcov[1][1]=1000.0;
    atr[n_track].flag=1;
    n_track++;
  }
  
  for(int jjj=1;jjj<=MaxIter;jjj++){
    // (re)initialization of the vertex covariance before iteration 
    
    for(i=0;i<3;i++) for(j=0;j<3;j++) Vcov[i][j]=0.0;
    Vcov[0][0]=vertex.cv[0];Vcov[1][1]=vertex.cv[2];Vcov[2][2]=vertex.cv[5];

    // Part one: vertex position estimation 
    
    for(itr=0;itr<n_track;itr++){
      // Construct covariance from vertex and track parts 
      
      for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
      for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
      for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
      
      // Construct state vector from vertex and track parts
      
      for (i=0;i<3;i++) Rk[i]=Vpar[i];
      Rk[3]=atr[itr].Tpar[0];Rk[4]=atr[itr].Tpar[1];
      
      // Linearized measurement matrix 
      
      dz=atr[itr].zf-Rk[2];ax=Rk[3];ay=Rk[4];
      
      for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
      Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
      Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
      
      // Check if track is in the validation region
      
      for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
      
      L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
      L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
      L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0]=L[0][1];L[2][2]=Gk[3][3];
      L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
      L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
      L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
      L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
      L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
      
      for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];
      
      four_dim_inv(L);
      
      ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
      for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
      lambda=0.0;
      for(i=0;i<4;i++) for(j=0;j<4;j++) lambda+=res[i]*res[j]*L[i][j];
      if (lambda>dchi2) {
	//	printf("Track %d out of val. region\n",itr);
	atr[itr].flag=0;
	continue;
      }
      else if (atr[itr].flag){
	// Probability of correct classification

	beta1=1.0-Pfalse/(Pfalse+exp(-lambda));
	
	// Kalman gain
	
	Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
	Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
	Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
	Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
	Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
	Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
	Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
	Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
	Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
	Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
	Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++){
	  Kk[i][j]=0.0;
	  for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	}
	
	// New estimation of the vertex position 
	
	for(i=0;i<5;i++) 
	  for(j=0;j<4;j++) 
	    Rk[i]+=beta1*Kk[i][j]*res[j];
	
	// New covariance matrix
	
	for(i=0;i<5;i++) for(j=0;j<5;j++){
	  KH[i][j]=0.0;
	  for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	}
	
	for(i=0;i<5;i++) for(j=0;j<5;j++){
	  dG[i][j]=0.0;
	  for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	}
	
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=beta1*dG[i][j];
	
	// Extract vertex part
	
	for(i=0;i<3;i++) Vpar[i]=Rk[i];
	for(i=0;i<3;i++) for (j=0;j<3;j++) Vcov[i][j]=Gk[i][j];
      }
    }
    
    // Part two: smoothing track parameters using estimated vertex position
    
    for(itr=0;itr<n_track;itr++) if(atr[itr].flag){
      // Construct covariance from vertex and track parts
      
      for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
      for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
      for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
      
      // Construct state vector from vertex and track parts
      
      for (i=0;i<3;i++) Rk[i]=Vpar[i];
      Rk[3]=atr[itr].Tpar[0];Rk[4]=atr[itr].Tpar[1];
      
      // Linearized measurement matrix
      
      dz=atr[itr].zf-Rk[2];ax=Rk[3];ay=Rk[4];
      
      for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
      Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
      Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
      
      // Check if track is still in the validation region
      
      for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
      
      L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
      L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
      L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0]=L[0][1];L[2][2]=Gk[3][3];
      L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
      L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
      L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
      L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
      L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
      
      for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];
      
      four_dim_inv(L);
      
      ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
      for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
      lambda=0.0;
      for(i=0;i<4;i++) for(j=0;j<4;j++) lambda+=res[i]*res[j]*L[i][j];
      if(lambda>dchi2) {
	//	printf("Track %d out of val. region\n",itr);
	atr[itr].flag=0;
	continue;
      }
      else{
	if(jjj==MaxIter) atr[itr].ChiInc=lambda;
	// Probability of correct classification
	beta1=1.0-Pfalse/(Pfalse+exp(-lambda));
	
	// Kalman gain
	
	Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
	Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
	Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
	Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
	Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
	Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
	Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
	Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
	Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
	Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
	Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++){
	  Kk[i][j]=0.0;
	  for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	}
	
	// New estimation
	
	for(i=0;i<5;i++) for(j=0;j<4;j++) Rk[i]+=beta1*Kk[i][j]*res[j];
	
	// New covariance matrix
	
	for(i=0;i<5;i++) for(j=0;j<5;j++){
	  KH[i][j]=0.0;
	  for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	}
	
	for(i=0;i<5;i++) for(j=0;j<5;j++){
	  dG[i][j]=0.0;
	  for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	}
	
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=beta1*dG[i][j];
	
	// Extract track and vertex parts
	
	atr[itr].Tpar[0]=Rk[3];atr[itr].Tpar[1]=Rk[4];
	//        Vpar[0]=Rk[0];Vpar[1]=Rk[1];Vpar[2]=Rk[2];
	//
	//atr[itr].Tcov[0][0]=Gk[3][3];atr[itr].Tcov[1][1]=Gk[4][4];
	//atr[itr].Tcov[0][1]=Gk[3][4];atr[itr].Tcov[1][0]=Gk[4][3];
	//
      }
    }
    else 
      ;
      //printf("Track was excluded from the fit\n");
  }
  chi2=0.0;nValidTrack=0;
  for(int itr=0; itr<ntrincluster; itr++) 
    if (atr[itr].flag!=0) 
      {
	chi2+=atr[itr].ChiInc;
	nValidTrack++;
      }
  chi2/=(4.0*nValidTrack-3.0-2.0*nValidTrack);
  
  vertex.xv = Vpar[0];
  vertex.yv = Vpar[1];
  vertex.zv = Vpar[2];

  vertex.cv[0] = Vcov[0][0];     
  vertex.cv[1] = Vcov[0][1];
  vertex.cv[2] = Vcov[1][1];
  vertex.cv[3] = Vcov[0][2];     
  vertex.cv[4] = Vcov[1][2];
  vertex.cv[5] = Vcov[2][2];
  vertex.nt = 0;
  int refntrincluster = 0;
  for(int itr=1; itr<=ntrincluster; itr++){
    if (atr[itr-1].flag==0)
      continue;
    vertex.nt++;
    Tracks_mi it = trackincluster[itr];
    (it->second).used = 1;
    if ((it->second).nhitb > 4) // 6
      refntrincluster++;
  }

  if (refntrincluster>=mintracksinvertex){
    if (!RoverHidden){
      ArteTable<RVER>::push_back(vertex);
    }

    vertex_ok = 1;
    //    cout<<"Chi2="<<chi2<<endl;
    nver++;

    for(int itr=1; itr<=ntrincluster; itr++){
      if (atr[itr-1].flag==0)
	continue;
      Tracks_mi it = trackincluster[itr];
      if ((it->second).rtra_rseg){
	  // state relations to RTRA
	if (!RoverHidden){
	  ArtePointer<RVER> rver(ArteTable<RVER>::size());
	  RREL relation;
	  relation.chi  = 0. ;
	  relation.flag = 1  ;
	  ArteTable<RREL>::push_back(relation);
	  
	  ArtePointer<RREL> rrel(ArteTable<RREL>::size());
	  ArtePointer<RTRA> rtra = (it->second).rtra;
	  createRelation(rver, rrel, &RVER::rrel);
	  createRelation(rtra, rrel, &RTRA::rrel);
	  //rtra->ever = rver;
	}
	rtra_to_rver[(it->second).rtra_index] = nver;
      }
      else{
	rseg_to_rver[(it->second).rseg_index] = nver;
      }
    }
  }
  else{
    vertex_ok = 0;
    for(int itr=1; itr<=ntrincluster; itr++){
      Tracks_mi it = trackincluster[itr];
      (it->second).used = 1;// remove ALL tracks of the cluster
    }
  }

  return; 
}


//===================================================================

void Lover::four_dim_inv(float a[4][4]){
  /**** Gaussian algorithm for 4x4 matrix inversion ****/
  
  int i,j,k,l;
  float factor;
  float temp[4];
  float b[4][4];
  
  for(i=0;i<4;i++) for(j=0;j<4;j++) 
    if(i==j) b[i][j]=1.0; else b[i][j]=0.0;
  
  for(i=0;i<4;i++){
    for(j=i+1;j<4;j++) 
      if (fabs(a[i][i])<fabs(a[j][i])){
	for(l=0;l<4;l++) temp[l]=a[i][l];
	for(l=0;l<4;l++) a[i][l]=a[j][l];
	for(l=0;l<4;l++) a[j][l]=temp[l];
	for(l=0;l<4;l++) temp[l]=b[i][l];
	for(l=0;l<4;l++) b[i][l]=b[j][l];
	for(l=0;l<4;l++) b[j][l]=temp[l];
      }
    factor=a[i][i];
    for(j=4-1;j>-1;j--) {
      b[i][j]/=factor;a[i][j]/=factor;
    }
    for(j=i+1;j<4;j++) {
      factor=-a[j][i];
      for(k=0;k<4;k++){
	a[j][k]+=a[i][k]*factor;b[j][k]+=b[i][k]*factor;
      }
    }
  }
  for(i=3;i>0;i--){
    for(j=i-1;j>-1;j--){
      factor=-a[j][i];
      for(k=0;k<4;k++){
	a[j][k]+=a[i][k]*factor;b[j][k]+=b[i][k]*factor;
      }
    }
  }
  for(i=0;i<4;i++) for(j=0;j<4;j++) a[i][j]=b[i][j];
}

//===================================================================

void Lover::three_dim_inv(float a[3][3]){
  /**** Gaussian algorithm for 3x3 matrix inversion ****/
  
  int i,j,k,l;
  float factor;
  float temp[3];
  float b[3][3];
  
  for(i=0;i<3;i++) for(j=0;j<3;j++) 
    if(i==j) b[i][j]=1.0; else b[i][j]=0.0;
  
  for(i=0;i<3;i++){
    for(j=i+1;j<3;j++) 
      if (fabs(a[i][i])<fabs(a[j][i])){
	for(l=0;l<3;l++) temp[l]=a[i][l];
	for(l=0;l<3;l++) a[i][l]=a[j][l];
	for(l=0;l<3;l++) a[j][l]=temp[l];
	for(l=0;l<3;l++) temp[l]=b[i][l];
	for(l=0;l<3;l++) b[i][l]=b[j][l];
	for(l=0;l<3;l++) b[j][l]=temp[l];
      }
    factor=a[i][i];
    for(j=3-1;j>-1;j--) {
      b[i][j]/=factor;a[i][j]/=factor;
    }
    for(j=i+1;j<3;j++) {
      factor=-a[j][i];
      for(k=0;k<3;k++){
	a[j][k]+=a[i][k]*factor;b[j][k]+=b[i][k]*factor;
      }
    }
  }
  for(i=2;i>0;i--){
    for(j=i-1;j>-1;j--){
      factor=-a[j][i];
      for(k=0;k<3;k++){
	a[j][k]+=a[i][k]*factor;b[j][k]+=b[i][k]*factor;
      }
    }
  }
  for(i=0;i<3;i++) for(j=0;j<3;j++) a[i][j]=b[i][j];
}

//============================================================

float Lover::distll(Tracks_t& t1,Tracks_t& t2,float& xm,float& ym,float& zm)
{
 
 //solve equation a*crossproduct(w1,w2) + b*w1 - c*w2 = v2-v1
 //for tracks v1+b*w1,v2+c*w2, the closest distance is given
 //by ||a*crossproduct(w1,w2)||, b and c are the footpoints
 //of the distance vector on the tracks
 
 vector<double> a(9,0.);
 vector<double> b(3,0.);

 a[0] = double(t1.ty-t2.ty);
 a[1] = double(t2.tx-t1.tx);
 a[2] = double(t1.tx*t2.ty-t1.ty*t2.tx);
 a[3] = double(t1.tx);
 a[4] = double(t1.ty);
 a[5] = double(1.0);
 a[6] = double(-t2.tx);
 a[7] = double(-t2.ty);
 a[8] = double(-1.0);


 b[0] = double(t2.x-t1.x-t2.tx*t2.z+t1.tx*t1.z);
 b[1] = double(t2.y-t1.y-t2.ty*t2.z+t1.ty*t1.z);
 b[2] = double(0.0);

 if(lusolve(a,3,b))
   {
    return -1.;
   } 

 float d = fabs(b[0])*sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
 
 xm = (t1.x-t1.z*t1.tx+b[1]*t1.tx + t2.x-t2.z*t2.tx+b[2]*t2.tx)/2.;
 ym = (t1.y-t1.z*t1.ty+b[1]*t1.ty + t2.y-t2.z*t2.ty+b[2]*t2.ty)/2.;
 zm = (b[1]+b[2])/2.;

 return d;
 
}



/***************** Graphics ******************/



/******************************************/
/************ graph_ini() ******************/
/******************************************/



void Lover::graph_ini(void){

  iginit(0);
  igsse(0, 7);


  iscr(1, Black       ,0.0000, 0.0000, 0.0000);
  iscr(1, DarkGray    ,0.5000, 0.5400, 0.5300);
  iscr(1, Blue        ,0.0000, 0.0000, 1.0000);
  iscr(1, Green       ,0.1333, 0.5451, 0.1333);
  iscr(1, Cyan        ,0.0000, 0.8078, 0.8196);
  iscr(1, Red         ,1.0000, 0.0000, 0.0000);
  iscr(1, Magenta     ,1.0000, 0.0000, 1.0000);
  iscr(1, Brown       ,0.5000, 0.1647, 0.1647);
  iscr(1, LightBlue   ,0.6784, 0.8471, 0.9020);
  iscr(1, LightGreen  ,0.5961, 0.9843, 0.5961);
  iscr(1, LightCyan   ,0.4980, 1.0000, 0.8314);
  iscr(1, LightRed    ,0.9804, 0.5020, 0.4471);
  iscr(1, LightMagenta,0.8549, 0.4392, 0.8392);
  iscr(1, Yellow      ,1.0000, 1.0000, 0.0000);
  iscr(1, LightGray   ,0.8275, 0.8275, 0.8275);
  iscr(1, White       ,1.0000, 1.0000, 1.0000);


  float beam_x = 0.0, beam_y = 0.0;  // beam center for simu data
  
  //  if (!ArteTable<MCIN>::booked()){
    beam_x =  0.30;                  // beam center for real data
    beam_y = -0.05;
    //  } // the sam for real and simu data


  iswn(2, beam_x-0.5, beam_x+0.5, beam_y-0.5, beam_y+0.5);
  isvp(2,0.0,1.0,0.0,1.0);
  iselnt(2);

  isfais(1);

  isfaci(LightGray);
  isplci(Red);

  igbox(beam_x-0.5, beam_x+0.5, beam_y-0.5, beam_y+0.5);

  iuwk(1,1);

}
/**************************************************/
/************** graph_close() *********************/
/**************************************************/

void Lover::graph_close(void){
  igend();
}

void Lover::draw_event(void){
  
  float xpl[100],ypl[100],xpm[10],ypm[10];

  isfaci(White);
  isplci(Red);

  float beam_x = 0.0, beam_y = 0.0;  // beam center for simu data
  
  //if (!ArteTable<MCIN>::booked()){
    beam_x =  0.30;                  // beam center for real data
    beam_y = -0.05;
    //}
  igbox(beam_x-0.5, beam_x+0.5, beam_y-0.5, beam_y+0.5);
       

  
  for(Target_mi tarit = targets.begin(); tarit != targets.end(); ++tarit){
    
    float xt   = (tarit->second).x;
    float yt   = (tarit->second).y;
    float zt   = (tarit->second).z;
    float dxt  = (tarit->second).dx;
    float dyt  = (tarit->second).dy;

    isplci(Yellow);
    if ((tarit->second).flaghorver==0){
      xpl[0] = xt-0.2;
      ypl[0] = yt;
      xpl[1] = xt+0.2;
      ypl[1] = yt;
    }
    else{
      xpl[0] = xt;
      ypl[0] = yt-0.2;
      xpl[1] = xt;
      ypl[1] = yt+0.2;
    }
    ipl(2,xpl,ypl);
    
    if (zt>-2.5)
      isplci(Black);
    else
      isplci(DarkGray);

    xpl[0] = xt-dxt;
    ypl[0] = yt-dyt;
    xpl[1] = xt+dxt;
    ypl[1] = yt-dyt;
    xpl[2] = xt+dxt;
    ypl[2] = yt+dyt;
    xpl[3] = xt-dxt;
    ypl[3] = yt+dyt;
    xpl[4] = xt-dxt;
    ypl[4] = yt-dyt;
    ipl(5,xpl,ypl);
  }


  ismk(3);
  ismksc(0.5);
  ispmci(Cyan);

  for (ArteTable<MTRA>::iterator mtraIt = ArteTable<MTRA>::begin();
       mtraIt != ArteTable<MTRA>::end(); ++mtraIt){
    if ((*mtraIt)->topp->chrg == 0 )
      continue;

    int nhitb = 0;
    for ( ArteRelationIterator<MIMP> mimpIt = (*mtraIt)->mimp.begin();
	  mimpIt != (*mtraIt)->mimp.end(); ++mimpIt){
      if ((*mimpIt)->gede->cmp != 1) continue;
      nhitb += distance((*mimpIt)->hitb.begin(),(*mimpIt)->hitb.end()); 
    }
    if (nhitb<5)
      continue;
    
    float px = (*mtraIt)->px / (*mtraIt)->pz;
    float py = (*mtraIt)->py / (*mtraIt)->pz;
    float xm = (*mtraIt)->bver->x;
    float ym = (*mtraIt)->bver->y;
    float zm = (*mtraIt)->bver->z;

    float zt = zsta;

    xpm[0]  = xm+(zt-zm)*px;
    ypm[0]  = ym+(zt-zm)*py;
    ipm(1,xpm,ypm);
  }

  if (!MCdata){

    ismk(3);
    ismksc(0.5);
    ispmci(LightBlue);
    isplci(LightBlue);
    
    for(Tracks_mi it = tracks.begin();it != tracks.end();++it){
      trackpropagate(it->second,zsta);
      xpm[0] = (it->second).x;
      ypm[0] = (it->second).y;
      ipm(1,xpm,ypm);
      
      float xm = (it->second).x;
      float ym = (it->second).y;
      float dx = sqrt((it->second).C[0]);
      float dy = sqrt((it->second).C[2]);
      
      igarc(xm, ym, 0.5*(dx+dy),0.5*(dx+dy), 0.0, 0.0);
      
    }
  }

  ismk(28);
  ismksc(5.0);

  for( ArteTable<MVER>::iterator mverIt = ArteTable<MVER>::begin();
       mverIt != ArteTable<MVER>::end();++mverIt){
    if( (*mverIt)->flag != 1) continue;

    int mtrafound = 0;
    for (ArteTable<MTRA>::iterator mtraIt = ArteTable<MTRA>::begin();
	 mtraIt != ArteTable<MTRA>::end(); ++mtraIt){
      if ((*mtraIt)->bver != (*mverIt)) continue;
      int nhitb = 0;
      for ( ArteRelationIterator<MIMP> mimpIt = (*mtraIt)->mimp.begin();
	    mimpIt != (*mtraIt)->mimp.end(); ++mimpIt){
	if ((*mimpIt)->gede->cmp != 1) continue;
	nhitb += distance((*mimpIt)->hitb.begin(),(*mimpIt)->hitb.end()); 
      }
      if (nhitb>=5){
	mtrafound++;
      }
    }
    if (mtrafound<mintracksinvertex) continue;
    
    if ((*mverIt)->z>-2.5)
      ispmci(Black);
    else
      ispmci(DarkGray);

    xpm[0]=(*mverIt)->x;
    ypm[0]=(*mverIt)->y;

    ipm(1,xpm,ypm);
  }

  iuwk(1,1);
}

