
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "arte/Arte.h"
#include "auto/MEMO.h"
#include "auto/MEMA.h"
#include "auto/MEMS.h"
#include "auto/MEIO.h"
#include "dbarte/DbArteBridge.h"
#include "arte/ArteFlag.h"
#include "arte/mecCount.h"

void DbCompLedaRelTab(int *buffer, int tab, int *keys, int rno, int cno);
void DbDecompLedaRelTab(int *buffer, int tab, int *keys, int rno, int cno, int *);
void DbWRelTab(int lun, int *buffer, int Nrows, DbTableDesc *TD, 
		int major, int minor, int *ierr);
void DbRRelTab(int lun, int **Buffer, int *nRows, DbTableDesc *TD, 
		int major, int minor, char *opt, int *ierr);
void chhtot_(char *, char *, int *);
DbTableDesc *DbLedaRelTabDesc(char *tName1, char *tName2, 
										int **pos1, int **pos2);

void mmirel_(int *);
void mmbrel_(int *, int *, int *);


void DbStoreLedaRel(int lun, int reltab, int minor, int major, int *ierr)
{
  /*************************************************************************
   *
   * Author: A. Semenov (JINR, Dubna)
   * Created 07.06.99: 
   *
   * Description: Create LEDA relation table and store to MIZZI database.
   *
   * Called from: anywhere.
   *
   * External references: DbLedaRelTabDesc, DbCompLedaRelTab,
   *                      DbWRelTab.
   *
   * Argument in/out description
   *  lun     in     LUN
   *  reltab  in     ARTE relation table number
   *  major   in     major revision number
   *  minor   in     minor revision number
   *  ierr    out    operation status
   *              0  OK
   *              1  table has no keys
   *              2  Bank name too long for MIZZI
   *              3  LUN is not connected to data file/server
   *              4  Can not book table
   *              5  Table already exists (at interactive mode, user 
   *					refused overwriting)
   *              6  Can not allocate memory for buffer
   *              7  field type not implemented 
   *              8  lun is opened read-only
   *
   * FUNCTION DESCRIPTION
   *
   ************************************************************************/
   
   char RName[DB_MAX_NAME_SIZE+1], TName1[DB_MAX_NAME_SIZE+1],
    	TName2[DB_MAX_NAME_SIZE+1];
   int tab1, tab2;
   int pos1, pos2;
   int *key1, *key2;
   int var, i, j;
   int *Link, *rLink;
   int cno1, cno2, rno;
   int *buffer;
   DbTableDesc *TD;
   int perm=0;
   
   var = DB_MAX_NAME_SIZE/4;
   chhtot_(MEMO(reltab).name, RName, &var );
   
   perm=strncmp(MEMO(MEMO(reltab).sup[0]).name,
		   MEMO(MEMO(reltab).sup[1]).name,DB_MAX_NAME_SIZE);
		   
   if(perm<0)
   {
	   tab1=MEMO(reltab).sup[0];
	   pos1=MEMO(tab1).mema+MEMO(reltab).pos[0]-1;
	   chhtot_(MEMO(tab1).name, TName1, &var );
	   
/*	   chhtot_(MEMA(pos1).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[0], RName,
			 MEMA(pos1).type, 
        	vvv(MEMA(pos1).type),
			  MEMO(reltab).pos[0]>MEMO(tab1).ocol?"no output":"is output" );*/

	   tab2=MEMO(reltab).sup[1];
	   pos2=MEMO(tab2).mema+MEMO(reltab).pos[1]-1;
	   chhtot_(MEMO(tab2).name, TName2, &var );
	   
/*	   chhtot_(MEMA(pos2).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[1], RName,
			 MEMA(pos2).type, 
        	vvv(MEMA(pos2).type),
			  MEMO(reltab).pos[1]>MEMO(tab2).ocol?"no output":"is output"  );*/
	   perm=0;
	}
	else
	{
		
	   tab1=MEMO(reltab).sup[1];
	   pos1=MEMO(tab1).mema+MEMO(reltab).pos[1]-1;
	   chhtot_(MEMO(tab1).name, TName1, &var );

/*	   chhtot_(MEMA(pos1).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[0], RName,
			 MEMA(pos1).type, 
        	vvv(MEMA(pos1).type),
			  MEMO(reltab).pos[1]>MEMO(tab1).ocol?"no output":"is output" );*/

	   tab2=MEMO(reltab).sup[0];
	   pos2=MEMO(tab2).mema+MEMO(reltab).pos[0]-1;
	   chhtot_(MEMO(tab2).name, TName2, &var );
/*	   
	   chhtot_(MEMA(pos2).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[1], RName,
			 MEMA(pos2).type, 
        	vvv(MEMA(pos2).type),
			  MEMO(reltab).pos[0]>MEMO(tab2).ocol?"no output":"is output"  );*/
	   perm=1;
	}
   
   TD=DbLedaRelTabDesc(TName1,TName2,&key1,&key2);
   
   if(TD==NULL)
   {
	   *ierr=7;
	   return;
   }
   
   var = DB_MAX_NAME_SIZE/4;
   chhtot_(MEMO(reltab).name, TD->Name, &var );
   sprintf(TD->Comm,"Relations %s <-> %s",TName1,TName2);
   
   {
	   static int errPtr;
	   static char mess[200];
       sprintf(mess,"DbStoreLedaRel: writing table %s (relations %s <-> %s)",
		 	  RName,TName1,TName2);
	   mecWarningCount(&errPtr, mess);
   }
   
   if(key1==NULL)
   {
	   static int errPtr;
	   static char mess[200];
	   sprintf(mess,"DbStoreLedaRel: Table %s has no defined keys\n",TName1);
	   mecErrorCount(&errPtr, mess);
	   *ierr=1;
	   return;
   }
   
   if(key2==NULL)
   {
	   static int errPtr;
	   static char mess[200];
	   sprintf(mess,"DbStoreLedaRel: Table %s has no defined keys\n",TName2);
	   mecErrorCount(&errPtr, mess);
	   *ierr=1;
	   return;
   }
   
   for(cno1=0;key1[cno1];cno1++);
   for(cno2=0;key2[cno2];cno2++);
   
    if(MEMA(pos1).type==4 && MEMA(pos2).type==4)
    {
	
	   var=MEMO(reltab).sup[1];
	   Link = gcbank_.iq + *((int *)((artelink_.larln1 + var) * 4));
	   rLink= gcbank_.iq + *((int *)((artelink_.larln1 + reltab) * 4));
	      
	   rno=0;
	   
	   for(i=0;i<N_U(Link);i++)
	   {
		   int ref;
		   ref=Link[2+MEMO(var).col*i+(MEMO(reltab).pos[1]-1)];
		   while(ref){
			   int rr;
			   rr=rLink[2+4*(ref-1)];
			   /*printf("%d -> %d\n",i+1,rr);*/
			   ref=rLink[2+4*(ref-1)+1];
			   rno++;}
	   }
	   
	   
	   if(rno==0)
		   return;

	   buffer=malloc(rno*(cno1+cno2)*sizeof(int));
	   if(buffer==NULL)
	   {
		   *ierr=6;
		   return;
	   }
	   
	   j=0;
	   for(i=0;i<N_U(Link);i++)
	   {
		   int ref;
		   ref=Link[2+MEMO(var).col*i+(MEMO(reltab).pos[1]-1)];
		   while(ref)
		   {
			   int rr;
			   rr=rLink[2+4*(ref-1)];
			   buffer[j*(cno1+cno2)]=perm?i+1:rr;
			   buffer[j*(cno1+cno2)+cno1]=perm?rr:i+1;
			   ref=rLink[2+4*(ref-1)+1];
			   j++;
		   }
	   }
		
	}
	else
	{   
   
	   var=MEMO(reltab).sup[1];
	   Link = gcbank_.iq + *((int *)((artelink_.larln1 + var) * 4));

	   rno=0;

	   for(i=0;i<N_U(Link);i++)
	   {
		   int ref;
		   ref=Link[2+MEMO(var).col*i+(MEMO(reltab).pos[1]-1)];
		   if(ref){
			 /*  printf("%d %s %d\n",i+1,perm?"<-":"->",ref);*/
			   rno++;}
	   }
	   
	   if(rno==0)
		   return;
	   
	   buffer=malloc(rno*(cno1+cno2)*sizeof(int));
	   if(buffer==NULL)
	   {
		   *ierr=6;
		   return;
	   }
	   
	   j=0;
	   for(i=0;i<N_U(Link);i++)
	   {
		   int ref;
		   ref=Link[2+MEMO(var).col*i+(MEMO(reltab).pos[1]-1)];
		   if(ref)
		   {
			   buffer[j*(cno1+cno2)]=perm?i+1:ref;
			   buffer[j*(cno1+cno2)+cno1]=perm?ref:i+1;
			   j++;
		   }
	   }
	   
   }
  
   DbCompLedaRelTab(buffer     ,tab1,key1,rno,cno1+cno2);

   DbCompLedaRelTab(buffer+cno1,tab2,key2,rno,cno1+cno2);

   *ierr=0;
   
   DbWRelTab(lun, buffer, rno, TD, major, minor, ierr);

   
}

void DbLoadLedaRel(int lun, int reltab, int minor, int major, char *opt, 
		int *ierr)
{
  /*************************************************************************
   *
   * Author: A. Semenov (JINR, Dubna)
   * Created 09.06.99: 
   *
   * Description: Read LEDA relation table from MIZZI database.
   *
   * Called from: anywhere.
   *
   * External references: DbLedaRelTabDesc, DbRRelTab,
   *                      DbDecompLedaRelTab.
   *
   * Argument in/out description
   *  lun     in     LUN
   *  reltab  in     ARTE relation table number
   *  major   in     major revision number
   *  minor   in     minor revision number
   *  ierr    out    operation status
   *              0  OK
   *              1  table has no keys
   *              2  'opt' has incorrect value
   *              3  bank name too long for MIZZI
   *              4  LUN is not connected to data file/server
   *              5  bank does not exist in file/server
   *              6  Can not find requested major/minor table
   *              7  Can not allocate memory for buffer
   *              8  Field type is not implemented
   *              9  Database table format differs from Arte table
   *             10  Key tuple not found in table
   *
   * FUNCTION DESCRIPTION
   *
   ************************************************************************/
   
   char RName[DB_MAX_NAME_SIZE+1], TName1[DB_MAX_NAME_SIZE+1],
    	TName2[DB_MAX_NAME_SIZE+1];
   int tab1, tab2;
   int pos1, pos2;
   int *key1, *key2;
   int var, i;
   int *Link, *Mlink;
   int cno1, cno2, rno;
   int *buffer;
   DbTableDesc *TD;
   int perm=0;
   
   var = DB_MAX_NAME_SIZE/4;
   chhtot_(MEMO(reltab).name, RName, &var );
   
   perm=strncmp(MEMO(MEMO(reltab).sup[0]).name,
		   MEMO(MEMO(reltab).sup[1]).name,DB_MAX_NAME_SIZE);
		   
   if(perm<0)
   {
	   tab1=MEMO(reltab).sup[0];
	   pos1=MEMO(tab1).mema+MEMO(reltab).pos[0]-1;
	   chhtot_(MEMO(tab1).name, TName1, &var );
	   
/*	   chhtot_(MEMA(pos1).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[0], RName,
			 MEMA(pos1).type, 
        	vvv(MEMA(pos1).type),
			  MEMO(reltab).pos[0]>MEMO(tab1).ocol?"no output":"is output" );*/

	   tab2=MEMO(reltab).sup[1];
	   pos2=MEMO(tab2).mema+MEMO(reltab).pos[1]-1;
	   chhtot_(MEMO(tab2).name, TName2, &var );
/*
	   chhtot_(MEMA(pos2).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[1], RName,
			 MEMA(pos2).type, 
        	vvv(MEMA(pos2).type),
			  MEMO(reltab).pos[1]>MEMO(tab2).ocol?"no output":"is output"  );*/
	   
	   perm=0;
	}
	else
	{
		
	   tab1=MEMO(reltab).sup[1];
	   pos1=MEMO(tab1).mema+MEMO(reltab).pos[1]-1;
	   chhtot_(MEMO(tab1).name, TName1, &var );
	   
/*	   chhtot_(MEMA(pos1).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[0], RName,
			 MEMA(pos1).type, 
        	vvv(MEMA(pos1).type),
			  MEMO(reltab).pos[1]>MEMO(tab1).ocol?"no output":"is output" );*/

	   tab2=MEMO(reltab).sup[0];
	   pos2=MEMO(tab2).mema+MEMO(reltab).pos[0]-1;
	   chhtot_(MEMO(tab2).name, TName2, &var );

/*	   chhtot_(MEMA(pos2).name, RName, &var );
	   printf("field %d '%s', type %d (%s), %s\n", MEMO(reltab).pos[1], RName,
			 MEMA(pos2).type, 
        	vvv(MEMA(pos2).type),
			  MEMO(reltab).pos[0]>MEMO(tab2).ocol?"no output":"is output"  );*/
	   perm=1;
	}
   
   TD=DbLedaRelTabDesc(TName1,TName2,&key1,&key2);
   
   if(TD==NULL)
   {
	   *ierr=8;
	   return;
   }
   
   var = DB_MAX_NAME_SIZE/4;
   chhtot_(MEMO(reltab).name, TD->Name, &var );
   sprintf(TD->Comm,"Relations %s <-> %s",TName1,TName2);

   {
	   static int errPtr;
	   static char mess[200];
       sprintf(mess,"DbLoadLedaRel: reading table %s (relations %s <-> %s)",
		 	  RName,TName1,TName2);
	   mecWarningCount(&errPtr, mess);
   }


   if(key1==NULL)
   {
	   static int errPtr;
	   static char mess[200];
	   sprintf(mess,"DbLoadLedaRel: Table %s has no defined keys\n",TName1);
	   mecErrorCount(&errPtr, mess);
	   *ierr=1;
	   return;
   }
   
   if(key2==NULL)
   {
	   static int errPtr;
	   static char mess[200];
	   sprintf(mess,"DbLoadLedaRel: Table %s has no defined keys\n",TName2);
	   mecErrorCount(&errPtr, mess);
	   *ierr=1;
	   return;
   }
       
   for(cno1=0;key1[cno1];cno1++);
   for(cno2=0;key2[cno2];cno2++);
   
   *ierr=0;
   
   DbRRelTab(lun, &buffer, &rno, TD, major, minor, opt, ierr);
   
   
   if(*ierr)
	   return;
   if(rno==0)
	   return;
      
   DbDecompLedaRelTab(buffer     ,tab1,key1,rno,cno1+cno2, ierr);

   DbDecompLedaRelTab(buffer+cno1,tab2,key2,rno,cno1+cno2, ierr);

   if(*ierr)
	   return;

   
    if(MEMA(pos1).type==4 && MEMA(pos2).type==4)
    {
		   
	   for(i=0;i<rno;i++)
	   {
		   int s1,s2;
		   s1=buffer[i*(cno1+cno2)+(perm?cno1:0)];
		   s2=buffer[i*(cno1+cno2)+(perm?0:cno1)];
		   /*printf(" %d -> %d\n",s1,s2);*/
		   mmbrel_((int *)((artelink_.larln1 + reltab) * 4), &s1, &s2);
	   }

		
	}
	else
	{   
   
	   int t1,t2;
	   
	   t1=N_U(gcbank_.iq + 
			   *((int *)((artelink_.larln1 + MEMO(reltab).sup[0]) * 4)));
	   
	   var=MEMO(reltab).sup[1];
	   Link = gcbank_.iq + *((int *)((artelink_.larln1 + var) * 4));
	   t2=N_U(Link);

	   
	   for(i=0;i<rno;i++)
	   {
		   int s1, s2;
		   s1=buffer[i*(cno1+cno2)];
		   s2=buffer[i*(cno1+cno2)+cno1];

		   if(perm)
		   {
				Link[2+MEMO(var).col*(s1-1)+(MEMO(reltab).pos[1]-1)]=s2;
		   }
		   else
		   {
				Link[2+MEMO(var).col*(s1-1)+(MEMO(reltab).pos[1]-1)]=s2;
		   }
		   
	   }
	   
   }
  
   
    if(MEMA(pos1).type!=4 && MEMA(pos2).type!=4)
	{
	Mlink=(int *)((artelink_.larln1 + reltab) * 4);
    mmirel_(Mlink);
    }
}

