
// ******************************************************************
//
// source:    DQ_FARM_hit_init.C
//
// type:      source code
//
// created:   1999
//
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// called by: DQ_init()
//
// changes:
// 07 Sep 1999 (AG) add processing time histo
// 09 Sep 1999 (AG) polishing
// 04 Oct 1999 (AG) polishing
// 07 Oct 1999 (AG) decrease range for processing time
// 21 Oct 1999 (AG) polishing
// 27 Oct 1999 (AG) use defines from DQ.h
// 29 Oct 1999 (AG) redefine histogram IDs
//  1 Nov 1999 (AG) get rid of dq_hid.h
// 22 Feb 2000 (AG) change scale of time
// 05 Apr 2000 (AG) change booking of times
// 12 May 2000 (AG) change booking of processing time from 10 to 20
// 15 May 2000 (AG) corrected binning
// 17 May 2000 (AG) sizes of some raw tables
// 22 May 2000 (AG) change scale of 4LT_node histogram
// 24 May 2000 (AG) histogram bytes
// 31 May 2000 (AG) remove RECO time here (now in DQ_FARM_rec)
// 31 May 2000 (AG) change scale for DSVD, DITR, DOTR
//
// ******************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ONLINE includes
#include "rhp.h"

// Arte includes
#include "arte/arte.hh"
#include "dq/DQ.h"
#include "dq/DQ_FARM.h"

//
// function code
//
int
DQ_FARM_hit_init()
{
  // show up
  cout << "DQ_FARM_hit_init()" << endl;
  
  // book histograms
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 1 ), "FARM: Trigger code", 16, -0.5, 15.5 );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 2 ), "FARM: Target Rate [MHz]", 50, 0, 50. );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 3 ), "FARM: SLP_node", 240, -0.5, 239.5 );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 4 ), "FARM: 4LT_node", 110, -0.5, 109.5 );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 6 ), "FARM: Raw event size [kB]", 40, 0.0, 200 );
  
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 15 ), "FARM: DQ Time [sec]", 50, 0.0, 1.0 );

  // size of some raw tables
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 101 ), "FARM: DSVD [kB]", 40, 0,  20 );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 102 ), "FARM: DITR [kB]", 50, 0,  50 );
  rhp_hbook1( DQ_ID( COMP_FARM, LVL_HIT, 104 ), "FARM: DOTR [kB]", 50, 0, 100 );

  return 0;
}
