 
// *****************************************************************************
//
// source:    DQ_FARM_rec_evnt.C
//
// type:      source code
//
// created:   17 Nov 1999
//
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// called by: DQ_event()
//
// changes:
// 17 Nov 1999 (AG) creation
// 13 Mar 2000 (AG) monitor RTRA, RSEG, RCCL, RVER
// 08 May 2000 (AG) event classification flag
// 12 May 2000 (AG) histogram 0 if no class bit set
// 18 May 2000 (AG) timings of reco modules
// 22 May 2000 (AG) comment 2-d histograms; polish printout
// 24 May 2000 (AG) histogram bytes
// 26 May 2000 (AG) histogram time_rcevnt here
// 31 May 2000 (AG) change scale for Rnnn
//  5 Jun 2000 (AG) add timing routine
//  6 Jun 2000 (AG) determine OTR and RICH reco rejection
//  7 Jun 2000 (AG) use new timing table EVTI
//  8 Jun 2000 (AG) some polishing of printout
// 16 Jun 2000 (AG) improve mean calculation
//
// ******************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ONLINE includes
#include "rhp.h"
#include "grr.h"

// Arte includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "dq/DQ.h"
#include "dq/DQ_FARM.h"

// prototypes
void DQ_FARM_rec_timing();
void DQ_FARM_rec_rejection();

// externals
extern char *LOGFILE; // logfile
extern int  FATIME;   // flag for printout of timing

//
// function code
//
int
DQ_FARM_rec_evnt()
{
  // DST flag
  ArtePointer<EVHD> evhd( 1 );
  int flagdst = evhd->flagdst;

  // l4t event classification
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 1 ), 0, 1 );
  for( int bit=0; bit<32; bit++ ){
    bool j = ( flagdst >> bit ) & 1;
    if( j ) rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 1 ), bit + 1, 1 );
  }

  // reco tables
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 10 ), ArteTable<RTRA>::size(), 1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 11 ), ArteTable<RSEG>::size(), 1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 12 ), ArteTable<RCCL>::size(), 1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 13 ), ArteTable<RVER>::size(), 1 );

  // EVRC
  if( ArteTable<EVRC>::size() == 0 ) return 0;
  ArtePointer<EVRC> evrc(1);

  // EVTI
  if( ArteTable<EVTI>::size() == 0 ) return 0;
  ArtePointer<EVTI> evti(1);

  // will be removed soon
  rhp_hf1( DQ_ID( COMP_FARM, LVL_HIT, 5 ), evrc->cputime, 1 );

  // reco standalone modules
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 100 ), evrc->cputime, 1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 101 ), evti->vds,     1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 104 ), evti->ranger,  1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 105 ), evti->rich,    1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 106 ), evti->trd,     1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 107 ), evti->ecal,    1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 108 ), evti->muon,    1 );

  // reco modules
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 112 ), evti->marple, 1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 114 ), evti->refit,  1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 116 ), evti->pid,    1 );
  rhp_hf1( DQ_ID( COMP_FARM, LVL_REC, 117 ), evti->vertex, 1 );

  // monitor timing
  DQ_FARM_rec_timing();

  // monitor rejection
  DQ_FARM_rec_rejection();
  
  return 0;
}

//
// statistics
//
#include <math.h>

// update mean
inline float mean( float mean, float x, int n ){
  return ( float( n - 1 ) / n ) * mean + fabsf( x ) / n;
}

// update variance 
//inline float vari( float vari, float mean, float x, int n ){
//  return ( float( n - 1) / n ) * vari + ( fabsf( x ) - mean ) * ( fabsf( x ) - mean ) / n;
//}

//
// function DQ_FARM_rec_timing()
//
void
DQ_FARM_rec_timing()
{
  // EVRC
  if( ArteTable<EVRC>::size() == 0 ) return;
  ArtePointer<EVRC> evrc(1);

  // EVTI
  if( ArteTable<EVTI>::size() == 0 ) return;
  ArtePointer<EVTI> evti(1);

  // initialization
  static int n = 0;
  
  static float mean_cputime = 0;
  static float mean_vds     = 0;
  static float mean_ranger  = 0;
  static float mean_rich    = 0;
  static float mean_marple  = 0;
  static float mean_refit   = 0;
  static float mean_pid     = 0;
  static float mean_vertex  = 0;

  /*
  static float vari_cputime = 0;
  static float vari_vds     = 0;
  static float vari_ranger  = 0;
  static float vari_rich    = 0;
  static float vari_marple  = 0;
  static float vari_refit   = 0;
  static float vari_pid     = 0;
  static float vari_vertex  = 0;
  */
  // counter
  n++;

  // mean
  mean_cputime = mean( mean_cputime, evrc->cputime, n );
  mean_vds     = mean( mean_vds,     evti->vds,     n );
  mean_ranger  = mean( mean_ranger,  evti->ranger,  n );
  mean_rich    = mean( mean_rich,    evti->rich,    n );
  mean_marple  = mean( mean_marple,  evti->marple,  n );
  mean_refit   = mean( mean_refit,   evti->refit,   n );
  mean_pid     = mean( mean_pid,     evti->pid,     n );
  mean_vertex  = mean( mean_vertex,  evti->vertex,  n );
  
  // variance
  //vari_cputime = vari( vari_cputime, mean_cputime, evrc->cputime, n );
  //vari_vds     = vari( vari_vds,     mean_vds,     evti->vds,     n );
  //vari_ranger  = vari( vari_ranger,  mean_ranger,  evti->ranger,  n );
  //vari_rich    = vari( vari_rich,    mean_rich,    evti->rich,    n );
  //vari_marple  = vari( vari_marple,  mean_marple,  evti->marple,  n );
  //vari_refit   = vari( vari_refit,   mean_refit,   evti->refit,   n );
  //vari_pid     = vari( vari_pid,     mean_pid,     evti->pid,     n );
  //vari_vertex  = vari( vari_vertex,  mean_vertex,  evti->vertex,  n );

  // text
  char time_txt[300];
  char mean_txt[300];
  //char vari_txt[300];
  memset( time_txt, '\0', 300 );
  memset( mean_txt, '\0', 300 );
  //memset( vari_txt, '\0', 300 );
  sprintf( time_txt,
	   "DQ_FARM_rec: [sec] RECO=%3.1f VDS=%3.1f Ranger=%3.1f RICH=%3.1f Marple=%3.1f Refit=%3.1f Pid=%3.1f Vertex=%3.1f",
	   evrc->cputime,
	   evti->vds,
	   evti->ranger,
	   evti->rich, 
	   evti->marple,
	   evti->refit,
	   evti->pid,
	   evti->vertex
	   );
  sprintf( mean_txt,
	   "DQ_FARM_rec: [sec] RECO=%3.1f VDS=%3.1f Ranger=%3.1f RICH=%3.1f Marple=%3.1f Refit=%3.1f Pid=%3.1f Vertex=%3.1f",
	   mean_cputime,
	   mean_vds,
	   mean_ranger,
	   mean_rich, 
	   mean_marple,
	   mean_refit,
	   mean_pid,
	   mean_vertex
	   );
  /*
  sprintf( vari_txt,
	   "DQ_FARM_rec: [sec] RECO=%3.1f VDS=%3.1f Ranger=%3.1f RICH=%3.1f Marple=%3.1f Refit=%3.1f Pid=%3.1f Vertex=%3.1f",
	   sqrt( vari_cputime ),
	   sqrt( vari_vds ),
	   sqrt( vari_ranger ),
	   sqrt( vari_rich ), 
	   sqrt( vari_marple ),
	   sqrt( vari_refit ),
	   sqrt( vari_pid ),
	   sqrt( vari_vertex )
	   );
	   */
  // printout
  if( FATIME ){
    cout << " " << time_txt << endl;
    cout << " " << mean_txt << endl;
    //cout << " " << vari_txt << endl;
  }

  // erwin & log
  if( ( n == 50 ) || !( n%500 ) ){
    ARTE_error( GRR_SINFO, mean_txt, 0 );
    ARTE_log( LOGFILE, mean_txt );
  }
}

//
// DQ_FARM_rec_rejection()
//
#include <unistd.h>

void
DQ_FARM_rec_rejection()
{
  // counter
  static int n = 0;
  n++;

  //
  // OTR
  //

  // OTR cutoff
  int  cutOTR   = 9999999;
  int  cutOTR10 = 10000;
  int  cutOTR12 = 12000;
  int  cutOTR15 = 15000;
  int  cutOTR20 = 20000;
  char *ccutOTR = getenv( "OtrOccupancyCut" );
  if( ccutOTR ) cutOTR = atoi( ccutOTR );
  
  // OTR hits
  int hitsOTR = 0;
  for( ArteTableIterator<HITB> hitb; hitb.valid(); hitb.next() ){ 
    ArtePointer<GEDE> gede( hitb->gede ); 
    if( gede->cmp == 2 ) hitsOTR++;
  }

  // OTR count
  static int rejOTR   = 0;
  static int rejOTR10 = 0;
  static int rejOTR12 = 0;
  static int rejOTR15 = 0;
  static int rejOTR20 = 0;
  if( hitsOTR > cutOTR )   rejOTR++;
  if( hitsOTR > cutOTR10 ) rejOTR10++;
  if( hitsOTR > cutOTR12 ) rejOTR12++;
  if( hitsOTR > cutOTR15 ) rejOTR15++;
  if( hitsOTR > cutOTR20 ) rejOTR20++;

  // OTR print
  char otr_txt[200];
  memset( otr_txt, '\0', 200 );
  sprintf( otr_txt,
	   "DQ_FARM_rec: OTR  rejection [%%] cut=%6d rej=%3.1f 10000=%3.1f 12000=%3.1f 15000=%3.1f 20000=%3.1f",
	   cutOTR,
	   float( rejOTR )   / n * 100,
	   float( rejOTR10 ) / n * 100,
	   float( rejOTR12 ) / n * 100,
	   float( rejOTR15 ) / n * 100,
	   float( rejOTR20 ) / n * 100 );

  //
  // RICH
  //

  // RICH cutoff
  int  cutRICH   = 5000;
  int  cutRICH10 = 1000;
  int  cutRICH15 = 1500;
  int  cutRICH20 = 2000;
  int  cutRICH30 = 3000;

  // RICH hits
  int hitsRICH = 0;
  for( ArteTableIterator<HITB> hitb; hitb.valid(); hitb.next() ){ 
    ArtePointer<GEDE> gede( hitb->gede ); 
    if( gede->cmp == 5 ) hitsRICH++;
  }
  //cout << "DQ_FARM_rec: hitsRICH=" << hitsRICH << endl;

  // RICH count
  static int rejRICH   = 0;
  static int rejRICH10 = 0;
  static int rejRICH15 = 0;
  static int rejRICH20 = 0;
  static int rejRICH30 = 0;
  if( hitsRICH > cutRICH )   rejRICH++;
  if( hitsRICH > cutRICH10 ) rejRICH10++;
  if( hitsRICH > cutRICH15 ) rejRICH15++;
  if( hitsRICH > cutRICH20 ) rejRICH20++;
  if( hitsRICH > cutRICH30 ) rejRICH30++;

  // RICH print
  char rich_txt[200];
  memset( rich_txt, '\0', 200 );
  sprintf( rich_txt,
	   "DQ_FARM_rec: RICH rejection [%%] cut=%6d rej=%3.1f 1000=%3.1f 1500=%3.1f 2000=%3.1f 3000=%3.1f",
	   cutRICH,
	   float( rejRICH )   / n * 100,
	   float( rejRICH10 ) / n * 100,
	   float( rejRICH15 ) / n * 100,
	   float( rejRICH20 ) / n * 100,
	   float( rejRICH30 ) / n * 100 );

  //
  // printout
  //
  if( FATIME ){
    cout << " DQ_FARM_rec: OTR  hits=" << hitsOTR  << endl;
    //cout << " DQ_FARM_rec: RICH hits=" << hitsRICH << endl;
    cout << " " << otr_txt  << endl;
    //cout << " " << rich_txt << endl;
  }

  //
  // erwin & logfile
  //
  if( ( n == 50 ) || !( n%500 ) ){
    ARTE_error( GRR_SINFO, otr_txt,  0 );
    //ARTE_error( GRR_SINFO, rich_txt, 0 );
    ARTE_log( LOGFILE, otr_txt );
    //ARTE_log( LOGFILE, rich_txt );
  }
}
