
/******************************************************************* 
 *
 * D. Samtleben 10 Aug 1999
 * A. Gellrich
 *
 * Filling of histograms
 *
 * hints:
 *        components: COMP_SVD
 *                    COMP_ITR
 *                    COMP_HIPT
 *                    COMP_OTR
 *                    COMP_RICH
 *                    COMP_TRD
 *                    COMP_ECAL
 *                    COMP_MUON
 *                    COMP_FLT
 *                    COMP_SLT
 *                    COMP_TLT
 *                    COMP_FARM
 *
 *         levels:    LVL_RAW
 *                    LVL_HIT
 *                    LVL_REC
 *                         
 * changes:                
 * 22 Oct 1999 (AG) code polishing
 * 27 Oct 1999 (AG) use defines from DQ.h
 *  1 Nov 1999 (AG) get rid of dq_hid.h
 *                                       
 *******************************************************************/

/* standard includes */
#include <stdlib.h>
#include <stdio.h>

/* ONLINE includes */
#include "rhp.h"
#include "event.h"
#include "etb.h"
#include "grr.h"

/* Arte includes */
#include "dq/DQ.h"
#include "dq/DQ_MUON.h"

#define IN_CRATES 36

int ARTE_get_feds_DMUT( int max, ETB_fed *feds, int *n );
int ARTE_get_feds_DMUP( int max, ETB_fed *feds, int *n );
int ARTE_get_feds_DMUX( int max, ETB_fed *feds, int *n );
int MUON_init();
int MUON_bad_fed();

/*====== structure of MUON daq information (Mu.fedconfig) */
struct daqinfo {
  int fed;                               /* fed name MuDecode.c           */
  int pos;                               /* position in readout */
  int fedid;                             /* fed id              */
  int compid;                            /* component id        */
  char sharc[10];                        /* sharc port          */
  char cluster[4];                       /* cluster card        */
} crate[IN_CRATES], daqpos[IN_CRATES];
int end_daq_table;        /* length of muon daq array */

/*====== data pointers */
ETB_fed tfeds[IN_CRATES], pfeds[IN_CRATES], xfeds[IN_CRATES];
int n_tfeds, n_pfeds, n_xfeds;

/*====== first event information */
int tfedids1[IN_CRATES], pfedids1[IN_CRATES], xfedids1[IN_CRATES];
int n_tfeds1, n_pfeds1, n_xfeds1, n_offset1;
int n_readfeds;


/*               */
/* function code */
/*               */
int DQ_MUON_raw_evnt( EVENT_HDR *event ) {

  static int mu_first = 1;
  static unsigned int badfed_id = 0;
  int trc, prc, xrc;
  int bxid1 = 0, goodcrate1;
  int i_fed;
  /* float h_badfed[5]; */

  /*  Initialize the MUON readout */
  if (mu_first) {
    MUON_init();
    badfed_id = DQ_ID(COMP_MUON,LVL_RAW,1);
    mu_first = 0;
  }

  /*============
   * Get the MUON data
   *============*/
  ARTE_get_feds_DMUT(IN_CRATES,tfeds,&n_tfeds);
  ARTE_get_feds_DMUP(IN_CRATES,pfeds,&n_pfeds);
  ARTE_get_feds_DMUX(IN_CRATES,xfeds,&n_xfeds);
  
  /*===========
   * Read the muon data
   *===========*/

  rhp_hf1(badfed_id,4.,1.);   /* Number of events counter */

  goodcrate1 = 1;

  /* tubes */
  for (i_fed = 0; i_fed < n_tfeds; i_fed++) {
    if (goodcrate1) bxid1 = tfeds[i_fed].hdr.bxid;

    trc = MUON_bad_fed(tfeds[i_fed],daqpos[i_fed].fedid,bxid1);
    if (goodcrate1 && !trc) goodcrate1 = 0;

    if (trc&1) rhp_hf1(badfed_id,0.,1.);
    if (trc&2) rhp_hf1(badfed_id,1.,1.);
    if (trc&4) rhp_hf1(badfed_id,2.,1.);
    if (trc&8) rhp_hf1(badfed_id,3.,1.);
  }

  /* pads */
  for (i_fed = 0; i_fed < n_pfeds; i_fed++) {
    if (goodcrate1) bxid1 = pfeds[i_fed].hdr.bxid;

    prc = MUON_bad_fed(pfeds[i_fed],daqpos[i_fed+n_tfeds1].fedid,bxid1);
    if (goodcrate1 && !prc) goodcrate1 = 0;

    if (prc&1) rhp_hf1(badfed_id,0.,1.);
    if (prc&2) rhp_hf1(badfed_id,1.,1.);
    if (prc&4) rhp_hf1(badfed_id,2.,1.);
    if (prc&8) rhp_hf1(badfed_id,3.,1.);
  }

  for (i_fed = 0; i_fed < n_xfeds; i_fed++){
    if (goodcrate1) bxid1 = xfeds[i_fed].hdr.bxid;

    xrc = MUON_bad_fed(xfeds[i_fed],daqpos[i_fed+n_offset1].fedid,bxid1);
    if (goodcrate1 && !xrc) goodcrate1 = 0;

    if (xrc&1) rhp_hf1(badfed_id,0.,1.);
    if (xrc&2) rhp_hf1(badfed_id,1.,1.);
    if (xrc&4) rhp_hf1(badfed_id,2.,1.);
    if (xrc&8) rhp_hf1(badfed_id,3.,1.);
  }

  /*
    rhp_hunpak(badfed_id,h_badfed);
    printf("Fed: %8.0f  Rec: %8.0f  BX: %8.0f  Len: %8.0f  Events: %8.0f\n",
    h_badfed[0],h_badfed[1],h_badfed[2],h_badfed[3],h_badfed[4]);
  */

  return 0; /* return error code (0 = okay) */
}

/*===============================================
 *      S U B R O U T I N E S
 *===============================================*/

int MUON_init() {

  int i, i_fed;
  int infed, innum, infedid, incompid; 
  char line[300];
  FILE *fp, *fopen();

  fp = fopen("/afs/desy.de/group/hera-b/MUON/db/Mu.fedconfig","r");
    
  if (fp == NULL) {
    printf("\nCan't find the file Mu.fedconfig "
	   "in /afs/desy.de/group/hera-b/MUON/db.\n");
    return 0;
  }

  i=0;
  while (fgets(line, sizeof(line), fp) != NULL) {
    if (sscanf(line, "FED %d POS %d FEDID %x COMPID %x "
	       "SHARC %s CLUSTER %s\n",
               &infed,&innum,&infedid,&incompid,
	       crate[i].sharc,crate[i].cluster) == 6) {
      crate[i].fed=infed; crate[i].pos=innum;
      crate[i].fedid=infedid; crate[i].compid=incompid;
      i++;
    }
  }
  end_daq_table = i;
  fclose(fp);

  for (i_fed = 0; i_fed < end_daq_table; i_fed++) {
    daqpos[crate[i_fed].pos] = crate[i_fed];
  }
  n_readfeds = end_daq_table;

  for (i_fed = 0; i_fed < n_readfeds; i_fed++) {
    switch(daqpos[i_fed].compid) {
    case 0x800: tfedids1[n_tfeds1++] = daqpos[i_fed].fedid; break;
    case 0xa00: pfedids1[n_pfeds1++] = daqpos[i_fed].fedid; break;
    case 0x900: xfedids1[n_xfeds1++] = daqpos[i_fed].fedid; break;
    default: break;
    }
  }
  n_offset1 = n_tfeds1 + n_pfeds1;
 
  return 0;
}

int MUON_bad_fed(ETB_fed Fed, int fedid1, int bxid1) {
 
  int is_bad=0;

  if (Fed.hdr.fedid != fedid1)  is_bad|=1;
  if (Fed.hdr.recid>>12 != 0x1) is_bad|=2;
  if (Fed.hdr.bxid != bxid1)    is_bad|=4;
  if (Fed.hdr.length != 136)    is_bad|=8;

  return is_bad;
}
