
// ******************************************************************
//
// D. Samtleben 10 Aug 1999
//
// Filling of histograms
//
// changes:
// 27 Oct 1999 (AG) use defines from DQ.h
//  1 Nov 1999 (AG) get rid of dq_hid.h
//  4 Nov 1999 (DD) beta=1 radius histogram 
//  5 Nov 1999 (AG) some polishing
// 16 Dec 1999 (RP) likelihood histograms DQ_RICH_rec_pid()
// 30 Mar 2000 (DD) fading histos  
// 11 May 2000 (RP) histograms redefined
// 25 May 2000 (DD) event efficiency monitoring, r2p2 particle id
// 16 Jun 2000 (AG) two new hostograms added sigma(tx/ty) vs momen
// ******************************************************************

// standard includes
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// ONLINE includes
#include "rhp.h"
#include "event.h"
#include "grr.h"

// ARTE includes
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#include "dq/DQ.h"
#include "dq/DQ_RICH.h"
#include "farm/farm.hh"
#include "rich/rich.h"
#include "rise/RingSearchARTE.h"

// externals
extern int FAGRR;

extern float DQ_RICH_rec_xmin;
extern float DQ_RICH_rec_xmax;
extern int   DQ_RICH_rec_nbin;

extern float DQ_RICH_rec_thc_xmin;
extern float DQ_RICH_rec_thc_xmax;
extern int   DQ_RICH_rec_thc_nbin;

// prototypes
int DQ_RICH_rec_beta1( ArtePointer<RSEG> rich );
int DQ_RICH_r2p2();
int DQ_RICH_rec_pid();
int DQ_RICH_rec_pid_rpnt();

//////////////// Fill RICH DQ histograms /////////////////////////////
int DQ_RICH_rec_evnt() {
  DQ_RICH_rec_pid_rpnt();
  DQ_RICH_rec_pid();
  int nrRICHrings=0; 
  for(ArteTable<RSEG>::iterator irseg=ArteTable<RSEG>::begin();
      irseg!=ArteTable<RSEG>::end(); ++irseg) {
    if ((*irseg)->cmp!=Rsegc::rich) continue;
    ArtePointer<RSEG> rich=(*irseg);
    nrRICHrings++;

    float tx=rich->txf;
    float ty=rich->tyf;
    rhp_hfill(DQ_ID(COMP_RICH, LVL_REC, 201),tx,ty,1);
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 202),rich->time,1);
    rhp_mult(DQ_ID(COMP_RICH, LVL_REC,2002),.9999);
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC,2002),rich->time,1);
    if (!DQ_RICH_rec_beta1(rich)) {
      rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 203),rich->time,1);
      rhp_mult(DQ_ID(COMP_RICH, LVL_REC,2003),.9999);
      rhp_hf1(DQ_ID(COMP_RICH, LVL_REC,2003),rich->time,1);
    }
  }
  rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 200),nrRICHrings,1);

  // 3=too low photon yield
  // 2=too high photon yield
  // 1=no rings but normal photon occupancy
  // 0=rings ok
  if (ArtePointer<EVHI>(5)->hit<25)
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2004),3.,1.);
  else if (ArtePointer<EVHI>(5)->hit>ArtePointer<RITH>(1)->par[3])
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2004),2.,1.);
  else if (nrRICHrings==0) 
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2004),1.,1.);
  else
    rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2004),0.,1.);

  DQ_RICH_r2p2();

  return 0; // return error code (0 = okay)
}

//////////////  Select RICH Beta=1 w/ ECAL ////////////////////////
int DQ_RICH_rec_beta1(ArtePointer<RSEG> rich) {
  static float max_dty=3e-3;  // angle match
  static float min_esum=8;    // minimum cluster energy
  static float max_p=200;     // maximum p
  static float min_p=10;      // minimum p
  for (ArteTable<RCCL>::iterator irccl=ArteTable<RCCL>::begin();
       irccl!=ArteTable<RCCL>::end(); ++irccl) {
    if ((*irccl)->esum<min_esum) continue;
    if (fabs((*irccl)->ty - rich->tyf)>max_dty) continue;
    float p=288/fabs((*irccl)->x - rich->txf*(*irccl)->z);
    if (p<min_p) continue;
    if (p>max_p) continue;
    return 0; // cluster found
  }
  return 1; // cluster not found
}

int DQ_RICH_r2p2() {

  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  for(ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
      irtra!=ArteTable<RTRA>::end(); ++irtra) {
    if ((*irtra)->fit!=0) continue;
    if (fabs((*irtra)->pf)>.15) continue;
    int richf=0, otrf=0, vdsf=0;
    ArtePointer<RSEG> rich;
    ArtePointer<RSEG> otr;
    ArtePointer<RSEG> vds;
    for (ArteRelationIterator<RSEG> irseg=(*irtra)->rseg.begin();
	 irseg!=(*irtra)->rseg.end(); ++irseg) {
      if ((*irseg)->cmp==(Rsegc::vxd|Rsegc::bit16))      {vdsf=1;  vds=(*irseg);}
      if ((*irseg)->cmp==Rsegc::rich)                    {richf=1; rich=(*irseg);}
      if ((*irseg)->cmp&Rsegc::patt && (*irseg)->fit!=7) {otrf=1;  otr=(*irseg); }
    }
    if (!richf || !otrf || !vdsf) continue;
    if (rich->ye>0) if (rich->ze/sqrt(rich->ye)<5) continue;

    float rr=rich->time*rich->time;
    float ip=(*irtra)->pf;
    float  p=1/fabs(ip);
    rhp_hfill(DQ_ID(COMP_RICH, LVL_REC, 2005),ip*ip,rr,1);
    RingSearchTrack::ip=ip;
    RICH_Search->makeR2P2lines();
    if (p<8)  rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2006),rr-RICH_Search->GetR2P2rad2(0),1);
    if (p<30) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2007),rr-RICH_Search->GetR2P2rad2(2),1);
    if (p<35) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2008),rr-RICH_Search->GetR2P2rad2(3),1);
    if (p<50) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2009),rr-RICH_Search->GetR2P2rad2(4),1);
    
    if (p<8 && fabs(rr-RICH_Search->GetR2P2rad2(0))<1.0e-4) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2010),0,1);
    if (p<30&& fabs(rr-RICH_Search->GetR2P2rad2(2))<1.0e-4) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2010),1,1);
    if (p<35&& fabs(rr-RICH_Search->GetR2P2rad2(3))<1.5e-4) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2010),2,1);
    if (p<50&& fabs(rr-RICH_Search->GetR2P2rad2(4))<2.0e-4) rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 2010),3,1);    

  }
  return 0;
}


int DQ_RICH_rec_pid(){ 
  float logp=0;

  static int   nbin=DQ_RICH_rec_nbin;
  static float xmin=DQ_RICH_rec_xmin;
  static float xmax=DQ_RICH_rec_xmax; 
  static float dx=xmax-xmin;

  int flr[nbin], tra[nbin];  
  for (int i=0;i<nbin;i++) {
    flr[i]=0;
    tra[i]=0;
  }
  int nflr=0;
  int ntra=0;
  for(ArteTable<RTRA>::iterator itra=ArteTable<RTRA>::begin();
      itra!=ArteTable<RTRA>::end(); ++itra) {
    ArtePointer<RTRA> rtra(itra); 
    int ok=1, bin=0;
    if (rtra->flr) { 
      nflr++;
      if ( fabs(rtra->pf) > 0.001)      logp=fabs(1/rtra->pf); else ok=0;
      if (logp>0) logp=log(logp); else ok=0;
      bin = int((logp-xmin)/dx);
      if ((bin>=nbin)||(bin<0)) ok=0;

      float lrother=1-
	( rtra->lre + rtra->lrmu + rtra->lrpi + rtra->lrk + rtra->lrp );

      float which = 1;
      float max = rtra->lre;
      if (rtra->lre>max)  { max= rtra->lre;  which=1;}
      if (rtra->lrmu>max) { max= rtra->lrmu; which=2;}
      if (rtra->lrpi>max) { max= rtra->lrpi; which=3;}
      if (rtra->lrk>max)  { max= rtra->lrk;  which=4;}
      if (rtra->lrp>max)  { max= rtra->lrp;  which=5;}
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 209),which+0.5,1);

      if (lrother  >max)  { max= lrother;    which=0;}   
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 210),which+0.5,1);
      rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 208),logp ,which+0.5, 1);

      
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 211),rtra->lre,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 212),rtra->lrmu,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 213),rtra->lrpi,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 214),rtra->lrk,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 215),rtra->lrp,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 216),lrother,1);
      rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 217),rtra->hri,1);

      if (ok) {
	flr[bin]++;
        float lrpi = 0;
        float lrk  = 0;
        if (rtra->lrpi!=0) lrpi  = log(rtra->lrpi);  
        if (rtra->lrk !=0)  lrk  = log(rtra->lrk);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 220),logp,lrk-lrpi,1);

	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 221),logp,rtra->lre,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 222),logp,rtra->lrmu,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 223),logp,rtra->lrpi,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 224),logp,rtra->lrk,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 225),logp,rtra->lrp,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 226),logp,lrother,1);
	rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 227),logp,rtra->hri,1);
	
      }
    }
    ntra++;
    if (ok) tra[bin]++;
  }

  rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 218), nflr+0.5 ,1);
  if (ntra) rhp_hf1(DQ_ID(COMP_RICH,LVL_REC, 219),     float(nflr)/ntra,1);
  for (int i=0;i<nbin;i++) {
    float x = xmin+dx/nbin*(i+0.5);
    rhp_hfill(  DQ_ID(COMP_RICH,LVL_REC, 228),x,flr[i],1);
    if (tra[i]) rhp_hfill(DQ_ID(COMP_RICH,LVL_REC, 229),x,float(flr[i])/tra[i],1);
  }
  return 0; 
} 

int DQ_RICH_rec_pid_rpnt(){
  
  const float zmin= 1001;  // middle of the rich
  for(ArteTable<RTRA>::iterator rtra1it=ArteTable<RTRA>::begin();
      rtra1it!=ArteTable<RTRA>::end();++rtra1it) {
    if ( (*rtra1it)->fit || !(*rtra1it)->flr  ) continue;      

    ArtePointer<RPNT> rpnt;
    
    for(ArteRelationIterator<RPNT> rpntit=(*rtra1it)->rpnt.begin();
	rpntit!=(*rtra1it)->rpnt.end();rpntit++) {
      if( ((*rpntit)->zf > rpnt->zf) && ((*rpntit)->zf < zmin) )  rpnt=(*rpntit);
    }
    
    if( rpnt->zf>500) {
      rhp_hf1( DQ_ID(COMP_RICH,LVL_REC, 230), sqrt(rpnt->cvf[5]),1);
      rhp_hf1( DQ_ID(COMP_RICH,LVL_REC, 231), sqrt(rpnt->cvf[9]),1);
      int ok = 1;
      float logp =0 ;
      if ( fabs(rpnt->pf) > 0.001)      logp=fabs(1/rpnt->pf); else ok=0;
      if (logp>0) logp=log(logp); else ok=0;
      rhp_hfill( DQ_ID(COMP_RICH,LVL_REC, 233), logp, sqrt(rpnt->cvf[5]),1);
      rhp_hfill( DQ_ID(COMP_RICH,LVL_REC, 234), logp, sqrt(rpnt->cvf[9]),1);
    }
  }
  return 0;
};




