
// *****************************************************************************
//
// source:    l4tEvnt.C
//
// type:      source code
//
// created:   14 Jan 1998
//
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    l4tEvnt
//
// purpose:   finally select events (L4 trigger)
//
// called by: anywhere after rcevnt_()
//
// hints:     classification is done in rcClassify()
//
// changes:
// 25 Nov 1999 (AG) move to l4tEvnt()
// 22 Feb 2000 (AG) handle various settings
// 13 Mar 2000 (AG) remove some printout
// 04 Apr 2000 (AG) return trigger result
// 27 Apr 2000 (AG) use ARTE_mark()
// 05 May 2000 (AG) set EDIR flag accoring to flagdst
// 08 May 2000 (AG) only flagdst needs to be set
// 16 May 2000 (AG) set flag = flagdst
// 24 May 2000 (AG) set classification bit for random triggers (8 + 9)
// 25 May 2000 (AG) set classification bit for FLT trigger (0 + 8)
// 29 May 2000 (AG) comment some printout
// 14 Jun 2000 (AG) polishing
// 16 Jun 2000 (AG) polishing
//
// *****************************************************************************

// standard includes
#include <iostream.h>

// Arte includes
#include "arte/arte.hh"
#include "arte/ArteUtil.h"
#include "farm/farm.hh"
#include "reco/eclass.hh"
#include "l4trig/l4t.hh"

// externals
extern int FATRG; // L4 trigger flag
extern int FADEB; // debug flag

//
// program code
//
int
l4tEvnt()
{
  // mark
  ARTE_mark( "l4tEvnt 0", 0 );

  // EVHD table and trigger code
  ArtePointer<EVHD> evhd( 1 );
  int tcode = faDDAQFCS_tcode();
  cout << " TRIGGER: tcode=" << tcode << endl;

  //===========================================================================
  //     set flags for random and FLT triggers
  //===========================================================================

  if( ( tcode == 8 ) ||
      ( tcode == 9 )    ) evhd->flagdst |= ECLASS_RNDTRG; // bit-wise OR

  if( ( tcode == 0 ) ||
      ( tcode == 8 )    ) evhd->flagdst |= ECLASS_FLTTRG; // bit-wise OR
  
  //===========================================================================
  //     set flag
  //===========================================================================

  evhd->flag = evhd->flagdst;  // NEEDED!
  if( FADEB ){
    cout << " TRIGGER: flagdst=0x" << hex << evhd->flagdst << dec << endl;
  }

  //===========================================================================
  //     L4 trigger stuff
  //===========================================================================

  // counters
  static int nAll = 0;
  static int nDST = 0;
  static int nOUT = 0;

  // output flag
  //cout << " TRIGGER: FATRG=" << FATRG << endl;
  arteutil_.eventout = 0;
  switch( FATRG ){

    // OFF
  case L4_OFF:
    arteutil_.eventout = 1;
    break;

    // ON
  case L4_ON:
    if( evhd->flagdst ) arteutil_.eventout = 1;
    break;

    // KEEP all
  case L4_KEEP:
    arteutil_.eventout = 1;
    break;

    // REJECT all
  case L4_REJECT:
    arteutil_.eventout = 0;
    break;

    // default
  default:
    arteutil_.eventout = 1;
  }

  // print
  cout << " TRIGGER: "
       << "TRG="            << FATRG                     << " "
       << "CLASS=0x" << hex << evhd->flagdst      << dec << " "
       << "OUT="            << arteutil_.eventout        << endl;

  // statistics
  nAll++;
  if( evhd->flagdst )      nDST++;
  if( arteutil_.eventout ) nOUT++;
  cout << " TRIGGER: Summary: "
       << "EVT="   << nAll << " "
       << "CLASS=" << nDST << " "
       << "OUT="   << nOUT << endl;

  // mark
  ARTE_mark( "l4tEvnt 1", 0 );

  // return
  return arteutil_.eventout; // return flag
}
