//##
//      /**Author:        Boris Fominykh
//
//      Date:             January 5, 2000
//
//      This code is for digi into hitb transformation
//      of muon detector hits, DataBase version.
//      history:
//      bug correction 25.05.2000 by Jing-Ge */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include "etb.h"

#include "arte/arte.hh"
#include "arte/ArteMessage.hh"

#include "murec/MuDbTubeArte.h"
#include "murec/MuDbPadArte.h"
#include "murec/MuDbPixelArte.h"

  static MuDbTubeArte  Tube;
  static MuDbPadArte   Pad;
  static MuDbPixelArte Pix;
  static MuPixChamber  PixCh;

extern "C" {
 int MuDigi2Hitb();
 int  MuD2Hini(int,int,char*);
 void MuPHini2(); 
 void TubeLoops2();
 void PadLoops2();
 void PixLoops2();
 void LoopsPW2(Word,Word,Word);
 int CheckBit2(Word,Word,LWord);
 void AddTube2(int,int);
 void AddPad2();
 void AddPix2(Word,Word,Word,Word);
}

extern "C" 
{

// main routine of digi into hitb transformation
int MuDigi2Hitb()
{
// no action for MC
  if(ArteTable<MTRA>::size() > 0) return 0;

// check that the objects are initialized
  if(Tube.Ivalid != 1) {
    arteMsg << "MuDigi2Hitb: tubes are not initialized" << endl;
    arteMsg.warn();
    return -1;
  }
  if(Pad.Ivalid != 1) {
    arteMsg << "MuDigi2Hitb: pads are not initialized" << endl;
    arteMsg.warn();
    return -2;
  }
  if(Pix.Ivalid != 1) {
    arteMsg << "MuDigi2Hitb: pixels are not initialized" << endl;
    arteMsg.warn();
    return -3;
  }
// set the initial pointers to hits
  MuPHini2();
// make digi into hits transformation
  TubeLoops2();
  PadLoops2();
  PixLoops2();
  return 0;
}

// initialization of the objects
int MuD2Hini(int KTstatus,int run,char* DBname) 
{
// no action for MC
  if(ArteTable<MTRA>::size() > 0) return 0;

  int ret=0;
// initialize the tubes
  Tube.Ivalid = -1;
  Tube.setDBname(DBname);
  Tube.setKTstatus(KTstatus);
  Tube.setRun(run);
  Tube.setMinor(0);
  ret=Tube.Ini();
  if(ret != 0) {
    arteMsg << "MuD2Hini: error of tube initialization, error= " << ret << endl;
    arteMsg.warn();
    return -1;
  }
//  cout << "MuD2Hini: Tubes are initialized" << endl;
//  Tube.Print();

// initialize the pads
  Pad.Ivalid = -1;
  Pad.setDBname(DBname);
  Pad.setKTstatus(KTstatus);
  Pad.setRun(run);
  Pad.setMinor(0);
  ret=Pad.Ini();
  if(ret != 0) {
    arteMsg << "MuD2Hini: error of pad initialization, error= " << ret << endl;
    arteMsg.warn();
    return -2;
  }
//  cout << "MuD2Hini: Pads are initialized" << endl;
  //Pad.Print();

// initialize the pads
  Pix.Ivalid = -1;
  Pix.setDBname(DBname);
  Pix.setKTstatus(KTstatus);
  Pix.setRun(run);
  Pix.setMinor(0);
  ret=Pix.Ini();
  if(ret != 0) {
    arteMsg << "MuD2Hini: error of pixel initialization, error= " << ret << endl;
    arteMsg.warn();
    return -3;
  }
//  cout << "MuDigi2Hitb: Pixels are initialized" << endl;
//  Pix.Print();

  if(KTstatus > 0) return 1;
  else             return 2;
}

// initialization of the pointers to hits 
void MuPHini2() 
{
// loops over GESL
  for ( ArteTableIterator<GESL> isl = ArteTable<GESL>::begin(); 
        isl != ArteTable<GESL>::end();
        ++isl ) {
    int   cmp  = (*isl)->cmp;
    if(cmp == 8 || cmp == 9) {
      int iv1 = (*isl)->fpl;
      int iv2 = iv1 + (*isl)->npl - 1;
      if(iv2 >= iv1) {
// loops over muon gede
        for ( int iv=iv1; iv<iv2; iv++) { 
          ArtePointer<GEDE> gede(iv);
          gede->hit1 = 0;
          gede->hitl = -1;
        }
      }
    }
  }
}

//*** loops over tube chambers
void TubeLoops2() 
{
  for(Word sl=0; sl<Tube.TubeGeoSl; sl++) {
    for(Word lay=0; lay<Tube.TubeGeoLay; lay++) {
      for(Word ud=0; ud<Tube.TubeGeoUpDown; ud++) {
        if(Tube.NchambT[sl][lay][ud] > 0) {
          for(Word sd=0; sd<Tube.TubeGeoSd; sd++) {
            for(Word nx=0; nx<Tube.NchambT[sl][lay][ud]; nx++) {
              Word ncon = Tube.NconnT[sl][lay][ud][nx];
              if(ncon > 0) {
                for(Word nc=0; nc<ncon; nc++) {
                  Word crate = Tube.CrateT[sl][lay][ud][nx];
                  Word nbit1 = Tube.Nbit1T[sl][lay][ud][nx][nc][sd];
                  Word wir1  = Tube.Wire1T[sl][lay][ud][nx][nc][sd];
                  int  step  = Tube.StepT[sl][lay][ud][nx];
                  int astp   = (step<0)? -step : step;
                  int gede = int(Tube.TubeGede[sl][lay][ud][nx][sd]);
                  for(Word wr=0; wr<Tube.NumbT; wr++) {
                    int wir=0;
                    if(step > 0) wir = int(wir1+wr);
                    if(step < 0) wir = int(wir1-wr);
                    LWord nbit = nbit1 + astp*wr;
// swapping
                    LWord nwd=nbit/32;
                    LWord nbt = nbit - 32*nwd;
                    nbit = 32*nwd + 31-nbt;
// check mask
                    Word mskb = astp*wr;
                    if(Tube.Nbit1T[sl][lay][ud][nx][nc][sd] > 
                       Tube.Nbit1T[sl][lay][ud][nx][nc][1-sd]) mskb += 1;
                    Word msk = 0;
                    Word one = 0x8000;
                    msk = msk + (one >> mskb);
                    Word mask = Tube.MaskT[sl][lay][ud][nx][nc] & msk;
                    if(mask != 0) {
// check DMUT
                      int check = CheckBit2(1,crate,nbit);
// add the tube wire into HITB
                      if(check == 1) AddTube2(gede,wir);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}

// loops over pad chambers
void PadLoops2()
{
  for(Word sl=0; sl<Pad.PadGeoSl; sl++) {
    for(Word udt=0; udt<Pad.PadGeoUpDown; udt++) {
      Word ud = Pad.PadGeoUpDown-1-udt;

// loops over wire connectors of side 0
      LoopsPW2(sl,ud,0);

// loops over pad connectors
      if(Pad.NchambP[sl][ud] > 0) {
        for(Word nx=0; nx<Pad.NchambP[sl][ud]; nx++) {
          Word ncon = Pad.NconPP[sl][ud][nx];
          if(ncon > 0) {
            Word crate= Pad.CratePP[sl][ud][nx];
            int gede1 = int(Pad.PadGede[sl][ud][nx][0]);
            int gede2 = int(Pad.PadGede[sl][ud][nx][1]);
            for(Word nc=0; nc<ncon; nc++) {
              Word nbit1 = Pad.Nbit1PP[sl][ud][nx][nc];
              Word pad1  = Pad.Pad1[sl][ud][nx][nc];
              int  npd   = Pad.NumbP[sl][ud][nx][nc];
              if(npd > 0) {
                for(Word pd=0; pd<npd; pd++) {
                  int pad = int(pad1+npd-pd-1);
                  LWord nbit = nbit1+pd;
// check mask
                  Word one = 0x1;
                  Word msk = 0;
                  if(npd >= 14 || npd < 12) msk = one << pd+15-npd;
//                  if(npd == 12) msk = one << pd;
                  if ((npd == 12)||(npd == 13)) msk = one << pd;
                  Word mask = Pad.MaskPP[sl][ud][nx][nc] & msk;
                  if(mask != 0) {
// check pad readout channel in DMUT
                    int check = CheckBit2(2,crate,nbit);
                    if(check == 1) {
// add two connected pads into HITB
                        AddPad2(gede1,pad);
                        AddPad2(gede2,pad);
                    }
                  }
                }
              }
            }
          }
        }
      }

// loops over wire connectors of side 1
      LoopsPW2(sl,ud,1);

    }
  }
}


// loops over pad chamber wire connectors of side sd
void LoopsPW2(Word sl,Word ud,Word sd) 
{
  if(Pad.NchambP[sl][ud] > 0) {
    for(Word nx=0; nx<Pad.NchambP[sl][ud]; nx++) {
      Word ncon = Pad.NconnP[sl][ud][nx];
      if(ncon > 0) {
        Word crate = Pad.CrateP[sl][ud][nx];
        int gede   = int(Pad.PadGede[sl][ud][nx][sd]);
        int  step  = Pad.StepP[sl][ud][nx];
        Word astp  = (step<0)? -step : step;
        for(Word nc=0; nc<ncon; nc++) {
          Word nbit1 = Pad.Nbit1P[sl][ud][nx][nc][sd];
          Word wir1  = Pad.Wire1P[sl][ud][nx][nc][sd];
          for(Word wr=0; wr<Pad.NumbT; wr++) {
            int wir=0;
            if(step > 0) wir = int(wir1+wr);
            if(step < 0) wir = int(wir1-wr);
            LWord nbit = nbit1 + astp*wr;
// swap the bytes
            LWord nwd = nbit/32;
            LWord nbt = nbit - 32*nwd;
            nbit      = 32*nwd + 31-nbt;
// check mask
            Word mskb = astp*wr;
            if(Pad.Nbit1P[sl][ud][nx][nc][sd] > 
               Pad.Nbit1P[sl][ud][nx][nc][1-sd]) mskb += 1;
            Word msk = 0;
            Word one = 0x8000;
            msk = msk + (one >> mskb);
            Word mask = Pad.MaskP[sl][ud][nx][nc] & msk;
            if(mask != 0) {
// check DMUT
              int check = CheckBit2(1,crate,nbit);
// add the pad wire into HITB
              if(check == 1) AddTube2(gede,wir);
            }
          }
        }
      }
    }
  }
}

//*** loops over pixel chambers
void PixLoops2() 
{
  Word one = 0x1;
  for(Word sl=0; sl<Pix.PixGeoSl; sl++) {
    for(Word qu=0; qu<Pix.PixGeoQu; qu++) {
//      int gede = int(Pix.PixGede[sl][qu]);
      for(Word nc=0; nc<Pix.PixNcon; nc++) {
        if(Pix.MaskX[sl][qu][nc] > 0) {
          Word crate = Pix.CrateX[sl][qu][nc];
          LWord  nbit1 = LWord(abs(Pix.Nbit1X[sl][qu][nc]));
          int step=1;
          if(Pix.Nbit1X[sl][qu][nc] < 0) step=-1;
          for(Word bt=0; bt<16; bt++) {
            LWord nbit = 0;
            if(step == 1)  nbit = nbit1 + bt;
            if(step == -1) nbit = nbit1 - bt;
// check mask
            Word msk = one << bt;
            Word mask = Pix.MaskX[sl][qu][nc] & msk;
            if(mask != 0) {
// check DMUX
              int check = CheckBit2(3,crate,nbit);
              if(check == 1) AddPix2(sl,qu,nc,15-bt);
            }
          }
        }
      }
    }
  }
}

// check the bit nbit in DMUT(table=1), DMUP(table=2), DMUX(table=3)
int CheckBit2(Word table,Word crate,LWord nbit) 
{
  int* len   = 0;
  if(table == 1) len=(int*)(ArtePointer<DMUT>(0)->bits);
  if(table == 2) len=(int*)(ArtePointer<DMUP>(0)->bits);
  if(table == 3) len=(int*)(ArtePointer<DMUX>(0)->bits);
  if(*len <= 0) return 0;
  LWord nmax = LWord(*len);

// crate offset
  LWord nword = 0;
  LWord ncrate = nmax/34;
  LWord* Lbuf = 0;
  LWord word = 0;
  LWord first=0;
// loops over crate headers
  for (LWord ncr=0; ncr<ncrate; ncr++) {
    first=2+34*ncr;
    if(table == 1) Lbuf=(LWord*)(ArtePointer<DMUT>(first)->bits);
    if(table == 2) Lbuf=(LWord*)(ArtePointer<DMUP>(first)->bits);
    if(table == 3) Lbuf=(LWord*)(ArtePointer<DMUX>(first)->bits);
// not change the bit order for Big endian   
    if(Pad.getEndian() == 0) word = *Lbuf;
// change the bit order for Little endian
    else word = ESP_SWAP_INT(*Lbuf);
    LWord tmp1 = word << 16;
    Word  tmp2=Word(tmp1 >> 16);
    if(tmp2 == crate) {
      nword=first+nbit/32+1;
      break;
    }
  }
  if(nword == 0) {
    // arteMsg << "CheckBit: no crate " << hex << crate << " in table " << table << endl;
    // arteMsg.warn();
    return 0;
  }
  if(nword > nmax) {
    // arteMsg << "CheckBit: nword= " << nword << " nmax= " << nmax << " table " << table  << endl;
    // arteMsg.warn();
    return 0;
  } 

// take the word
  if(table == 1) Lbuf=(LWord*)(ArtePointer<DMUT>(nword+1)->bits);
  if(table == 2) Lbuf=(LWord*)(ArtePointer<DMUP>(nword+1)->bits);
  if(table == 3) Lbuf=(LWord*)(ArtePointer<DMUX>(nword+1)->bits);
// not change the bit order for Big endian   
  if(Pad.getEndian() == 0) word = *Lbuf;
// change the bit order for Little endian
  else word = ESP_SWAP_INT(*Lbuf);
// check the bit
  if(word == 0) return 0;
  Word nw = Word(nbit/32);
  Word nbtt = nbit-32*nw;
  Word nbyte = nbtt/8;
  Word nbb = nbtt-8*nbyte;
  Word nb = 7-nbb;  // change the order of bits inside byte
  Word nbt = 8*nbyte+nb;
  static const LWord one = 0x1;
  LWord mask = one << nbt;
  LWord res = word & mask;
//  cout << "crate " << hex << crate << " nword " << dec << nword << " nbit " << nbit;
//  cout << " word " << hex << *Lbuf << "  " << word << " res " << res << endl;
  if(res == 0) return 0;
  return 1;
}

// add tube wire to HITB
void AddTube2(int iv, int iw)
{
  ArtePointer<GEDE> gede(iv);
  float dist = gede->pit;    // distance between wires
  float xfw  = gede->xfw[0]; // X-position of first wire
  float xd1  = gede->xmin;   // Xmin of chamber
  float x = xd1+xfw+dist*float(iw);  // x-position of wire
  float z = gede->zfw[0];    // z-position of wire

// add the hit to HITB
  struct HITB hitb;
  hitb.x    = x;
  hitb.y    = 0;
  hitb.z    = z;
  hitb.q    = 0;
  hitb.r    = 0;
  hitb.gede = iv;   
  ArtePointer<HITB>::append(hitb);

// append the pointers from GEDE to HITB
  int nhitb = ArtePointer<HITB>::nrow();
  if(*gede->hit1 == 0) gede->hit1=nhitb;
  gede->hitl=nhitb;

  return;  
}

// add pad to HITB
void AddPad2(int iv, int pad)
{
  ArtePointer<GEDE> gede(iv);
  int npmx = gede->nwi[2];   // number of pads in row
  if(pad < 0 || pad > 2*npmx-1) return;
  int nr = pad/npmx;         // raw number
  int np = pad-nr*npmx;      // pad number in the raw

  float xd1  = gede->xmin;   // Xmin of chamber
  float xd2  = gede->xmax;   // Xmax of chamber
  float yd1  = gede->ymin;   // Ymin of chamber
  float yd2  = gede->ymax;   // Ymax of chamber

// up - down 
  int ud = 0;
  if(fabsf(yd1) > fabsf(yd2)) ud=1;

  float yl2=0.01*float(gede->nwi[3]); // y-dist between pads
  float dxpad=gede->xfw[1];           // x-size of pad
  float dypd1=gede->xfw[2];           // y-size of first pad
  float dypd2=gede->xfw[3];           // y-size of standard pad
  float zpad =gede->zfw[1];           // z-position of pad
  float alsx =gede->sxmx - gede->sxmn;
  float xl1  =(xd2-xd1-alsx)/2.;      // x-dist from edge to pad
  //  float xl2  =alsx - 2.*dxpad;        // x-dist between pads
  float alsy =gede->symx - gede->symn;
  float yl1  =(yd2-yd1-alsy)/2.;       // y-dist from edge to pad
// y-size of last pad
  float dypd3=alsy-dypd1-yl2-(dypd2+yl2)*float(npmx-2);

// the pad position and size
  float xpad = 0.;
  if(nr == 0) xpad=xd1+xl1+dxpad*0.5;
  if(nr == 1) xpad=xd2-xl1-dxpad*0.5;
  float ypad = 0.;
  float dypad= 0.;
  if(np == 0) {
    if(ud == 0) ypad=yd1+yl1+0.5*dypd1;
    if(ud == 1) ypad=yd2-yl1-0.5*dypd1;
    dypad=dypd1;
  }
  if(np > 0 && np < npmx-1) {
    if(ud == 0) ypad=yd1+yl1+dypd1+yl2+(dypd2+yl2)*float(np-1)+0.5*dypd2;
    if(ud == 1) ypad=yd2-yl1-dypd1-yl2-(dypd2+yl2)*float(np-1)-0.5*dypd2;
    dypad=dypd2;
  }
  if(np == npmx-1) {
    if(ud == 0) ypad=yd2-yl1-0.5*dypd3;
    if(ud == 1) ypad=yd1+yl1+0.5*dypd3;
    dypad=dypd3;
  }

// add the hit to HITB
  struct HITB hitb;
  hitb.x    = xpad;
  hitb.y    = ypad;
  hitb.z    = zpad;
  hitb.q    = dxpad;
  hitb.r    = dypad;
  hitb.gede = iv;   
  ArtePointer<HITB>::append(hitb);

// append the pointers from GEDE to HITB
  int nhitb = ArtePointer<HITB>::nrow();
  if(*gede->hit1 == 0) gede->hit1=nhitb;
  gede->hitl=nhitb;

  return;  
}

// add pixel to HITB
void AddPix2(Word sl,Word qu,Word con,Word bit) 
{
  int  iv  = int(Pix.PixGede[sl][qu]);
  //  int  nx  = int(PixCh.Pnx1[sl][con][bit]-1);
  // modified by Jing-Ge Shiu on May 09, 2000 since the following caculation doesn't count this -1
  int  nx  = int(PixCh.Pnx1[sl][con][bit]);
  int  nxc = int(PixCh.NxC[sl]);
  if(nxc < 1) return;
  //  int  ny = int(PixCh.Pny1[sl][con][bit]-1);
  // modified by Jing-Ge Shiu on May 09, 2000 since the following caculation doesn't count this -1
  int  ny = int(PixCh.Pny1[sl][con][bit]);
  int  nyc = int(PixCh.NyC[sl]);
  if(nyc < 1) return;

  ArtePointer<GEDE> gede(iv);
  float xd1  = gede->xmin;   // Xmin of chamber
  float xd2  = gede->xmax;   // Xmax of chamber
  float yd1  = gede->ymin;   // Ymin of chamber
  float yd2  = gede->ymax;   // Ymax of chamber
  float zd1  = gede->zmin;   // Zmin of chamber
  float zd2  = gede->zmax;   // Zmax of chamber
  float xed  = gede->xfw[0]; // x-dist from edge to pixels
  float yed  = gede->xfw[1]; // y-dist from edge to pixels
  float xsz  = gede->xfw[2]; // x size of elemetary pixel
  float ysz  = gede->xfw[3]; // y size of elemetary pixel

// the relative hit position and size
  float dx = xsz*float(nxc);
  float x = 0.;
  if(sl > 1) x = xed+dx*(float(nx)-0.5);
  if(sl < 2 && nx == 1) {
    dx=xsz;  
     x=xed+dx*0.5;
  }
  if(sl < 2 && nx > 1) x=xed+xsz+dx*(float(nx)-1.5);

  float dy = ysz*float(nyc);
  float y = 0.;
  if(sl != 2) y = yed+dy*(float(ny)-0.5);
  if(sl == 2 && ny == 1) {
    dy=1.5*dy;  
     y=yed+dy*0.5;
  }
  if(sl == 2 && ny > 1) y=yed+dy*float(ny);

// hit coordinates
  if(qu == 0 || qu == 3) x=xd1+x; 
  else                   x=xd2-x; 
  if(qu == 0 || qu == 2) y=yd1+y; 
  else                   y=yd2-y; 

  float z = 0.5*(zd1+zd2);

// add the hit to HITB
  struct HITB hitb;
  hitb.x    = x;
  hitb.y    = y;
  hitb.z    = z;
  hitb.q    = dx;
  hitb.r    = dy;
  hitb.gede = iv;   
  ArtePointer<HITB>::append(hitb);

//  cout << "PIX hit at sl/qu/con/bit/x/y " << sl << "/" 
//       << qu << "/" << con << "/" << bit << "/" << x << "/" << y << endl;

// append the pointers from GEDE to HITB
  int nhitb = ArtePointer<HITB>::nrow();
  if(*gede->hit1 == 0) gede->hit1=nhitb;
  gede->hitl=nhitb;

  return;  
}
}

