/* HorizontalRowsIn_HPT_Chamber class
   -----------------------------------------
   intended to define horizontal rows in HPT chambers.
   It has 
     map<Index, HPT_Chamber::ISection_Cell, less<Index> > cell;
   so that cell[Index(ix,iy)] gives ISection_Cell of the cell
   positioned in the iy-th horizontal row at ix-th place.

   The following conventions for ix,iy numbering are used.
   Y-numbering:
   -IY, -IY+1, ..., -2, -1,   1, 2, IY-1, IY
   note that there is no 0-th row. This is done to have the
   symmetry between y<0 and y>0 halves.

   X-numbering: the boundary cell which is closest to the beam
   has #0, others has numbers in ascending order. 
   Thus for x>0 half: 0, 1, 2, ..., IX (ascending to the right)
        for x<0     : IX, ..., 2, 1, 0 (ascending to the left)

   Note that one (big) cell can be in several horizontal rows
   (as it is in the corners of inner HPT chambers). 
   Thus  (ix,iy) -> 0 or 1 ISection_Cell
   ISection_Cell -> 1, 2, ... (ix,iy)
    
   Creation via constructor or create() member function, e.g.:
     create(HPT_Chamber& chamber);


   HorizontalRowsIn_HPT_Setup class
   --------------------------------
   is a map<HPT_Setup::IChamber, HorizontalRowsIn_HPT_Chamber>
   Creation via constructor or create() member function, e.g.:
     create(HPT_Setup& setup);
   */

// Author V.Balagura (balagura@mail.desy.de)   (21.02.99).

#ifndef _horizontal_rows_in_HPT_Setup_hh
#define _horizontal_rows_in_HPT_Setup_hh 1

#include "ptset/high_pt.hh"

struct HorizontalRowsIn_HPT_Chamber {
  map<Index, HPT_Chamber::ISection_Cell, less<Index> > cell;
  void create(HPT_Chamber& chamber);
  HorizontalRowsIn_HPT_Chamber() {}
  HorizontalRowsIn_HPT_Chamber(HPT_Chamber& chamber) { create(chamber); }
};

struct HorizontalRowsIn_HPT_Setup : 
  map<HPT_Setup::IChamber, HorizontalRowsIn_HPT_Chamber,
  less<HPT_Setup::IChamber> > {
    HorizontalRowsIn_HPT_Setup() {}
    HorizontalRowsIn_HPT_Setup(HPT_Setup& setup) { create(setup); }
    void create(HPT_Setup& setup);
};
    
    
#endif
