// 
//  Author:  Rainer Mankel                     Created:  07-Mar-99
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// class rangerDomain

#include "rang/Domain.hh"
#include "rang/DomWin.hh"
#include "rang/ProxyWin.hh"
#include "rang/Geom.hh"


rangerDomain::rangerDomain(const string dName,
			   const int dScope,const int dActive,
			   const string sl1, const string sl2, 
			   const string sl3){
//   cout << "This is rangerDomain::rangerDomain(const string dName..." << endl;
//   cout << " dName=" << dName << endl;
  name = dName;
  scope = dScope;
  active = dActive;

  if (sl1 != "") {
//     cout << "...append SL " << sl1 << endl;
    slNames.push_back(sl1);
  }
  if (sl2 != "") {
//     cout << "...append SL " << sl2 << endl;
    slNames.push_back(sl2);
  }
  if (sl3 != "") {
//     cout << "...append SL " << sl3 << endl;
    slNames.push_back(sl3);
  }
  setZRef();
  dseg = 0;
  xWi = 0;
  number = 0;
}


rangerDomain* rangerDomain::cloneSl() const {
  switch (slNames.size()) {
  case 0:
    return new rangerDomain(name,scope,active,"");
  case 1:
    return new rangerDomain(name,scope,active,slNames[0]);
  case 2:
    return new rangerDomain(name,scope,active,slNames[0],slNames[1]);
  default:
    return new rangerDomain(name,scope,active,slNames[0],slNames[1],
			    slNames[2]);
  }
}

void rangerDomain::updSls() {
  sls.erase(sls.begin(),sls.end());
  for (vector<string>::iterator nIt=slNames.begin();
       nIt != slNames.end(); ++nIt) {
    
    int found = 0;
    for (ArteTableIterator<GESL> gsIt=ArteTable<GESL>::begin();
           gsIt != ArteTable<GESL>::end(); ++gsIt) {

      string mist((*gsIt)->name,4);
      if (mist == *nIt) {
// 	cout << "superlayer " << *nIt << " has number " << *gsIt << endl;
	sls.push_back(*gsIt);
	found = 1;
	break;
      }
    }
    if (! found) {
      cout << "Domain::updSls: no gesl entry found for " 
	   << *nIt << endl;

    }
  }
  setZRef();
}


int rangerDomain::addWin(const int cmp,const int firstSector,
			 const int lastSector,
			 const float sAlphaMin,const float sAlphaMax,
			 const int maxMod) {

  int nAdded = 0;

  // loop over superlayers in domain
  for (vector<ArtePointer<GESL> >::iterator sIt=sls.begin();
       sIt != sls.end(); ++sIt) {
    if ( cmp && (cmp != (*sIt)->cmp) ) continue;  // check cmp

    // loop over sectors
    for (int sec=firstSector; sec <= lastSector; ++sec) {
      int nModSec = 0;
      rangerDomWin* dw = 0;

      // loop over modules
      for (ArteRelationIterator<GEDE> gIt=((*sIt)->gede).begin();
	   gIt != ((*sIt)->gede).end(); ++gIt) {
	if (nModSec >= maxMod) break;

	if ((*gIt)->sec != sec) continue;  // check sector number
        float sinTot = rangerGeom::vol( (*gIt) )->Sin(); // check stereo angle
	if (sinTot < sAlphaMin || sinTot > sAlphaMax) continue;

	if (! dw) dw = new rangerDomWin(this,(*sIt)->cmp,sec);
	dw->accum(*gIt);
	++nModSec;
      }

      if (dw) {
	windows.push_back(dw);
	++nAdded;
      }
    }
  }
  return nAdded;
}    

void rangerDomain::setZRef(){
  zRef = 0;
  float ztot = 0;
  float ntot = 0;
  for (vector<ArtePointer<GESL> >::iterator sI=sls.begin();
       sI != sls.end(); ++sI) {
    ztot += 0.5 * ((*sI)->z1 + (*sI)->z2);
    ++ntot;
  }
  if (ntot) zRef = ztot / ntot;
}


int rangerDomain::addProxy(const int cmp,const int sec,const int position,
			   const float width,const int cmpP,const int secP){
  int nAdded = 0;
  for (vector <rangerDomWin*>::iterator wit=windows.begin();
       wit !=windows.end(); ++wit) {
    if ((*wit)->cmp() == cmp && (*wit)->sec() == sec) {
      for (vector <rangerDomWin*>::iterator witP=windows.begin();
	   witP !=windows.end(); ++witP) {
	if ((*witP)->cmp() == cmpP && (*witP)->sec() == secP) {
	  (*wit)->addProxy(new rangerProxyWin(*wit,position,width,*witP));
	  ++nAdded;
	}
      }
    }
  }
  return nAdded;
}


int rangerDomain::setProxyLayers(){
  int n = 0;
  for (vector <rangerDomWin*>::iterator wit=windows.begin();
	   wit !=windows.end(); ++wit) {
    n += (*wit)->setProxyLayers();
  }
  return n;
}


ostream& operator<<(ostream &os,const rangerDomain &d){
  os << "(" << d.name << ")";
  return os;
}

void rangerDomain::dump(const int level) const {
  if (level > 0) {
    cout << " Domain " << name << " z=" << zRef
	 << " scope=" << scope << " act=" << active
	 << " nb=" << number
	 << " windows " << windows.size() 
	 << endl;
    for (vector <rangerDomWin*>::const_iterator wIt=windows.begin();
	 wIt != windows.end(); ++wIt) {
      (*wIt)->dump(level-1);
    }
    if (level > 1) 
      cout << "---------------------------------------------------"
	   << endl;
  }
}
