// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// class rangerStateP
// state class for tracking with momentum

#ifndef rangerStateP_HH
#define rangerStateP_HH

#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/Vector.h"
#include "arte/arte.hh"
#include "rang/ScatThick.hh"
#include "rang/FitMode.hh"
#include "rang/State.hh"
#include "rang/HitBase.hh"

class rangerStateP : public rangerState {

public:
  rangerStateP();
  rangerStateP(ArtePointer<RHIT> rhit);
  rangerStateP(ArtePointer<HITB> hitb);
  rangerStateP(ArtePointer<RCCL> rccl,int charge);
  rangerStateP(const rangerStateP& s);  // copy constructor
  ~rangerStateP();
  rangerState& operator=(const rangerStateP& s);  // assignment
  rangerState& operator=(const rangerState& s);

  const double x() const { return p(1); } // parameter lookup accessors
  const double y() const { return p(2); }
  const double tx() const { return p(3); }
  const double ty() const { return p(4); }
  const double qP() const { return p(5); }
  const double I() const { return 0; }

  inline double& setX() { return p(1); }    // parameter update accessors
  inline double& setY() { return p(2); }
  inline double& setTx() { return p(3); }
  inline double& setTy() { return p(4); }
  inline double& setQp() { return p(5); }

  int dim() const {return 5;};   // dimension of state vector

  HepVector& vec();    // lookup function for state vector
  const HepVector& vec() const; // const lookup function for state vector
  HepSymMatrix& cov(); //  lookup function for covariance matrix
  const HepSymMatrix& cov() const; //  lookup function for covariance matrix

  rangerState* clone() const { return new rangerStateP(*this); };
                                   // produce copy
  void covInit();
  int predict(const double& zDestination,
	      const double& qPExp,rangerState*& targetState,
	      HepMatrix*& FTransport);  // perform Kalman filter prediction
  int predictOld(const double& zDestination,
	      const double& qPExp,rangerState*& targetState,
	      HepMatrix*& FTransport);  // perform Kalman filter prediction
  void project(HepVector& resid,HepMatrix& HH,rangerHitBase* hit)
    { hit->projectP(resid,HH,this); } ;
                                     // project into measurement space of hit 
  void applyScat(const rangerScatThin& scat);
                                        // apply m.s. correction 
  void applyScat(const rangerScatThick& scat);
                                        // apply m.s. correction 
  void applyRadEl(const rangerRadEl& rad);
                                  // apply continuous radiative energy loss correction 
  void applyDeDx(const float rl); // apply continuous ionization energy loss correction
  void applyRadPhoton(const float e,const float dE);
                                     // apply resolved radiated photon
  void storeRseg(ArtePointer<RSEG> rseg,int firstLast); 
                                          // store track parameters into RSEG
  void storeRtra(ArtePointer<RTRA> rtra,int firstLast); 
                                          // store track parameters into RTRA
  void storeRhit(ArtePointer<RHIT> rhit); // store parameters as RHIT

  ArtePointer<RPNT> storeRpnt(ArtePointer<RPNT> rpnt);
                                          // store parameters as RPNT
  void copyVec( const rangerState& s); // copy state vector from another state
                                       // (do not copy cov matrix)
  void linExtrapVec( const rangerState& s, float z); 
                                       // extrapolate state vector from 
                                       //another state
  inline rangerFitMode fitMode() { return ModeP; }
  void dump();

private:
  HepVector p;
  HepSymMatrix c;
};

#endif
