
// *****************************************************************************
//
// source:    rcClassify.C
//
// type:      source code
//
// created:   14 Jun 2000
//	
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcClassify
//
// called by: rcevnt_
//
// purpose:   event classification
//
// hints:     flagdst has 32 bits which are defined in eclass.hh.
//            Classification bits should be set by bitwise ORing (!)
//            flagdst with the appropriate mask (ECLASS_...) from eclass.hh.
//            Classification bits must NOT be set in other places.
//            l4trig/l4tEvnt.C sets classification bits for randomly
//            (ECLASS_RNDTRG) and FLT (ECLASS_FLTRG) triggered events.
//
//            Event selection to output is done in l4trig/l4tEvnt.C.
//
// changes:
// 14 Jun 2000 (AG) creation
// 16 Jun 2000 (AG) add timing
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteUtil.h"
#include "farm/farm.hh"
#include "reco/eclass.hh"
#include "reco/rc99Areco.hh"

// externals
extern int FADEB;                // debug flag
extern int rcSelectTracksStatus;

//
// program code
//
int
rcClassify()
{
  // show up
  ARTE_mark( "rcClassify 0", 0 );
  
  // EVHD table
  ArtePointer<EVHD> evhd( 1 );
  
  // flag
  if( FADEB ){
    cout << "CLASS: 0 flagdst=0x" << hex << evhd->flagdst << dec << endl;
  }

  //===========================================================================
  //     J/Psi -> e+e- candidates ECLASS_JPSIEE
  //===========================================================================

  // check track selection
  if( rcSelectTracksStatus ){
  
    // loop over tracks  
    for( ArteTable<RTRA>::iterator rtra = ArteTable<RTRA>::begin();
	 rtra != ArteTable<RTRA>::end(); ++rtra ){
      
      if( (*rtra)->fit == 2 ){
	evhd->flagdst |= ECLASS_JPSIEE; // bit-wise OR	
	cout << "CLASS: ECLASS_JPSIEE flagdst=0x" << hex << evhd->flagdst << dec << endl;
      }
    }
  }

  //===========================================================================
  //     J/Psi -> mu+mu- candidates ECLASS_JPSIMM
  //===========================================================================

  //evhd->flagdst |= ECLASS_JPSIMM; // bit-wise OR	
  //cout << "CLASS: ECLASS_JPSIMM flagdst=0x" << hex << evhd->flagdst << dec << endl;
  
  //===========================================================================
  //     one high E_T calorimeter cluster ECLASS_CAL1HT
  //===========================================================================
  
  //evhd->flagdst |= ECLASS_CAL1HT; // bit-wise OR	
  //cout << "CLASS: ECLASS_CAL1HT flagdst=0x" << hex << evhd->flagdst << dec << endl;

  // flag
  if( FADEB ){
    cout << "CLASS: 1 flagdst=0x" << hex << evhd->flagdst << dec << endl;
  }

  // show up
  ARTE_mark( "rcClassify 1", 0 );  

  // return
  return 0;    
}
