
// *****************************************************************************
//
// source:    rcFillEVTI.C
//
// type:      source code
//
// created:   16 Jun 2000
//	
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcFillEVTI
//
// called by: rcevnt_
//
// purpose:   fill time table EVTI
//
// hints:     EVTI is NOT written to output streams
//
// changes:
// 16 Jun 2000 (AG) creation
//
// *****************************************************************************

// ARTE includes
#include "arte/arte.hh"

// externals
extern float time_rc99AVDS;
extern float time_rc99AOTR;
extern float time_rc99ARICH;
extern float time_rc99ATRD;
extern float time_rc99AECAL;
extern float time_rc99AMUON;
extern float time_rc99Match;
extern float time_rc99Arefit;
extern float time_rc99AParticleID;
extern float time_rc99Avertex;

//
// program code
//
void
rcFillEVTI()
{
  // check EVTI size
  if( ArteTable<EVTI>::size() == 0 )
    mmbook_( ArtePointer<EVTI>::getLink().getRef(), 1 );
  
  // pointer to table EVTI
  ArtePointer<EVTI> evti( 1 );
  
  // fill in times    
  evti->vds    = time_rc99AVDS;
  evti->ranger = time_rc99AOTR;
  evti->refit  = time_rc99Arefit;
  evti->rich   = time_rc99ARICH;
  evti->ecal   = time_rc99AECAL;
  evti->trd    = time_rc99ATRD;
  evti->muon   = time_rc99AMUON;
  evti->marple = time_rc99Match;
  evti->vertex = time_rc99Avertex;
  evti->pid    = time_rc99AParticleID;
}
