// ********************************************************************
//
//  Routine:      rcSelectJpsi
//  Author:       T.Oest
//
//  Description:  search for J/psi candidates
//
//  Modification:
//  -- faster J/psi selection by changing order of cuts  
//                                                      01/12/99 T.Oest
//  -- changed the cuts in the J/psi selection, 
//     now each lepton must be identified either by
//     Bremsstrahlung, RICH or OTR                      05/06/00 
//                                                           D. Samtleben
//
// ********************************************************************


#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include "arte/arte.hh"
#include "reco/rc99Areco.hh"
#include "arte/RSEG_cmp.hh"
#include "arte/ArteUtil.h"


float           rc99Bremsstrahlung(ArtePointer<RTRA>);
extern "C" void rcv2fit_(const int   & rtra1, 
                         const int   & rtra2, 
                         const float & mass1,
                         const float & mass2,
                         const float & chi,
                         const float * array,
                         const float * array,
                         const float * array,
                         const float * array,
                         const float & mass,
                         const int   & flag);

//*****************************************************************
// function
//*****************************************************************

int rcSelectJpsi() {

static const int index[5][5] = { {0, 1,  3, 6,10},
                          {1, 2,  4, 7,11},
                          {3, 4,  5, 8,12},
                          {6, 7,  8, 9,13},
                          {10,11,12,13,14}};
//==================================================================
// 1. Cuts
//==================================================================
    float dyCut         = 3.;
    float pCut          = 5.;
    float ptCut         = 0.0;
    float EtCut         = 1.;
    float EoverPCut     = 0.36;
    float MassCut       = 2.0;
    float chi2VertexCut = -1.;


//==================================================================
// 
//==================================================================
    float px1, py1, pz1, p1, px2, py2, pz2, p2, pt1, pt2;
    float Et1, Et2, Ee1, Ee2, Eb1, Eb2;
    float px, py, pz, p, mee, meecorr, pt, mass;

    float dy1, dy2;

    for( ArteTable<RTRA>::iterator rtra1 = ArteTable<RTRA>::begin();
         rtra1 != ArteTable<RTRA>::end(); ++rtra1 ) {


     ArteRelationIterator<RCCL> rccl1     = (*rtra1)->rccl.end();
     ArteRelationIterator<RSEG> rVDS1     = (*rtra1)->rseg.end();
     ArtePointer<RSEG> rRICH1     = 0;
     ArtePointer<RSEG> rOTR1     = 0;

    if ((*rtra1)->rccl.begin() != (*rtra1)->rccl.end()) 
                          rccl1 = (*rtra1)->rccl.begin();


    for (ArteRelationIterator<RSEG> rseg2 = (*rtra1)->rseg.begin();
         rseg2 != (*rtra1)->rseg.end(); ++rseg2){

     if ((*rseg2)->fit <= 0 ||  (*rseg2)->fit == 8 ||
          (*rseg2)->fit == 10){

       if ((*rseg2)->cmp & Rsegc::vxd) rVDS1 = rseg2;
       if ((*rseg2)->cmp & Rsegc::rich) rRICH1 = *rseg2;
       if ((*rseg2)->cmp & Rsegc::patt) rOTR1 = *rseg2;
    };
    }

     p1  = fabs(1./(*rtra1)->pf);


     if (rVDS1 != (*rtra1)->rseg.end() && rccl1 != (*rtra1)->rccl.end()
         && p1 > pCut && fabs((*rccl1)->esum/fabs(p1)-1.)< EoverPCut){


     pz1 = p1 / sqrt(1.+(*rtra1)->txf*(*rtra1)->txf +
                        (*rtra1)->tyf*(*rtra1)->tyf);
     px1 = (*rtra1)->txf * pz1;
     py1 = (*rtra1)->tyf * pz1;
     pt1 = sqrt(px1*px1+py1*py1);
     Et1 = (*rccl1)->esum * 
           sqrt((*rccl1)->x*(*rccl1)->x+(*rccl1)->y*(*rccl1)->y)/(*rccl1)->z;
     dy1 = (*rccl1)->y -
           ((*rtra1)->yf + (*rtra1)->tyf *((*rccl1)->z - (*rtra1)->zf));


     if (pt1 > ptCut && Et1 > EtCut && fabs(dy1) < dyCut){




    for( ArteTable<RTRA>::iterator rtra2 = rtra1;
         rtra2 != ArteTable<RTRA>::end(); ++rtra2 ) {

     ArteRelationIterator<RCCL> rccl2     = (*rtra2)->rccl.end();
     ArteRelationIterator<RSEG> rVDS2     = (*rtra2)->rseg.end();
     ArtePointer<RSEG> rRICH2 = 0;
     ArtePointer<RSEG> rOTR2 = 0;

    if ((*rtra2)->rccl.begin() != (*rtra2)->rccl.end()) 
                          rccl2 = (*rtra2)->rccl.begin();

    for (ArteRelationIterator<RSEG> rseg2 = (*rtra2)->rseg.begin();
         rseg2 != (*rtra2)->rseg.end(); ++rseg2){

     if ((*rseg2)->fit <= 0 ||  (*rseg2)->fit == 8 ||
          (*rseg2)->fit == 10){

       if ((*rseg2)->cmp & Rsegc::vxd) rVDS2 = rseg2;
       if ((*rseg2)->cmp & Rsegc::rich) rRICH2 = *rseg2;
       if ((*rseg2)->cmp & Rsegc::patt) rOTR2 = *rseg2;
    };

    };

         p2  = fabs(1./(*rtra2)->pf);

     if (rVDS2 != (*rtra2)->rseg.end() && rccl2 != (*rtra2)->rccl.end()
         && (*rtra1)->pf * (*rtra2)->pf < 0 
         && p2 > pCut && fabs((*rccl2)->esum/fabs(p2)-1.) < EoverPCut){


     if ((*rccl1) != (*rccl2) && (*rVDS1) != (*rVDS2)){

         pz2 = p2 / sqrt(1.+(*rtra2)->txf*(*rtra2)->txf +
                            (*rtra2)->tyf*(*rtra2)->tyf);
         px2 = (*rtra2)->txf * pz2;
         py2 = (*rtra2)->tyf * pz2;
         pt2 = sqrt(px2*px2+py2*py2);
         Et2 = (*rccl2)->esum * 
               sqrt((*rccl2)->x*(*rccl2)->x+(*rccl2)->y*(*rccl2)->y)/(*rccl2)->z;
         dy2 = (*rccl2)->y -
               ((*rtra2)->yf + (*rtra2)->tyf *((*rccl2)->z - (*rtra2)->zf));

     if (pt2 > ptCut && Et2 > EtCut && fabs(dy2) < dyCut){


         px  = px1 + px2;
         py  = py1 + py2;
         pz  = pz1 + pz2;
         p   = sqrt(px*px+py*py+pz*pz);
         pt  = sqrt(px*px+py*py);


//-----------------------------------------------------------------
//   electron electron mass
//-----------------------------------------------------------------
         float corr1, corr2;

         Ee1    = p1;
         Ee2    = p2;
         Eb1    = rc99Bremsstrahlung((*rtra1));
         Eb2    = rc99Bremsstrahlung((*rtra2));

         mee    = sqrt( 2.*(Ee1*Ee2 - px1*px2 - py1*py2 - pz1*pz2));

         if (Ee1 != 0.0 && Ee2 != 0.0) {
          corr1     = 1. + Eb1 / Ee1;
          corr2     = 1. + Eb2 / Ee2;
          meecorr   = mee * sqrt(corr1 * corr2);
         } else {
	  meecorr   = 0.0;
         };

//-----------------------------------------------------------------
//   mass cut
//-----------------------------------------------------------------
    if (meecorr > MassCut
     	&&(Eb1>0||rRICH1!=0||rOTR1!=0)&&(Eb2>0||rRICH2!=0||rOTR2!=0)){
      //	     &&((Eb1>0||Eb2>0)||(rRICH1!=0&&rRICH2!=0))){
//-----------------------------------------------------------------
//   do a vertex fit
//-----------------------------------------------------------------

         int r1 = *rtra1 - *ArteTable<RTRA>::begin() +1;
         int r2 = *rtra2 - *ArteTable<RTRA>::begin() +1;
         float x[3], cov[3][3], vchi, ppsi[6], cpsi[5][5];
         int FitFlag;

         float m1 = 0.0;
         float m2 = 0.0;

         rcv2fit_(r1, r2, m1, m2, vchi, x, &cov[0][0],
                  ppsi, &cpsi[0][0], mass, FitFlag);

//-----------------------------------------------------------------
//   vertex cut
//-----------------------------------------------------------------
      if (chi2VertexCut < 0.0 ||
          (FitFlag == 0 && vchi < chi2VertexCut) ) {
            arteutil_.eventout=1;

 
//-----------------------------------------------------------------
//   fill RTRA and vertex bank
//-----------------------------------------------------------------
      RTRA rtra;

      rtra.fit = 2;

      rtra.hsi = 0;
      rtra.hca = 0;
      rtra.hin = 0;
      rtra.hou = 0;
      rtra.hri = 0;
      rtra.htr = 0;
      rtra.hmu = 0;

      rtra.flr = 0;
      rtra.flt = 0;
      rtra.fle = 0;
      rtra.flm = 0;
      rtra.fid = 0;

      rtra.lre = 0.;
      rtra.lrmu= 0.;
      rtra.lrpi= 0.;
      rtra.lrk = 0.;
      rtra.lrp = 0.;

      rtra.lmmu= 0;
      rtra.lmh = 0;

      rtra.lte = 0;
      rtra.lth = 0;

      rtra.lee = meecorr;
      rtra.leh = 0;
      rtra.lemi= 0;

      rtra.le  = 0.;
      rtra.lmu = 0.;
      rtra.lpi = 0.;
      rtra.lk  = 0.;
      rtra.lp  = 0.;

      rtra.xf  = ppsi[1];
      rtra.yf  = ppsi[2];
      rtra.zf  = ppsi[0];
      rtra.txf = ppsi[3];
      rtra.tyf = ppsi[4];
      rtra.chi = vchi;
      rtra.pf  = ppsi[5];

      for (int i=0; i<=3; i++){
      for (int j=0; j<=3; j++){
        rtra.cvf[index[i][j]] = cpsi[i+1][j+1];
      };};

      ArteTable<RTRA>::push_back( rtra );
      ArtePointer<RTRA> Prtra(ArteTable<RTRA>::size());

      // -- vertex --------------------------------

      if (FitFlag == 0) {

      RVER rver;

      rver.xv   = x[0];
      rver.yv   = x[1];
      rver.zv   = x[2];
      rver.mass = mass;
      rver.merr = 0.0;
      rver.nt   = 2;
      rver.btra = Prtra;

      rver.cv[0]= cov[0][0];
      rver.cv[1]= cov[0][1];
      rver.cv[2]= cov[1][1];
      rver.cv[3]= cov[0][2];
      rver.cv[4]= cov[1][2];
      rver.cv[5]= cov[2][2];

      ArteTable<RVER>::push_back( rver );
      ArtePointer<RVER> Prver(ArteTable<RVER>::size());

      Prver->btra = Prtra;
      Prtra->ever = Prver;

      // -- RREL tables --------------------------------

      RREL rrel;

      ArtePointer<TOPP> eTOPP (ArteTable<TOPP>::begin() + 18 - 1);
      ArtePointer<TOPP> pTOPP (ArteTable<TOPP>::begin() + 19 - 1);

      if ((*rtra1)->pf <  0) {rrel.topp =  eTOPP;}
      else                   {rrel.topp =  pTOPP;};
      rrel.flag = 1;
      rrel.chi  = 0.5*vchi;
      
      ArteTable<RREL>::push_back( rrel );
      ArtePointer<RREL> Prrel1(ArteTable<RREL>::size());


      if ((*rtra2)->pf <  0) {rrel.topp =  eTOPP;}
      else                   {rrel.topp =  pTOPP;};

      ArteTable<RREL>::push_back( rrel );
      ArtePointer<RREL> Prrel2(ArteTable<RREL>::size());


      // -- relations --------------------------------

      createRelation( Prver, Prrel1, &RVER::rrel );
      createRelation(*rtra1, Prrel1, &RTRA::rrel );

      createRelation( Prver, Prrel2, &RVER::rrel );
      createRelation(*rtra2, Prrel2, &RTRA::rrel );

      };
      };

  
     };};};
     };};};
     };};


 return 0;

};
