
// *****************************************************************************
// 
//  Author:  Rainer Mankel                     Created:  17-Apr-99
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
//
// rcevnt.C
// reconstruction main program
// dispatch selecting the appropriate main program according to
// time or geometry version
//
// changes:
// 17 Dec 1999 (AG) introduce timing
//  8 Feb 2000 (TO) reconstruction module is selected by flags
// 26 May 2000 (AG) store timing in global variable
//  7 Jun 2000 (AG) initialize timers to 0; fill new table EVTI
//  8 Jun 2000 (AG) use ARTE_mark()
// 14 Jun 2000 (AG) invoke event classification here
// 16 Jun 2000 (AG) introduce rcFillEVTI() 
//
// *****************************************************************************

// standard includes
#include <math.h>

// ONLINE includes
#include "rhp.h"

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteMessage.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "dq/DQ.h"
#include "reco/rcflags.hh"
#include "reco/reco.hh"
#include "reco/recoOptions.hh"
#include "wax/OTrSelect.h"

// prototypes
extern "C" void rcZevnt_();
unsigned int    acFlag();
int             rcClassify(); // event classification
void            rcFillEVTI(); // fill time table EVTI

// globals
float time_rcevnt     = 0;
float time_rcClassify = 0;

// externals
extern recoOptions RecoOpt;
extern int         FATIME;

//
// program code
//
extern "C" void
rcevnt_()
{
  // mark
  ARTE_mark( "rcevnt_ 0", 0 );

  //-------------------------------------------------------------------
  //    initialize the reco error flag
  //-------------------------------------------------------------------
  if( recoFlag ) rcSetError("init");

  //-------------------------------------------------------------------
  //    create new EVRC bank if reco is switched on 
  //-------------------------------------------------------------------
  ArtePointer<EVRC> evrc;

  if( recoFlag ){
    int size = ArteTable<EVRC>::size();
    mmbook_( ArtePointer<EVRC>::getLink().getRef(), 1 );
    evrc = ArtePointer<EVRC>( size + 1 );
    evrc->ipaddr   = arteflag_.ipaddr;
    evrc->errmask  = 0x0;
    evrc->date     = time( NULL );
    evrc->cputime  = 0.0;
    evrc->alca     = acFlag();
    evrc->artevers = arteflag_.artevers;
  }

  //-------------------------------------------------------------------
  //     start timing
  //-------------------------------------------------------------------
  static Chrono *cron_rcevnt     = 0;
  static Chrono *cron_rcClassify = 0;

  if( !cron_rcevnt ){
    ChronoManager CM;
    CM.instance();

    cron_rcevnt     = CM.createChrono( "rcevnt" );
    cron_rcClassify = CM.createChrono( "rcClassify" ); 
  }
  cron_rcevnt->start();

  //-------------------------------------------------------------------
  //     event reconstruction
  //-------------------------------------------------------------------
  if( recoFlag ){
    ARTE_mark( "rc99Aevnt", 0 );
    rc99Aevnt();
  } 
  else {
    if (RecoOpt.CheckActiveAlg("RITER")  || 
	RecoOpt.CheckActiveAlg("CATS")   ||
	RecoOpt.CheckActiveAlg("HOLMES") ||
	RecoOpt.CheckActiveAlg("MARPLE") ||
	RecoOpt.CheckActiveAlg("RIRE")   ||
	RecoOpt.CheckActiveAlg("RITER")  ||
	RecoOpt.CheckActiveAlg("VERTEX"))  rcZevnt_();
  }
  
  //-------------------------------------------------------------------
  //     event classification
  //-------------------------------------------------------------------
  if( recoFlag ){
    ARTE_mark( "rcClassify", 0 );
    cron_rcClassify->start();
    rcClassify();
    cron_rcClassify->stop();
    time_rcClassify = cron_rcClassify->lastTime();

    if( FATIME ) cout << " RECO: class=" << time_rcClassify << endl;
  } 

  //-------------------------------------------------------------------
  //       fill timing table EVTI
  //-------------------------------------------------------------------  
  if( recoFlag ) rcFillEVTI();

  //-------------------------------------------------------------------
  //     stop timing
  //-------------------------------------------------------------------
  cron_rcevnt->stop();
  time_rcevnt = cron_rcevnt->lastTime();

  //-------------------------------------------------------------------
  //     fill in EVRC bank 
  //-------------------------------------------------------------------
  if( recoFlag ) {
    evrc->cputime = time_rcevnt;
    evrc->errmask = recoErrorFlag();
    OTrEventTime( evrc->ottmean,  evrc->ottwidth);
  }

  // mark  
  ARTE_mark( "rcevnt_ 1", 0 );
}
