/////////////////////////////////////////////////////
//
//  Modified:
//    16 May, 2000 (DD) Filling of hot channels into DB
//    17 May, 2000 (DD) Keytable initialization
//    27 May, 2000 (DD) Bug fix
//
/////////////////////////////////////////////////////

#include "rich/RICHDB.h"
#ifdef USE_ARTE
#include "farm/farm.hh"
#endif
#include "grr.h"
#include "database/keytable.h"


int DAQ_INITIALIZATION::distribute=0;

int DAQ_INITIALIZATION::RichGeoServer     = -1;  
int DAQ_INITIALIZATION::RichASDtable      = -1; 
int DAQ_INITIALIZATION::RichCDIMtable     = -1; 
int DAQ_INITIALIZATION::RichCEFFtable     = -1; 
int DAQ_INITIALIZATION::RichCLENtable     = -1; 
int DAQ_INITIALIZATION::RichCPDWtable     = -1; 
int DAQ_INITIALIZATION::RichCPMDtable     = -1; 
int DAQ_INITIALIZATION::RichCPMItable     = -1; 
int DAQ_INITIALIZATION::RichCPMTtable     = -1; 
int DAQ_INITIALIZATION::RichCPOItable     = -1; 
int DAQ_INITIALIZATION::RichCSMItable     = -1; 
int DAQ_INITIALIZATION::RichCSPMtable     = -1; 
int DAQ_INITIALIZATION::RichCSSMtable     = -1; 
int DAQ_INITIALIZATION::RichCSUPtable     = -1; 
int DAQ_INITIALIZATION::RichCVEStable     = -1; 
int DAQ_INITIALIZATION::RichChannelStatusTable = -1; 
int DAQ_INITIALIZATION::RichM16WireMapTable    = -1; 
int DAQ_INITIALIZATION::RichM4WireMapTable     = -1; 




// ------------------------------------------------------------------
DAQ_INITIALIZATION::DAQ_INITIALIZATION(char *sours){
  if (!sours) {
    char dbsours[256];
    dbsours[0]='\0';
    char* RICHdbsours=getenv("RICH_INSTALLDB");
    if (RICHdbsours) strcpy(dbsours,RICHdbsours);
    else {
      fprintf(stderr,"Set environment variable RICH_INSTALLDB before starting the program\n");
      Exit(1);
    }
    if (RichGeoServer==-1) 
      tid = OpenTables(dbsours, 0); 
    else
      tid = OpenTables(KeyServerName(RichGeoServer),TABLEreadonly);
  }
  for (int j=0; j<3; j++) PMType[j]=NULL;
  for (int j=0; j<4; j++) ASDConn[j]=NULL;
  for (int j=0; j<4; j++) ASD[j]=NULL;
  SuperMod=NULL;
  PMBRow=NULL;
  PMBCol=NULL;
  MotherB=NULL;
  DaughterB=NULL;
  DBConn=NULL;
  PMT=NULL;
  Pad=NULL;
  Channel=NULL;
  Chan_ID=NULL;
  Chan_Stat=NULL;
  fsize=0;
  if (sours) {
    way=new char[256];
    way[0]='\0';
    strcpy (way,sours);
  }
  else way=0;
  file=0;
}


// ------------------------------------------------------------------
DAQ_INITIALIZATION::~DAQ_INITIALIZATION(){

  for (int j=0; j<3; j++) if (PMType[j])  delete [] PMType[j];
  for (int j=0; j<4; j++) if (ASDConn[j]) delete [] ASDConn[j];
  for (int j=0; j<4; j++) if (ASD[j])     delete [] ASD[j];
  if (SuperMod)  delete [] SuperMod;
  if (PMBRow)    delete [] PMBRow;
  if (PMBCol)    delete [] PMBCol;
  if (MotherB)   delete [] MotherB;
  if (DaughterB) delete [] DaughterB;
  if (DBConn)    delete [] DBConn;
  if (PMT)       delete [] PMT;
  if (Pad)       delete [] Pad;
  if (Channel)   delete [] Channel;
  if (Chan_ID)   delete [] Chan_ID;
  if (Chan_Stat) delete [] Chan_Stat;
  //if (fsize)   delete &fsize;
  if (way)  delete way; 
  else      CloseTables(tid);
  if (file) delete file;
}


// ------------------------------------------------------------------
int DAQ_INITIALIZATION::ASDInitialization() {
  int j;

  PrepareTable(tid,"ASD",RichASDtable);

  SkipTableField(tid,1);
  fsize=TableFieldSize(tid);
  for (j=0;j<3;j++) PMType[j]= new char[fsize];
  for (j=0;j<4;j++) ASDConn[j]= new char[fsize];
  SuperMod= new int[fsize];
  PMBRow= new int[fsize];
  PMBCol= new int[fsize];
  MotherB= new int[fsize];
  DaughterB= new int[fsize];
  DBConn= new int[fsize];
  if (!DBConn) { 
    fprintf(stderr,"DAQ_INITIALIZATION::ASDInitialization() Out of memory: Memory allocation error.\n");
    Exit(1);
  }
  //printf("Fieldsize ASD=%d\n", fsize);
  
  for (j=0; j<3; j++){
    GetTableField(tid,TABLEascii  ,fsize, PMType[j]);
    //printf("PMType[%d]=%s\n",j,PMType[j]);
	}
  for (j=0; j<4; j++)  GetTableField(tid,TABLEascii  ,fsize, ASDConn[j]);
  GetTableField(tid,TABLEint  ,fsize,SuperMod);
  GetTableField(tid,TABLEint  ,fsize,PMBRow);
  GetTableField(tid,TABLEint  ,fsize,PMBCol);
  GetTableField(tid,TABLEint  ,fsize,MotherB);
  GetTableField(tid,TABLEint  ,fsize,DaughterB);
  GetTableField(tid,TABLEint  ,fsize,DBConn);
  return 1;
}

// ------------------------------------------------------------------
int DAQ_INITIALIZATION::M16WInitialization() {
  int j;

  PrepareTable(tid,"M16WireMap",RichM16WireMapTable);


  SkipTableField(tid,1);
  fsize=TableFieldSize(tid);
  //printf("Fieldsize M16WireMap=%d\n", fsize);
  for (j=0;j<4;j++) ASD[j]= new char[fsize];
  PMT=new int[fsize];
  Pad=new int[fsize];
  Channel=new int[fsize];
  GetTableField(tid,TABLEint  ,fsize,PMT);
  GetTableField(tid,TABLEint  ,fsize,Pad);
  for (j=0; j<4; j++)  GetTableField(tid,TABLEascii  ,fsize, ASD[j]);
  GetTableField(tid,TABLEint  ,fsize,Channel);
  return 1;
}

// ------------------------------------------------------------------
int DAQ_INITIALIZATION::M4WInitialization() {

  PrepareTable(tid,"M4WireMap",RichM4WireMapTable);

  SkipTableField(tid,1);
  fsize=TableFieldSize(tid);
  //printf("Fieldsize M4WireMap=%d\n", fsize);
  PMT=new int[fsize];
  Pad=new int[fsize];
  Channel=new int[fsize];
  GetTableField(tid,TABLEint  ,fsize,PMT);
  GetTableField(tid,TABLEint  ,fsize,Pad);
  GetTableField(tid,TABLEint  ,fsize,Channel);
  return 1;
}
// --------------------------------------------------------------------
int DAQ_INITIALIZATION::BadTablesList() {
  int entr_num=BookTable(tid, "ChannelStatus", 0, 0); 
  //Put the TableStamp on the entry which has a 0 major and minor numbers:
  TableStamp(tid, 0, 0);
  printf("'ChannelStatus' tables, which are stored in the database:\n");
  while (MatchTable(tid, entr_num, 1)){
  printf("%d\t",TableMajor(tid));
  TableStamp(tid, TableMajor(tid), 0);
  }
  printf("\nThe number of the table is associated with the run number.\n");
  printf("Use RICH_BADTABLE=[number] environment variable to read selected table.\n");
  return 1;
}

// --------------------------------------------------------------------
int DAQ_INITIALIZATION::DefaultStatus() {
  RICHDetector& det=*RICH_Detector;
  int sm,r,c,p;
  fsize=det.NumChannel();
  Chan_ID= new int[fsize];
  Chan_Stat= new short[fsize];
  for (det.NewChannelLoop();det.ChannelLoop(&sm,&r,&c,&p);) {
    Chan_ID[det.ChannelIndex()]=ChannelID(sm,r,c,p);
    Chan_Stat[det.ChannelIndex()]=1;
  }
  return 1;
}
// --------------------------------------------------------------------
int DAQ_INITIALIZATION::StatusfromFile() {
  DefaultStatus();
  RICHDetector& det=*RICH_Detector;
  int idd,statt;
  char s[250], n[256];
  s[0]=0;
  n[0]=0;
  strcpy(n,way);
  char* badfile=getenv("RICH_BADFILE");
  if (badfile) {
    strcat(n,badfile);
    //printf("RICH_BADFILE=%s\n",n);
    file=new ScriptFile(n);
    while (file->getline(s)) {
      sscanf(s,"%d %d", &idd, &statt);
      if (det.ChannelExist(idd)) {
	Chan_Stat[det.ChannelIndex()]=statt+1;
      }
    }
    RichRun::bad_channels=2;
    return 1;
  }
  RichRun::bad_channels=0;
  return 0;
}

// --------------------------------------------------------------------
int DAQ_INITIALIZATION::ChannelStatus() {
  DefaultStatus();
  char* analize=getenv("RICH_OK");
  if (!analize) {
    RICHDetector& det=*RICH_Detector;
    int entr=1, rev=0, entr_num=-1, *chan_id;
    short *chan_stat;
    entr_num=BookTable(tid,  "ChannelStatus", 0, 0); 

    char* request=getenv("RICH_BADTABLE");
    if (request) { 
      entr=atoi(request); 
      RichRun::bad_channels=4; 
    }
    else if (RichChannelStatusTable!=-1) {
      entr=KeyTableMajor(RichChannelStatusTable); 
      rev =KeyTableMinor(RichChannelStatusTable); 
      RichRun::bad_channels=3; 
    }
    else  {
         TableStamp(tid, TABLEmaxrev, TABLEmaxrev);
         MatchTable(tid, entr_num, 2);
         entr=TableMajor(tid);
         rev =TableMinor(tid);
         RichRun::bad_channels=3; 
    }
    RichRun::bad_channels_reference=entr;

    //printf("RICH_BADTABLE=%d\n",entr);
    TableStamp    (tid, entr, rev);
    if (!FetchTable(tid,entr_num,0)) {
      fprintf(stderr,"DAQ_INITIALIZATION::ChannelStatus: No table with entry number %d\n",entr);
      Exit(1);
    }
    SkipTableField(tid,1);
    fsize=TableFieldSize(tid);
    //printf("Fieldsize Chan_Stat=%d\n", fsize);
    chan_id= new int[fsize];
    chan_stat= new short[fsize];
    GetTableField(tid,TABLEint  ,fsize,chan_id); 
    GetTableField(tid,TABLEshort,fsize,chan_stat); 
    for (int i=0; i<fsize; i++){
      if (det.ChannelExist(chan_id[i])) {
	Chan_Stat[det.ChannelIndex()]=chan_stat[i];
      }
    }
  }
  return 1;
}

// -------------------------------------------------------------------------
int DAQ_INITIALIZATION::sours_selector (char* subject){	
  if (way){
    char n[256];
    n[0]=0;
    strcpy(n,way);
    if (!strcmp(subject,"status")) { StatusfromFile(); return 1; }
    else if (!strcmp(subject,   "asd")) strcat(n,"ASD.dat");
    else if (!strcmp(subject,   "m16")) strcat(n,"m16wiremap.dat");
    else if (!strcmp(subject,    "m4")) strcat(n,"m4wiremap.dat");
    file=new ScriptFile(n);
  }
  else {
    if (!strcmp(subject,"status")) ChannelStatus();
    else if (!strcmp(subject,   "asd")) ASDInitialization();
    else if (!strcmp(subject,   "m16")) M16WInitialization();
    else if (!strcmp(subject,    "m4")) M4WInitialization();
  }                        
  return 1;
}

// -------------------------------------------------------------------------
int DAQ_INITIALIZATION::asd_scan(char* ftag,char* fasd, int* fsm, 
				 int *fr, int *fc,int *ffedid, int *fd, int *fcon) { 
  if (way){
    char s[256];
    s[0]=0;
    if(!file->getline(s))return 0;
    if (sscanf(s,"%s %s %d %d %d %d %d %d",ftag,fasd,fsm,fr,
	            fc,ffedid,fd,fcon)!=8)
      {      
	fprintf(stderr,"ASD.dat: Error scanning %s\n",s);
	Exit(1);
      }
    return 1;
  }
  else {
    static int i=0;
    char pmt[5], conn[6];
    pmt[0]=0;
    conn[0]=0;
    if (i<FS()){
      for (int j=0; j<3; j++) pmt[j]=PMType[j][i]; pmt[3]='\0';
      sscanf(pmt,"%s",ftag);
      for (int j=0; j<4; j++) conn[j]=ASDConn[j][i]; conn[4]='\0';
      sscanf(conn,"%s",fasd);
      *fsm=SuperMod[i];
      *fr=PMBRow[i];
      *fc=PMBCol[i];
      *ffedid=MotherB[i];
      *fd=DaughterB[i];
      *fcon=DBConn[i];
      //printf("%d,%s, %s, %d, %d, %d, %d, %d, %d\n",i,ftag,fasd,*fsm,*fr,*fc,*ffedid,*fd,*fcon);
      i++;
    }
    else return 0; 
    return 1;
  }
}		


// -------------------------------------------------------------------------
int DAQ_INITIALIZATION::m16_scan(char* s) { 
  if (way){
    if (!file->getline(s)) return 0;
    return 1;
  }
  else {
    static int i=0;
    char conn[6];
    if (i<FS()){
      for (int j=0; j<4; j++) conn[j]=ASD[j][i]; conn[4]='\0';
      sprintf(s,"%d\t%d\t%s\t%d",PMT[i],Pad[i],conn,Channel[i]);
      //printf("%d %d\n",i,FS());
      i++;
    }
    else return 0;
    return 1;
  }
}              


// -------------------------------------------------------------------------
int DAQ_INITIALIZATION::m4_scan(char* s) { 
  if (way){
    if (!file->getline(s)) return 0;
    return 1;
  }
  else {
    static int i=0;
    if (i<FS()){
      sprintf(s,"%d\t%d\t%d",PMT[i],Pad[i],Channel[i]);
      //printf("%d %d\n",i,FS());
      i++;
    }
    else return 0; 
    return 1;
  }
}              

// -------------------------------------------------------------------------
int DAQ_INITIALIZATION::channelstatus_scan(int* id, short* stat) { 
  RICHDetector& det=*RICH_Detector;
  static int i=0;
  if (i<det.NumChannel()){
    *id=Chan_ID[i];
    *stat=Chan_Stat[i];
    i++;
  }
  else return 0; 
  return 1;
}       

// --------------------------------------------------------------------

#ifndef USE_ARTE
static void ARTE_error( int sev, char* msg ,int ret) {
  printf(msg);
  printf("\n");
}
#endif
 

void DAQ_INITIALIZATION::channelstatus_fill(int *chid, short int *chstat, int nchan, int run_no) {
  int verb=0;

  char dbsite[256];
  dbsite[0]=0;
  char *getdbsite=getenv("RICH_INSTALLDB");
  if (!getdbsite) {
    ARTE_error( GRR_SERROR," channelstatus_fill: RICH_INSTALLDB NOT defined, cannot update database" , -1);
    return;
  }
  strcpy(dbsite,getdbsite);

  Tables x=OpenTables(dbsite,0);
  if (!x) { fprintf(stderr,"Cannot fatch a table for run %d/table not found\n",run_no); return; }

  if (distribute) {
    ARTE_error( GRR_SINFO," channelstatus_fill:Distributing table " , -1);
    DistributeTables(x);
  }

  int tx=BookTable(x,"ChannelStatus",0,0);
  if (!tx) { 
    fprintf(stderr," $$$ DAQ_INITIALIZATION::channelstatus_fill: Cannot fatch a table for run %d/table not found\n",
	    run_no); 
    return; 
  }

  TableStamp(x,run_no,TABLEmaxrev);
  if (!FetchTable(x,tx,TABLEle)) { 
    fprintf(stderr," $$$ DAQ_INITIALIZATION::channelstatus_fill: Cannot fatch a table for run %d/table not found\n",
	    run_no); 
    return; 
  }

  SkipTableField(x,1);
  int   major=TableMajor(x);
  int   minor=TableMinor(x);  
  int   run_ver=0;
  if (major==run_no) run_ver=minor+1;


  TableStamp(x,run_no,run_ver);
  SkipTableField(x,1);

  if (verb>1) {
    fprintf(stderr," $$$ DAQ_INITIALIZATION::channelstatus_fill: New parameters for run %d (minor=%d)\n",
	    run_no,run_ver);
    if (major==run_no)
      fprintf(stderr," $$$ DAQ_INITIALIZATION::channelstatus_fill: Old version run %d (minor=%d)\n",
	      major,minor);
  }

  PutTableField(x,  TABLEint,nchan,  chid);
  PutTableField(x,TABLEshort,nchan,chstat);

  StoreTable(x,tx,0);
  if (verb>1) PrintTables(x,1);
  CloseTables(x);
  
  ARTE_error( GRR_SINFO, " $$$ channelstatus_fill: RICH hot-channel list has been updated!", 0);
}



// --------------------------------------------------------------------


int  DAQ_INITIALIZATION::PrepareTable(Tables &table, char *tname, int RichTable) {
  if (RichTable==-1) {
    int nent = BookTable(table, tname, 0, 0);
    TableStamp(table, 1, 0);
    FetchTable(table, nent,0);
  }
  else {
    if (!KeyTableNew(RichTable)) return 0;
    int x = BookTable(table, tname, 0, 0);
    TableStamp(table, KeyTableMajor(RichTable), KeyTableMinor(RichTable));    
    if (!FetchTable(table,x,0)) { 
      ARTE_error( GRR_SERROR," $$$ PrepareTable: Cannot fatch a table for assigned key", -1);
      return -1;
    }
  }
  return 0;
}

// --------------------------------------------------------------------

int DAQ_INITIALIZATION::RichKeyTableInit(int &RichTable, char* TableName) {
  RichTable=KeyFindTable(RichGeoServer, TableName);
  if (RichTable==-1) {
    ARTE_error( GRR_SWARNING," $$$ RichKeyTableInit: KeyTable not found", -1);
    return 1;
  }
  return 0;
}

int DAQ_INITIALIZATION::RichKeyInit() {

  char *dbgeosite=getenv("RICH_INSTALLDB");
  if (!dbgeosite) {
    ARTE_error( GRR_SERROR," $$$ RichKeyInit: RICH_GEO not defined" , -1);
    return -1;
  }
  RichGeoServer = KeyFindServer(dbgeosite);
  if (RichGeoServer==-1) {
    ARTE_error( GRR_SWARNING," $$$ RichKeyInit: KeyTable server not found", -1); 
    return -1;
  }
  int ret=0;
  // key1: fast
  if (RichKeyTableInit(RichChannelStatusTable,"ChannelStatus")) ret++;
  
  // key 2: medium
  if (RichKeyTableInit(       RichASDtable,"ASD"))   ret++;
  //if (RichKeyTableInit(RichM16WireMapTable,"ASD"))   ret++;
  //if (RichKeyTableInit( RichM4WireMapTable,"ASD"))   ret++;

  // key 3: slow
  if (RichKeyTableInit(RichCDIMtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCEFFtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCLENtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCPDWtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCPMDtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCPMItable,"CVES")) ret++;
  if (RichKeyTableInit(RichCPMTtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCPOItable,"CVES")) ret++;
  if (RichKeyTableInit(RichCSMItable,"CVES")) ret++;
  if (RichKeyTableInit(RichCSPMtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCSSMtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCSUPtable,"CVES")) ret++;
  if (RichKeyTableInit(RichCVEStable,"CVES")) ret++;
  return ret;
}

void DAQ_INITIALIZATION::PrintTable(int RichTable,char *TableName) {
  if (RichTable!=-1) cout<<"$ "<<TableName<<": "<<KeyTableMajor(RichTable)<<","<<KeyTableMinor(RichTable)<<endl;
  else               cout<<"$ Keytable for "<<TableName<<" not assigned"<<endl;
}


void DAQ_INITIALIZATION::PrintUsedTables() {

  PrintTable(RichASDtable,"ASD");
  PrintTable(RichCVEStable,"CVES");
  PrintTable(RichChannelStatusTable,"ChannelStatus");
  
}
