#include <riter/RITERMain.hh>
#ifdef RITER_DEBUG
#include "arte/Chrono.hh"
#endif

#include "dq/DQ.h"
#include "dq/DQ_RICH.h"
#include "rhp.h"
#include "riter/RITERhbook.hh"
#define NBIN   256



int RITERRingTableFill(  Rvector<RITERTracks> *RTracks,
		       Rvector<RITERHits>   *RHits,
		       Rvector<RITERTable>  *Rrich ){


  float bin   =.45e-3;                  //! bin size
 
  Rvector<RITERTracks> &tracks = *RTracks;
  Rvector<RITERHits>   &fotons = *RHits;
  Rvector<RITERTable>  &rich   = *Rrich;

  float th_cer=0,fi_cer=0;
  int   itra   = 0;
  int   calcok = 0;
  int   is_inside_flag= 0;
  float sigma         = 0;
  int   iplm          = 0;
  int   isfm          = 0;




  for(Rvector<RITERTracks>::iterator it=tracks.begin();
      it!=tracks.end();++it){    
    int ifot=0;
    for(Rvector<RITERHits>::iterator hit=fotons.begin();
	hit!=fotons.end();++hit){


        calcok=RITERAngleCalculate((*hit),(*it),
				  th_cer,fi_cer,
				  is_inside_flag,
				  sigma, iplm, isfm );

      if (calcok==RITER_OK) {

        register int ir= (int) (th_cer/bin);
	if (ir > (NBIN-1) ) continue;
	register int icha=0;
	register float  utez=(*hit).utez;
	register float  w=(*hit).w;
	if (utez>0) {
	  rich.push_back(
	  RITERTable(ir,itra,ifot,icha,utez,th_cer,fi_cer,w,is_inside_flag, sigma, iplm, isfm ));
	}
      }
      ifot++;
    }         //hits
    itra++;
  }           //tracks
  return RITER_OK;

}


float  RITERIterate(  
		       Rvector<RITERTracks> *RTracks,
		       Rvector<RITERHits>   *RHits, 
		       Rvector<RITERTable>  *Rrich
		       ){

// iterational procedure

  Rvector<RITERTracks> &tracks = *RTracks;
  Rvector<RITERHits>   &fotons = *RHits;
  Rvector<RITERTable>  &rich   = *Rrich;

  int tablesize = rich.size();
  if (!(tablesize>0)) return RITER_OK;
  int nhits = fotons.size();
  int ntra  = tracks.size();
  
  //------------------------------------------------------
  // filling of histograms pf
  //------------------------------------------------------
  float pf[NBIN][ntra];   // histogram (radius,track)
  vzero(&pf[0][0],NBIN*ntra);
  for(Rvector<RITERTable>::iterator hit=rich.begin();
      hit!=rich.end();++hit){
    pf[(*hit).ir][(*hit).itra]+=(*hit).w;
  }

  //------------------------------------------------------
  // filling of  spf
  //------------------------------------------------------

  float spf[nhits];	  // sum of probabilities over tracks for i-th photon
  vzero(spf,nhits);
  
  for(Rvector<RITERTable>::iterator hit=rich.begin();
      hit!=rich.end();++hit){
    spf[(*hit).ifot]+=pf[(*hit).ir][(*hit).itra];
  }

  //------------------------------------------------------
  // new weigths;
  //------------------------------------------------------

  float wdiff=0;
  for(Rvector<RITERTable>::iterator hit=rich.begin();hit!=rich.end();++hit){
    int ifot   =(*hit).ifot;
    int ir     =(*hit).ir;
    int itra   =(*hit).itra;
    float utez =(*hit).utez;
    float w_old=(*hit).w;
    float w=0;
    if(spf[ifot]>0) w=pf[ir][itra]/spf[ifot]*utez;
    wdiff=max(wdiff,fabs(w-w_old));
    (*hit).w = w;
  }
  return wdiff;
}

int RITERLikelihoods(
		       Rvector<RITERTracks> *RTracks,
		       Rvector<RITERHits>   *RHits, 
		       Rvector<RITERTable>  *Rrich
		       ){
  Rvector<RITERTracks> &tracks = *RTracks;
  Rvector<RITERHits>   &fotons = *RHits;
  Rvector<RITERTable>  &rich   = *Rrich;
  
  // fill number of photons first
  for(Rvector<RITERTable>::iterator hit=rich.begin();
      hit!=rich.end();++hit){
    int   flag =  (*hit).flag;
    if (!flag) continue;
    int   itra = (*hit).itra;
    float wi   = (*hit).w;
    
    int type = ( flag > 0) ? 2 : 3; 
    tracks[itra].Likelihood()->Fill(type,wi);
    
  }  
  
  
  /// Calculate background photons
  for (Rvector<RITERTracks>::iterator itra=tracks.begin();
       itra!=tracks.end();++itra){
    (*itra).Likelihood()->SetBackgroundPhotons();
  }
  
  ///! fill log likelihoods
  const float pcut = 15; // momentum cut for histos 
  for(Rvector<RITERTable>::iterator hit=rich.begin();
      hit!=rich.end();++hit){
    register float thc   = (*hit).thc;
    register float wi    = (*hit).w;
    register float sigma = (*hit).sigma;
    
    register int   itra  = (*hit).itra;
    register float p     = tracks[itra].p;
    register int   ifot  = (*hit).ifot;
    register int pmtype  = fotons[ifot].Type(); 

    if ( RITERhisto ) { 
      if (p>pcut) {
	register float fic   = (*hit).fic;
	rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 206 + pmtype),thc,wi);
	rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 204 + pmtype),thc,1);
	rhp_hfill(DQ_ID(COMP_RICH, LVL_REC, 1001 +  pmtype),fic,thc,wi);
	if ( RITERalign ) {
	  register int iplm  = (*hit).iplm; 
	  register int isfm  = (*hit).isfm; 
	  rhp_hfill(DQ_ID(COMP_RICH, LVL_REC, 400 + iplm),fic,thc,wi);
	  rhp_hfill(DQ_ID(COMP_RICH, LVL_REC, 300 + isfm),fic,thc,wi);
	}
      }
    }

    int   flag =  ( (*hit).flag > 0 );
    if (!flag) continue;
    tracks[itra].Likelihood()->FillLKH(sigma,thc,wi); 

  }

  ///! calculate extended  likelihood function   
  for (Rvector<RITERTracks>::iterator itra=tracks.begin();
       itra!=tracks.end();++itra){     
    if ( ( RITERhisto ) && ( (*itra).p> pcut ) ) {
      register float expected_photons =  (*itra).stf[3];
      rhp_hf1(DQ_ID(COMP_RICH, LVL_REC, 1003 ),expected_photons,1);      
    }
    (*itra).CalculateLKH();
  }

  return RITER_OK;
}

