#ifndef _riter_tracks_h_
#define _riter_tracks_h_
#include <riter/RITER.hh>
#include <arte/arte.hh>
#include <riter/RITERGeometry.hh>
#include <riter/RITERUtils.hh>
#include <riter/RITERLikelihood.hh>

/// description of a track   
class RITERTracks { 
private:
  RITERLikelihood *likelihood;
public:

  /// RICH entrance point of the track
  RITERVec r;
  /// direction of the track
  RITERVec dir;       
  /// momentum of the track
  float p ;               

  /// likelihoods for hypothesis {bgr=0,e=1,mu=2,pi=3,k=4,p=5} hypothesis
  float plkh[NHIP];
  /// measured number of signal photons  
  float sfot;
  /// expected number of signal photons for  {bgr=0,e=1,mu=2,pi=3,k=4,p=5} hypothesis
  float stf[NHIP];         
  /// fraction  of m16 hits for different hypothesis for geo acceptance !=0
  float acgm16[NHIP];   
  /// geometrical acceptance for  {..}        
  float acg[NHIP];
  /// pointer to the RITER geometry 
  RITERGeometry *geo;
  /// track direction error 
  float sig2[2];
  /// particle type (unknown=0,e=1,mu=2,pi=3,k=4,p=5)
  int  itype;            
  /// mean photon emission point
  RITERVec  r0;        
  /// particle hit on (up, down) spherical  mirror 
  RITERVec  tr1[2];   
  /// direction after reflection (up, down)  spherical  mirror
  RITERVec  tdir1[2]; 
  /// ring centers (up and down detector)
  RITERVec  trc[2]  ;  
  /// which mirror was hit =1 if particle hits upper sph. mirror
  short  wm ; 
  /// track length in the radiator          
  float tpath;
  /// which planar mirror was hit (up and down detector)
  short id_plm[2] ;
  /// which spherical mirror was hit (up and down detector)
  short id_sfm[2];
   
  /// expected cherenkov angles for different hypothesis {0,e,mu,pi,k,p,e+e-} 
  float thc[NHIP];      
  /// expected min cerenkov angles for {...} 
  float thmin[2][NHIP];      
  /// expected max cerenkov angles for {..}  
  float thmax[2][NHIP];      
  ///  sigmas  for different hypothesis {0,e,mu,pi,k,p,e+e-} 
  float sig_mean[NHIP];     
 
  /// pointer to RTRA ARTE bank
  ArtePointer<RTRA> rtra;
  /// calculate track parameters
  int Calculate();
  /// calculate acceptance
  float  Acceptance(float, float *);
  /// calculate expected photons
  int    ExpectedPhotons(float thc,float path,
			 float &stfs, float &acg,float &acg16);
  float  acc_the(float theta);
  /// calculate expected background photons
  float  ExpectedBackground(int i,float *thc,float *sig,float *acg);
  /// print the contents of the class
  int    Print();
  /// expected cherenkov angle of the track with momentum p 
  float  CherenkovAngle(float p,int itype,float ref_ind); 
  /// pointer to  RITERLikelihood class
  RITERLikelihood *Likelihood(); 
  /// calculate extended likelihood 
  int CalculateLKH();
  /// return sigma^2 for tx and ty 
  float Sigma2(int direction){ return sig2[direction]; };
  int Test(int );

  /// constructor  of the class
  RITERTracks(const RITERVec &rtrack, const RITERVec &dirtrack,  
	      float momentum, float sigma2tx, float sigma2ty,  
              int   type=0, 
	      ArtePointer<RTRA> artetrack = ArtePointer<RTRA>() );
  /// destructor of the class
  ~RITERTracks(){ if (likelihood) delete likelihood; };
  /// empty constructor
  RITERTracks(){ };

};


#endif




