#ifndef _L3_STATE_HH_
#define _L3_STATE_HH_
// *****************************************************************************
//
// source:    L3_State.hh
//
// type:      include file
//
// created:   06 Dec 1999
//
// authors:   Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// changes:
//
// *****************************************************************************

#include "L3_Defines.hh"

#include "L3_Event.hh"
#include "L3_Timer.hh"

class L3_State {
public:
  L3_State();
  L3_State(const L3_State&);
  L3_State& operator = (const L3_State&);

  void update(const L3_Event&);
  void read(const L3_Timer&);

  int nEvents() const;
  int nEventsAccepted() const;
  int nSpacepoints() const;
  int nTracks() const;
  float dtPointReco() const;
  float dtTrackReco() const;
  float dtVertexReco() const;
  float dtTrigger() const;

  void print() const;
  void printEVHD() const;
  void printSummary() const;
private:
  int NumberOfEventsProcessed;
  int NumberOfEventsAccepted;
  int TotalNumberOfReconstructedSpacepoints;
  int TotalNumberOfReconstructedTracks;
  float dt_PointReco;
  float dt_TrackReco;
  float dt_VertexReco;
  float dt_Trigger;
};

#endif
