*CMZ :          07/11/96  16.34.53  by  T. Oest
*-- Author :    T. Oest   29/10/96

      SUBROUTINE V0CHI2(TRK1, TRK2, NZ1, ZMUSC1, ANG1,
     +                  NZ2, ZMUSC2, ANG2, PAR,
     +                  CHI2V0, IRET)

************************************************************************
*
* Description: returns new vertex/track parameters or the
*              covariance matrix
*
* Called from:
*
* External references:
*
* Argument in/out description
* INIT     in     = .TRUE.    do initialization
* COVAR    in     = .FALSE.   calculate new parameters
*                 = .TRUE.    calculate covariance matrix
* TRK1     in     first track
* TRK2     in     second track
* NZ1,2    in     number of multiple scattering planes for track 1,2
* ZMUSC1,2 in     zposition of multiple scattering planes for track 1,2
* ANG1,2   in     (multiple scattering angle)**2 for 1 GeV, tx=ty=0
* PAR      out    vertex, track parameters at START
*                 (ZSTART,x,y,tx1,ty1,1/p1,tx2,ty2,1/p2)
* VPAR     out    covariance matrix
* WPAR     out    (VPAR)**-1
* DCHI2    out    increase in CHI2  (COVAR=.FALSE.)
* CHI2V0   out    CHI2 of fit (COVAR=.TRUE.)
* IRET     out    = 0 OK
*
************************************************************************

      IMPLICIT NONE

#include "arte/Arte.fi"
#include "auto/EV.fi"

      LOGICAL          COVAR, INIT
      INTEGER          TRK1, TRK2, IRET, TRK(2), NZ1, NZ2
      DOUBLE PRECISION PAR(9), VPAR(9,9), WPAR(9,9), DCHI2, CHI2V0
      DOUBLE PRECISION ZMUSC1(*), ZMUSC2(*), ANG1(*), ANG2(*)


      INTEGER          I, J, K, L, M, IERROR, INDEX(5,5), IC1, IC2
      LOGICAL          COVMS, SCATT
      DOUBLE PRECISION VTRK(2,5,5), WTRK(2,5,5)
      DOUBLE PRECISION A(2,5,3), B(2,5,3), W(2,5,5)
      DOUBLE PRECISION E(2,3,3), EINV(2,3,3), DD(2,5), DQ(2,3)
      DOUBLE PRECISION D0(3,3), D(2,3,3)
      DOUBLE PRECISION DX(3)
      DOUBLE PRECISION MA(6,3), MB(6,3), MVMS(6,6), MS(2,5,5)
      DOUBLE PRECISION PIN(6), POUT(6), ZOUT, DZ, PREC(4)
      DOUBLE PRECISION MW5(5,5), RW5(5), RW3(3), MW3(3,3)
      DOUBLE PRECISION WA(2,5,3), WB(2,5,3), BE(2,5,3), DSUM

      DOUBLE PRECISION C00(3,3), C11(3,3), C22(3,3)
      DOUBLE PRECISION C10(3,3), C20(3,3), C21(3,3)


      SAVE C00, EINV, D, DD, W, VTRK, WTRK

#include "arte/Arte.ff"
#include "auto/EV.ff"

      DATA INDEX /  1,  2,  4,  7, 11,
     +              2,  3,  5,  8, 12,
     +              4,  5,  6,  9, 13,
     +              7,  8,  9, 10, 14,
     +             11, 12, 13, 14, 15 /

      DATA SCATT /.TRUE./

C-----------------------------------------------------------------
C  initialization
C-----------------------------------------------------------------
      IRET   = 0
      TRK(1) = TRK1
      TRK(2) = TRK2

      IF (PRTRA_HITB(TRK1).EQ.0) THEN
       IC1 = 0
      ELSE
       IC1 = 1
      ENDIF
      IF (PRTRA_HITB(TRK2).EQ.0) THEN
       IC2 = 0
      ELSE
       IC2 = 1
      ENDIF
C-----------------------------------------------------------------
C  covariance matrix
C-----------------------------------------------------------------

       DO I=1,2

         DO J=1,5
         DO K=J,5
          MW5(J,K) = DBLE(FRTRA_CVF(INDEX(J,K), TRK(I)))
          MW5(K,J) = MW5(J,K)
          VTRK(I,J,K) = MW5(J,K)
          VTRK(I,K,J) = MW5(K,J)
         ENDDO
        ENDDO

         CALL DINV(5, MW5, 5, RW5, IERROR)
         IF (IERROR.NE.0) THEN
          IRET = -1
          GOTO 9000
         ENDIF

         DO J=1,5
         DO K=1,5
          WTRK(I,J,K) = MW5(J,K)
         ENDDO
         ENDDO

       ENDDO
c       GOTO 9000

c      ENDIF
C-----------------------------------------------------------------
C  new parameter or covariance matrix ?
C-----------------------------------------------------------------
C-----------------------------------------------------------------
C  swim track parameter to measured tracks
C         --> DD, A, B
C-----------------------------------------------------------------
      PIN(1) = PAR(1)
      PIN(2) = PAR(2)
      PIN(3) = PAR(3)

      DO I=1,2

       PIN(4) = PAR(1+I*3)
       PIN(5) = PAR(2+I*3)
       PIN(6) = PAR(3+I*3)
       ZOUT   = DBLE(FRTRA_ZF(TRK(I)))
       DZ     = 10. * ABS(PIN(1) - ZOUT)
       PREC(1)= SQRT(FRTRA_CVF( 1,TRK(I))) / DZ
       PREC(2)= SQRT(FRTRA_CVF( 3,TRK(I))) / DZ
       PREC(3)= SQRT(FRTRA_CVF( 6,TRK(I))) / DZ
       PREC(4)= SQRT(FRTRA_CVF(10,TRK(I))) / DZ

       IF (I.EQ.1) THEN
        CALL V0FEXT(IC1, SCATT, NZ1, ZMUSC1, ANG1, PIN, ZOUT, PREC, 
     +              POUT, MA, MB, COVMS, MVMS, IERROR)
       ELSE
        CALL V0FEXT(IC2, SCATT, NZ2, ZMUSC2, ANG2, PIN, ZOUT, PREC, 
     +              POUT, MA, MB, COVMS, MVMS, IERROR)
       ENDIF
       IF (IERROR.NE.0) THEN
        IRET = -2
        GOTO 9000
       ENDIF

       DO J=1,5
       DO K=1,5
         MS(I,J,K) = MVMS(J+1,K+1)
       ENDDO
       ENDDO

       DD(I,1) = FRTRA_XF (TRK(I)) - POUT(2)
       DD(I,2) = FRTRA_YF (TRK(I)) - POUT(3)
       DD(I,3) = FRTRA_TXF(TRK(I)) - POUT(4)
       DD(I,4) = FRTRA_TYF(TRK(I)) - POUT(5)
       DD(I,5) = FRTRA_PF (TRK(I)) - POUT(6)


      ENDDO

C-----------------------------------------------------------------
C  fill matrices
C-----------------------------------------------------------------

C ---   W1, W2

        DO I=1,2

         IF (COVMS) THEN
          DO J=1,5
          DO K=J,5
           MW5(J,K) = VTRK(I,J,K) + MS(I,J,K)
           MW5(K,J) = MW5(J,K)
          ENDDO
          ENDDO

          CALL DINV(5, MW5, 5, RW5, IERROR)
          IF (IERROR.NE.0) THEN
           IRET = -3
           GOTO 9000
          ENDIF

          DO J=1,5
          DO K=1,5
           W(I,J,K) = MW5(J,K)
          ENDDO
          ENDDO

         ELSE
          DO J=1,5
          DO K=1,5
           W(I,J,K) = WTRK(I,J,K)
          ENDDO
          ENDDO
         ENDIF


        ENDDO

C-----------------------------------------------------------------
C chi2
C-----------------------------------------------------------------


      CHI2V0 = 0.0
      DO I=1,5
      DO J=1,5
       CHI2V0 = CHI2V0 + DD(1,I)*W(1,I,J)*DD(1,J)
     +                 + DD(2,I)*W(2,I,J)*DD(2,J)
      ENDDO
      ENDDO


C-----------------------------------------------------------------
C  the end
C-----------------------------------------------------------------


 9000 END

************************************************************************
* End of V0FITI
************************************************************************
