*CMZ :          06/11/96  12.55.46  by  T. Oest
*-- Author :    T. Oest   24/10/96

      SUBROUTINE V0FEXT(ICHRG, SCATT, NZ, ZMUSC, ANG,
     +                  PIN, ZOUT, PREC, POUT, MA, MB, 
     +                  COVAR, MVMS, IERROR)

************************************************************************
*
* Description: extrapolate a track to a new z value. This routine
*              is written for the transport of track parameters
*              from a vertex position to the first track hit.
*
*              walls and detector positions are not yet treated
*              properly in subroutines V0WALL and V0GEDE
*
*              the inhomogenoeus field is used for the region 
*              ZMAG +- DZMAG
*
* Called from:
*
* External references:
*
* Argument in/out description
* ICHRG    in     particle charge
* SCATT    in     = .TRUE.  with multiple scattering
* NZ       in     number of multiple scattering planes
* ZMUSC    in     zposition of multiple scattering planes
* ANG      in     (multiple scattering angle)**2 for 1 GeV, tx=ty=0
* PIN(6)   in     input parameters (z,x,y,tx,ty,rho)
* ZOUT     in     z value for output parameters
* PREC     in     precision required for the transport
* POUT(6)  out    output parameters (z,x,y,tx,ty,rho)
* MA(6,3)  out    derivatives d POUT / d PIN(z,x,y)
* MB(6,3)  out    derivatives d POUT / d PIN(tx,ty,rho)
* COVAR    out    = .TRUE. if MVMS.ne.0
* MVMS(6,6)out    covariance matrix for multiple scattering at ZOUT
* IERROR   out    = 0 OK
*
*
************************************************************************

      IMPLICIT NONE

#include "arte/Arte.fi"
#include "auto/GEO.fi"
#include "gean/ZGESL.fi"
#include "gean/ZGBPIP.fi"

      LOGICAL          SCATT
      INTEGER          ICHRG, IERROR, NZ
      DOUBLE PRECISION PIN(6), POUT(6), ZOUT, PREC(4)
      DOUBLE PRECISION MA(6,3), MB(6,3), MVMS(6,6)
      DOUBLE PRECISION ZMUSC(*), ANG(*)


      INTEGER          I, J, K, L, IRET, ICOVAR
      LOGICAL          FIRST
      LOGICAL          MHIT, COVAR
      DOUBLE PRECISION MVMSO(6,6), MSERR, MSWALL, MSGEDE, MSPIPE
      DOUBLE PRECISION FSTEP, Z, PTMP(6), RKD(6,6), RKDTOT(6,6)
      DOUBLE PRECISION RKH(6,6)
      DOUBLE PRECISION DZMAG, ZMAG, ZMMIN, ZMMAX
      DOUBLE PRECISION ZMAT, MSMAT

      SAVE FIRST, ZMMIN, ZMMAX

#include "arte/Arte.ff"
#include "auto/GEO.ff"

      DATA DZMAG / 280. /
      DATA FIRST /.TRUE./
*-----------------------------------------------------------------------
* first call
*-----------------------------------------------------------------------
      IF (FIRST) THEN
       FIRST = .FALSE.
       ZMAG  = -10000.

        IF (NGESL().GT.0) THEN
         DO I=1,NGESL()
          IF (IGESL_CMP(I) .EQ. LGESL_FLD) THEN
            ZMAG  = DBLE(FGESL_Z1(I))
          ENDIF
         ENDDO
        ENDIF

       IF (ZMAG.EQ.-10000.) THEN
        ZMMIN = -99999.
        ZMMAX =  99999.
       ELSE
        ZMMIN = ZMAG - DZMAG
        ZMMAX = ZMAG + DZMAG
       ENDIF

      ENDIF
*-----------------------------------------------------------------------
* initialization
*-----------------------------------------------------------------------
c      print*,'v0fext ',pin(1),zout
      IERROR = 0
      COVAR  = .FALSE.
      MHIT   = .FALSE.
      FSTEP  = 20.
      DO I=1,6
       PTMP(I) = PIN(I)
       DO J=1,6
        MVMS(I,J)   = 0.0D0
        RKDTOT(I,J) = 0.0D0
       ENDDO
       RKDTOT(I,I) = 1.0
      ENDDO

*-----------------------------------------------------------------------
* loop
*-----------------------------------------------------------------------
      Z = ZOUT

 100  CONTINUE

*-----------------------------------------------------------------------
* next step
*-----------------------------------------------------------------------

c      print*,'z,fs ',sngl(z),sngl(fstep)
c      print*,'pin  ',sngl(ptmp(1)),sngl(ptmp(2)),sngl(ptmp(3))
c      print*,'prec     ',sngl(prec(1)),sngl(prec(2)),sngl(prec(3)),
c     +                   sngl(prec(4))

      IF ( ICHRG.EQ.0 .OR.
     +     (PTMP(1) .LT. ZMMIN .AND. Z .LT. ZMMIN) .OR.
     +     (PTMP(1) .GT. ZMMAX .AND. Z .GT. ZMMAX)) THEN
       CALL V0LEXT(Z, PTMP, POUT, RKD, IRET)
c       print*,' v0lext '
      ELSE
       ICOVAR = 2
       CALL RKMAI2(ICOVAR,Z, FSTEP, PTMP, PREC, POUT, RKD, IRET)
c       print*,' rkmai2 '
      ENDIF
c      print*,' iret ',iret
c      print*,'pout ',sngl(pout(1)),sngl(pout(2)),sngl(pout(3))

      IF (IRET.NE.0) THEN
       IERROR = -1
       GOTO 9000
      ENDIF

c      print*,' rkd '
c      do i=1,2
c       j = 3*i-3
c      print*,'     ',sngl(rkd(J+1,1)),sngl(rkd(J+1,2)),sngl(rkd(J+1,3))
c      print*,'     ',sngl(rkd(J+2,1)),sngl(rkd(J+2,2)),sngl(rkd(J+2,3))
c      print*,'     ',sngl(rkd(J+3,1)),sngl(rkd(J+3,2)),sngl(rkd(J+3,3))
c      print*,'     ',sngl(rkd(J+1,4)),sngl(rkd(J+1,5)),sngl(rkd(J+1,6))
c      print*,'     ',sngl(rkd(J+2,4)),sngl(rkd(J+2,5)),sngl(rkd(J+2,6))
c      print*,'     ',sngl(rkd(J+3,4)),sngl(rkd(J+3,5)),sngl(rkd(J+3,6))
c      ENDDO
*-----------------------------------------------------------------------
* check for next wall and detector plane
*-----------------------------------------------------------------------
      
      IF (SCATT .AND. NZ.GT.0 .AND. .NOT. MHIT) THEN
       IF (POUT(1).GT.PTMP(1)) THEN
        DO I=1,NZ
         IF (ZMUSC(I).GT.PTMP(1).AND.ZMUSC(I).LT.POUT(1)) THEN
          Z     = ZMUSC(I)
          MSMAT = ANG(I)
          MHIT  = .TRUE.
          GOTO 100
         ENDIF
        ENDDO
       ELSE
        DO I=NZ,1,-1
         IF (ZMUSC(I).LT.PTMP(1).AND.ZMUSC(I).GT.POUT(1)) THEN
          Z     = ZMUSC(I)
          MSMAT = ANG(I)
          MHIT  = .TRUE.
          GOTO 100
         ENDIF
        ENDDO
       ENDIF
      ENDIF
*-----------------------------------------------------------------------
* update covariance matrix
*-----------------------------------------------------------------------

      DO I=1,6
      DO J=1,6
       RKH(I,J) = RKDTOT(I,J)
      ENDDO
      ENDDO

      DO I=1,6
      DO J=1,6
       RKDTOT(I,J) = 0.0
       DO K=1,6
        RKDTOT(I,J) = RKDTOT(I,J) + RKD(I,K)*RKH(K,J)
       ENDDO
      ENDDO
      ENDDO




      IF (COVAR) THEN
       DO I=1,6
       DO J=1,6
        MVMSO(I,J) = MVMS(I,J)
       ENDDO
       ENDDO

       DO I=1,6
       DO J=I,6
        MVMS (I,J) = 0.0
        DO K=1,6
        DO L=1,6
         MVMS(I,J) = MVMS(I,J) + RKD(I,K)*MVMSO(K,L)*RKD(J,L)
        ENDDO
        ENDDO
        MVMS(J,I) = MVMS(I,J)
       ENDDO
       ENDDO
      ENDIF

*-----------------------------------------------------------------------
* add multiple scattering error covariance matrix
*-----------------------------------------------------------------------
      IF (MHIT) THEN
       MHIT  = .FALSE.
       COVAR = .TRUE.
       MSERR = MSMAT * PIN(6)**2
       MVMS(4,4) = MVMS(4,4) + MSERR * (1.+POUT(4)**2)**2
       MVMS(5,5) = MVMS(5,5) + MSERR * (1.+POUT(5)**2)**2
      ENDIF

*-----------------------------------------------------------------------
* next step
*-----------------------------------------------------------------------
      IF (ZOUT.EQ.Z) GOTO 1000


      DO I=1,6
       PTMP(I) = POUT(I)
      ENDDO

      Z = ZOUT

      GOTO 100
*-----------------------------------------------------------------------
* end of transport
*-----------------------------------------------------------------------
 1000 CONTINUE
c      print*,'****** rkdtot *******'
c      print*,rkdtot(1,1), rkdtot(1,2), rkdtot(1,3)
c      print*,rkdtot(2,1), rkdtot(2,2), rkdtot(2,3)
c      print*,rkdtot(3,1), rkdtot(3,2), rkdtot(3,3)
c      print*,rkdtot(4,1), rkdtot(4,2), rkdtot(4,3)
c      print*,rkdtot(5,1), rkdtot(5,2), rkdtot(5,3)
c      print*,rkdtot(6,1), rkdtot(6,2), rkdtot(6,3)
c      print*,''
c      print*,rkdtot(1,4), rkdtot(1,5), rkdtot(1,6)
c      print*,rkdtot(2,4), rkdtot(2,5), rkdtot(2,6)
c      print*,rkdtot(3,4), rkdtot(3,5), rkdtot(3,6)
c      print*,rkdtot(4,4), rkdtot(4,5), rkdtot(4,6)
c      print*,rkdtot(5,4), rkdtot(5,5), rkdtot(5,6)
c      print*,rkdtot(6,4), rkdtot(6,5), rkdtot(6,6)
      DO I=1,6
      DO J=1,3
       MA(I,J) = RKDTOT(I,J)
       MB(I,J) = RKDTOT(I,J+3)
      ENDDO
      ENDDO
*-----------------------------------------------------------------------
* the end
*-----------------------------------------------------------------------
 9000 END

************************************************************************
* End of V0FEXT
************************************************************************
